#include "lp.h"
#include <iostream>
#include <sstream>
#include <algorithm>
#include <ext/functional>
#include <ClpSimplex.hpp>
#include <CoinBuild.hpp>




int LP::lp_solve ()
{
  result = model.initialSolve();

  if (result)
    cerr << endl << "WARNING: An optimal solution to the LP has NOT been found" << endl << endl;

  if (apply)
    updateFlows();

  return result;
}




/*
 * Rearrange e2e flows of given demands to minimize the maximum Utot
 * based on a Linear Program
 * Sd is an (optional) vector including the set of links, for each demand (in the
 * same order as in the demands map),
 * that are allowed to carry the flow of the corresponding demand
 * If capacity is true, the capacity constraints are included (which may
 * make the LP infeasible)
 * Includes the constraint M<=maxUtot
 * Changes are applied only if an optimal solution is found, hence
 * the resulting max Utot < maxUtot
 * In that case, maxUtot is updated to the new max Utot
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
MinMaxUtot::MinMaxUtot(Topology* _T, map<pair<int,int>,double>* _demands, double* _maxUtot, vector< set< pair< int, int > > > Sd, 
                       bool capacity, vector<float> *nodePowerConstr, bool _apply)
  : LP (_T, _apply),
    demands(_demands),
    maxUtot(_maxUtot)
{
  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->GetNumNodes() + 1];

  cumNumNeighbs[0] = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
  	  cumNumNeighbs[u + 1] = cumNumNeighbs[u] + T->Node(u)->GetNumNeighbors();

  // the number of variables is numDemands*numLinks (f^ab(u->v)) + 1 (M)
  Ncol = numDemands*cumNumNeighbs[T->GetNumNodes()]+1;

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  M
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
//      int i=1;
      int i=0;
      for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u << "," << v->Id() << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }
  
  int* colno = new int[Ncol];
  double *coeff = new double[Ncol], *coeff2 = new double[Ncol];
  
  /* set the objective function: min M*/
//   coeff[0] = 1;
//   colno[0] = Ncol;
//   set_obj_fnex(lp, 1, coeff, colno);
  model.setObjectiveCoefficient(Ncol-1, 1);

  /* set the object direction to minimize */
//   set_minim(lp);
  model.setOptimizationDirection(1);
  
//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  /* M < maxUtot (through M <= maxUtot - 0.001) */
//    if (maxUtot < numeric_limits<REAL>::max())
//   {
//     coeff[0] = 1;
//     colno[0] = Ncol;
//     add_constraintex(lp, 1, coeff, colno, LE, maxUtot+0.0001);
//   }

  vector<vector<int> > colno_v(T->GetNumNodes()*numDemands);
  vector<vector<double> > coeff_v(T->GetNumNodes()*numDemands);

  CoinBuild buildObject;

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      int d=0;
      for (FlowDemandsIterator di=demands->begin(); di!=demands->end(); di++, d++) {
      // for (unsigned int d=0; d<numDemands; d++) {
	dist = distance(T->Node(u)->NeighborBegin(), v);
	/* flow leaving node u */
	coeff_v[u*numDemands+d].push_back(1);
// 	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[u*numDemands+d].push_back((cumNumNeighbs[u] + dist)*numDemands + d);
	/* flow entering node v */
	coeff_v[v->Id()*numDemands+d].push_back(-1);
// 	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[v->Id()*numDemands+d].push_back((cumNumNeighbs[u] + dist)*numDemands + d);
	/* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	if ( (u == di->first.second) || (v->Id() == di->first.first) )
// 	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	  model.setColumnUpper((cumNumNeighbs[u] + dist)*numDemands + d, 0.);
//	else
//	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, di->second);
      }
    }

  /* Add the flow conservation constraints one-by-one */
  int k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      k = 0;
      for (int x=0; x<T->GetNumNodes(); x++)
	for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	  if ( T->LinkItrs(T->Node(x),y).IsInTheCollDomainOf(T->LinkItrs(T->Node(u),v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->Rate());
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->NeighborBegin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x] + distance(T->Node(x)->NeighborBegin(),y))*numDemands +i;
	    k += numDemands;
	  }
      coeff[k] = -1;
//       colno[k++] = Ncol;         // -M
      colno[k++] = Ncol-1;         // -M
//       add_constraintex(lp, k, coeff, colno, LE, 0 );
      buildObject.addRow(k, colno, coeff, -COIN_DBL_MAX, 0.);
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because the goal of minimizing M
   * will force to minimize the e2e flows
   */
  if (capacity)
    for (int u=0; u<T->GetNumNodes(); u++)
      for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
	fill(coeff, coeff+numDemands, 1);
	for (int i=0; i<numDemands; i++)
// 	  colno[i] = (cumNumNeighbs[u->info->id] + distance(u->NeighborBegin(),v))*numDemands +i +1;
	  colno[i] = (cumNumNeighbs[u] + distance(T->Node(u)->NeighborBegin(),v))*numDemands +i;

// 	add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
	buildObject.addRow(numDemands, colno, coeff, 0., v->Rate());
      }

  /* constraints to only have non-null flow on the links belonging to Sd */
  if (!Sd.empty())
    for (int d=0; d<numDemands; d++) {
      for (int u=0; u<T->GetNumNodes(); u++)
	for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
	  if (Sd[d].find(pair<int,int>(u,v->Id())) == Sd[d].end()) {
	    dist = distance(T->Node(u)->NeighborBegin(), v);
// 	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	    model.setColumnUpper((cumNumNeighbs[u] + dist)*numDemands + d, 0.);
	  }
    }

  /* constraints on the node power consumption */
  if (nodePowerConstr)
  {
    map<TRadio, set<TLinkItrs> > Lri, Lro;  // maps each radio to the set of incoming/outgoing links using it

    /* compute Lri and Lro for all the radios */
    for (unsigned int u=0; u<T->GetNumNodes(); u++)
      for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
      {
	/* insert u->v in the set Lro of the radio on u */
	Lro[TRadio(u,distance(T->Node(u)->Info()->Radio(0),RadioIfaceIterator(v->Radio())))].insert(T->LinkItrs(T->Node(u),v));
	/* insert u->v in the set Lri of the radio on v */
// 	pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
	int ri = v->Info()->findRadioWithChannel(v->RadioChannel());
// 	if (ch.first)
	if (ri >= 0)
	  Lri[TRadio(v->Id(), ri)].insert(T->LinkItrs(T->Node(u),v));
      }

    int pos_i;
    /* constraints tau_TX(r) + tau_RX(r) <= 1 (for each active radio) and node power constraints */
    for (unsigned int u=0; u<T->GetNumNodes(); u++)
    {
      k=0;

      /* add the constraint tau_TX(r) + tau_RX(r) <= 1 for each active radio */
      for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	if (T->Node(u)->Info()->Radio(i)->Channel())
	{
	  int j1=k;
	  RadioIfaceIterator ri = T->Node(u)->Info()->Radio(i);

	  if (Lro.find(TRadio(u,i)) != Lro.end())
	    for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->Rate());
	      // coeff2 is used for the power consumption constraint
	      fill_n(coeff2+k, numDemands, ((*ri->Rho_Tx())(DbmToW(WMS.TxPowerLevel(l->second->TxPower()))) - ri->Rho_Idle()) / l->second->Rate());
	      pos_i = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i));
	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  int j2=k;

	  if (Lri.find(TRadio(u,i)) != Lri.end())
	    for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->Rate());
	      fill_n(coeff2+k, numDemands, (ri->Rho_Rx() - ri->Rho_Idle()) / l->second->Rate());
	      pos_i = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i));
	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  if (k != j1)
	  {
	    /* make a copy of coeff and colno because add_constraintex sorts the element in the passed vectors */
	    double coeff_c[k-j1];
	    int colno_c[k-j1];

	    copy (coeff+j1, coeff+k, coeff_c);
	    copy (colno+j1, colno+k, colno_c);

// 	    add_constraintex(lp, k-j1, coeff_c, colno_c, LE, 1);
	    buildObject.addRow(k-j1, colno_c, coeff_c, 0., 1.);
	  }

	  /* multiply coefficients by the power consumption (needed for node power constraint) */
// 	  for (int j=j1; j<j2; j++)
// 	    coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	  for (int j=j2; j<k; j++)
// 	    coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);
	}

      /* if node u has active radios, add the node power constraint */
      if ((pos_i=T->Node(u)->Info()->countRadiosWithChannel(0)) < T->Node(u)->Info()->numAvailRadios())
// 	add_constraintex(lp, k, coeff, colno, LE, nodePowerConstr->at(u) - T->pm.BasePowerConsumption - pos_i*T->pm.rho_sleep - (T->node[u].info->numAvailRadios() - pos_i)*T->pm.rho_idle);
// 	buildObject.addRow(k, colno, coeff, 0., nodePowerConstr->at(u) - T->pm.BasePowerConsumption - pos_i*T->pm.rho_sleep - (T->node[u].info->numAvailRadios() - pos_i)*T->pm.rho_idle);
      {
	float constr = nodePowerConstr->at(u) - T->Node(u)->Info()->GetBasePowerConsumption();

	RadioIfaceIterator ri = T->Node(u)->Info()->Radio(0);
	for (int rr = 0; rr < T->Node(u)->Info()->GetNumRadios(); ri++, rr++)
	    constr -= (ri->Channel() ? ri->Rho_Idle() : ri->Rho_Sleep());

	buildObject.addRow(k, colno, coeff2, 0., constr);
      }
    }
  }

  model.addRows(buildObject);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}


void MinMaxUtot::updateFlows ()
{
  if (result)
  {
    cerr << endl << "WARNING: MinMaxUtot: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
    return;
  }

  /* If an optimal solution has been found, then update the flows, the e2e flows and the Utot */

  if (maxUtot)
    *maxUtot = model.objectiveValue();

  double *ptr_var, *ptr_constr;
  unsigned int numDemands = demands->size();

  ptr_var = model.primalColumnSolution();
  ptr_constr = model.primalRowSolution();

  unsigned int linkId = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
    {
      SetFlow (v, 0.);
      ClearFlowDemands (v);

      int i=0;
      for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++)
	if (ptr_var[linkId*numDemands + i] > 0.)
	{
	  SetFlowDemand (v, d->first, ptr_var[linkId*numDemands + i]);
	  SetFlow (v, v->Flow() + ptr_var[linkId*numDemands + i]);
	}
      
      SetUtot (v, ptr_constr[T->GetNumNodes()*numDemands+linkId] + ptr_var[Ncol-1]);

      linkId++;
    }
}
  



/*
 * Rearrange e2e flows to minimize the sum of the flows on the links L of a radio
 * such that no Utot above u0 increases nor exceeds u0
 * If the minimum is zero, returns true (the radio can be turned off)
 */
TurnOffRadio::TurnOffRadio (Topology* _T, map<pair<int,int>,double>* _demands, set<TLinkItrs>* L, double u0, bool _apply)
  : LP (_T, _apply),
    demands(_demands),
    success(false)
{
  bool success = false;

  cout << "TurnOffRadio: set L: ";
  for (set<TLinkItrs>::iterator l=L->begin(); l!=L->end(); l++)
    cout << "(" << l->first->Id() << "," << l->second->Id() << ")[" << l->second->RadioChannel() << "] ";
  cout << endl;


  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();

  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->GetNumNodes() + 1];

  cumNumNeighbs[0] = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
    cumNumNeighbs[u + 1] = cumNumNeighbs[u] + T->Node(u)->GetNumNeighbors();

  // the number of variables is numDemands*numLinks (f^ab(u->v))
  Ncol = numDemands*cumNumNeighbs[T->GetNumNodes()];

  /* find the subset of demands routed on the links in L
   * (moving the other demands is just overhead)
   * subdemands is a map containing ALL the demands, each of which
   * is associated with TRUE if it is routed on the links in L
   */
  map<pair<int,int>,bool> subdemands;

  for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++)
    subdemands[d->first] = false;
  for (set<TLinkItrs>::iterator l=L->begin(); l!=L->end(); l++)
    for (FlowDemandsIterator f=l->second->F_e2eBegin(); f!=l->second->F_e2eEnd(); f++) {
      subdemands[f->first] = true;
      cout << "demand [" << f->first.first << f->first.second << "] true" << endl;
    }

  numSubDemands = count_if(subdemands.begin(), subdemands.end(), __gnu_cxx::compose1(bind2nd(equal_to<bool>(),true), __gnu_cxx::select2nd<map<pair<int,int>, bool>::value_type>() ) );

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  M
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
//       int i=1;
      int i=0;
      for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u << "," << v->Id() << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }

  int* colno = new int[Ncol];
  double* coeff = new double[Ncol];

  /* set the objective function: min sum_l\in L sum_ab\in subDemands f^ab(l) */
  int k = 0;
  for (set<TLinkItrs>::iterator l=L->begin(); l!=L->end(); l++) {
//     int i=1;
    int i=0;
    for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++)
      if (subdemands[d->first]) {
	coeff[k] = 1;
	colno[k++] = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(), l->second))*numDemands + i;
      }
  }

//   set_obj_fnex(lp, k, coeff, colno);

  for (int i=0; i<k; i++)
    model.setObjectiveCoefficient(colno[i], coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);

  model.setOptimizationDirection(1);

//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;

  vector<vector<int> > colno_v(T->GetNumNodes()*numSubDemands);
  vector<vector<double> > coeff_v(T->GetNumNodes()*numSubDemands);

  /* flow conservation constraints  ( |V| * numSubDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
//       int d=1, sd=0;
      int d=0, sd=0;
      for (FlowDemandsIterator di=demands->begin(); di!=demands->end(); di++, d++) 
	/* if the demand is routed on the current link u->v */
	if (subdemands[di->first]) {
	  dist = distance(T->Node(u)->NeighborBegin(), v);
	  /* flow leaving node u */
	  coeff_v[u*numSubDemands+sd].push_back(1);
	  colno_v[u*numSubDemands+sd].push_back((cumNumNeighbs[u] + dist)*numDemands + d);
	  /* flow entering node v */
	  coeff_v[v->Id()*numSubDemands+sd].push_back(-1);
	  colno_v[v->Id()*numSubDemands+sd].push_back((cumNumNeighbs[u] + dist)*numDemands + d);
	  /* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	  if ( (u == di->first.second) || (v->Id() == di->first.first) )
// 	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d, 0.);
	    model.setColumnUpper((cumNumNeighbs[u] + dist)*numDemands + d, 0.);
//	  else
//	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d, di->second);
	  sd++;
      }
      else {
	/* the corresponding variable must be set to the current value */
	dist = distance(T->Node(u)->NeighborBegin(), v);
// 	set_bounds(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d, v->f_e2e[di->first], v->f_e2e[di->first]);
	model.setColumnBounds((cumNumNeighbs[u] + dist)*numDemands + d, v->F_e2e(di->first), v->F_e2e(di->first));
	// DEBUG cout << "set_bounds " << cumNumNeighbs[u->info->id] << " " << dist << " " << numDemands << " " << d << endl;
      }
    }

  /* Add the flow conservation constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++)
      if (subdemands[d->first]) {
	copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
	copy(colno_v[k].begin(), colno_v[k].end(), colno);

	/* if i is the source of the demand d */
	if (i == d->first.first)
// 	  add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	  buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
	/* if i is the destination of the demand d */
	else if (i == d->first.second)
// 	  add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	  buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
	else
// 	  add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	  buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
	k++;
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because of the
   * constraints on the total utilizations
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->NeighborBegin(); v!=u->NeighborEnd(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->NeighborBegin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      k = 0;
      for (int x=0; x<T->GetNumNodes(); x++)
	for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	  if ( T->LinkItrs(T->Node(x),y).IsInTheCollDomainOf(T->LinkItrs(T->Node(u),v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->Rate());
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->NeighborBegin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x] + distance(T->Node(x)->NeighborBegin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, max(v->Utot, u0) );
      buildObject.addRow(k, colno, coeff, 0., max(v->UTot(), u0));
    }

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model.addRows(buildObject);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}

void TurnOffRadio::updateFlows ()
{
  if (result)
    cerr << endl << "WARNING: TurnOffRadio: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;

  /* If an optimal solution has been found that zeroes the flow
   * on all the links of the extracted radio,
   * update the flows, the e2e flows and the Utot */
  if ((!result) && (model.objectiveValue() < almost_null))
  {
    cout << endl << "TurnOffRadio: the LP was able to turn off the radio!" << endl;
    success = true;

    double *ptr_var, *ptr_constr;
    unsigned int numDemands = demands->size();

    ptr_var = model.primalColumnSolution();
    ptr_constr = model.primalRowSolution();

    unsigned int linkId = 0;

    for (int u=0; u<T->GetNumNodes(); u++)
      for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
      {
      SetFlow (v, 0.);
      ClearFlowDemands (v);

	int i=0;
	for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++)
	  if (ptr_var[linkId*numDemands + i])
	  {
	    SetFlowDemand (v, d->first, ptr_var[linkId*numDemands + i]);
	    SetFlow (v, v->Flow() + ptr_var[linkId*numDemands + i]);
	  }

	SetUtot (v, ptr_constr[T->GetNumNodes()*numSubDemands+linkId]);

	linkId++;
      }
  }
}


/*
 * Rearrange e2e flows to minimizes the total power consumption subject
 * to constraints on node power and maximum total utilization.
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
TurnOffRadioPower::TurnOffRadioPower(Topology* _T, map<pair<int,int>,double>* _demands, const TRadio& r0, double u0,
                                     double* _maxUtot, double* _Ptot, vector<float>* nodePowerConstr, bool _apply)
  : LP(_T, _apply),
  demands(_demands),
  maxUtot(_maxUtot),
  Ptot(_Ptot)
{
  powerStandbyNodes = 0.;

  map<TRadio, set<TLinkItrs> > Lri, Lro;  // maps each radio to the set of incoming/outgoing links using it

  /* compute Lri and Lro for all the radios */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
    {
      /* insert u->v in the set Lro of the radio on u */
      Lro[TRadio(u,distance(T->Node(u)->Info()->Radio(0),RadioIfaceIterator(v->Radio())))].insert(T->LinkItrs(T->Node(u),v));
      /* insert u->v in the set Lri of the radio on v */
//       pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
      int ri = v->Info()->findRadioWithChannel(v->RadioChannel());
      if (ri >= 0)
	Lri[TRadio(v->Id(), ri)].insert(T->LinkItrs(T->Node(u),v));
    }

  cout << "TurnOffRadioPower: Outgoing links from r0: ";
  if (Lro.find(r0) != Lro.end())
    for (set<TLinkItrs>::iterator l=Lro[r0].begin(); l!=Lro[r0].end(); l++)
      cout << "(" << l->first->Id() << "," << l->second->Id() << ")[" << l->second->RadioChannel() << "] ";
  cout << endl << "TurnOffRadioPower: Incoming links to r0: ";
  if (Lri.find(r0) != Lri.end())
    for (set<TLinkItrs>::iterator l=Lri[r0].begin(); l!=Lri[r0].end(); l++)
      cout << "(" << l->first->Id() << "," << l->second->Id() << ")[" << l->second->RadioChannel() << "] ";
  cout << endl;


  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();

  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->GetNumNodes() + 1];

  cumNumNeighbs[0] = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
  	  cumNumNeighbs[u + 1] = cumNumNeighbs[u] + T->Node(u)->GetNumNeighbors();

  // the number of variables is numDemands*numLinks (f^ab(u->v))
  Ncol = numDemands*cumNumNeighbs[T->GetNumNodes()];

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);
  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
//       int i=1;
      int i=0;
      for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u << "," << v->Id() << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }

  int* colno = new int[Ncol+1];
  double *coeff = new double[Ncol+1], *coeff2 = new double[Ncol+1];
  
  vector<vector<double> > coeff_tau;
  vector<vector<int> > colno_tau;

  vector<vector<double> > coeff_power;
  vector<vector<int> > colno_power;
  vector<double> lhs_power;

  int k, pos_i;


  /* With one iteration over all the radios, we determine the coefficients used for the objective
   * function, the constraints tau_TX(r) + tau_RX(r) <= 1 (for each active radio) and node power constraints */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    if (T->Node(u)->Info()->countRadiosWithChannel(0) == T->Node(u)->Info()->numAvailRadios())
//       numStandbyNodes++;
      powerStandbyNodes += T->Node(u)->Info()->GetStandbyPowerConsumption();
    else if (u == r0.first && T->Node(u)->Info()->countRadiosWithChannel(0) == T->Node(u)->Info()->numAvailRadios()-1 && T->Node(u)->Info()->Radio(r0.second)->Channel())
//       numStandbyNodes++;
      powerStandbyNodes += T->Node(u)->Info()->GetStandbyPowerConsumption();
    else
    {
      k = 0;

      /* for each active radio */
      for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	if (T->Node(u)->Info()->Radio(i)->Channel() && TRadio(u,i) != r0)
	{
	  int j1=k;
	  RadioIfaceIterator ri = T->Node(u)->Info()->Radio(i);

	  if (Lro.find(TRadio(u,i)) != Lro.end())
	    for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->Rate());
	      // coeff2 is used for the power consumption constraint
	      fill_n(coeff2+k, numDemands, ((*ri->Rho_Tx())(DbmToW(WMS.TxPowerLevel(l->second->TxPower()))) - ri->Rho_Idle()) / l->second->Rate());
	      pos_i = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i));
	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  int j2=k;

	  if (Lri.find(TRadio(u,i)) != Lri.end())
	    for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->Rate());
	      fill_n(coeff2+k, numDemands, (ri->Rho_Rx() - ri->Rho_Idle()) / l->second->Rate());
	      pos_i = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i));
	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  if (k != j1)
	  {
	    /* make a copy of coeff and colno so that later we can add the tau constraint */
	    coeff_tau.push_back(vector<double>(coeff+j1,coeff+k));
	    colno_tau.push_back(vector<int>(colno+j1,colno+k));
	  }

	  /* multiply coefficients by the power consumption (needed for node power constraint) */
// 	  for (int j=j1; j<j2; j++)
// 	    coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	  for (int j=j2; j<k; j++)
// 	    coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);
	}

      /* stores the coefficients for the power constraint on node u */
      if (k)
      {
	pos_i = T->Node(u)->Info()->countRadiosWithChannel(0);
	if (u == r0.first && T->Node(u)->Info()->Radio(r0.second)->Channel())
	  pos_i++;

	coeff_power.push_back(vector<double>(coeff2,coeff2+k));
	colno_power.push_back(vector<int>(colno,colno+k));

	float _lhs = T->Node(u)->Info()->GetBasePowerConsumption();

	RadioIfaceIterator ri = T->Node(u)->Info()->Radio(0);
	for (int rr = 0; rr < T->Node(u)->Info()->GetNumRadios(); ri++, rr++)
	  _lhs += (ri->Channel() ? ri->Rho_Idle() : ri->Rho_Sleep());

// 	lhs_power.push_back(T->pm.BasePowerConsumption + pos_i*T->pm.rho_sleep + (T->node[u].info->numAvailRadios() - pos_i)*T->pm.rho_idle);
	lhs_power.push_back (_lhs);
      }
    }

//   for (int i=1; i<=Ncol; i++)
  for (int i=0; i<Ncol; i++)
    coeff[i] = 0.;               // reset the elements of coeff

  for (unsigned int i=0; i<coeff_power.size(); i++)
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);

  /* set the objective function: min sum_u\in V P(u) */
//   set_obj_fn(lp, coeff);

  for (int i=0; i<Ncol; i++)
    model.setObjectiveCoefficient(i, coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);
  model.setOptimizationDirection(1);


//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      k = 0;
      for (int x=0; x<T->GetNumNodes(); x++)
	for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	  if ( T->LinkItrs(T->Node(x),y).IsInTheCollDomainOf(T->LinkItrs(T->Node(u),v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->Rate());
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->NeighborBegin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x] + distance(T->Node(x)->NeighborBegin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, max(maxUtot, u0) );
      buildObject.addRow(k, colno, coeff, max(*maxUtot, u0), max(*maxUtot, u0));
    }

  /* no flow on the links using radio r0 */
  if (Lro.find(r0) != Lro.end())
    for (set<TLinkItrs>::iterator l=Lro[r0].begin(); l!=Lro[r0].end(); l++)
//       for (unsigned int d=1; d<=numDemands; d++)
      for (unsigned int d=0; d<numDemands; d++)
// 	set_upbo(lp, (cumNumNeighbs[l->first->info->id] + distance(l->first->NeighborBegin(),l->second))*numDemands + d, 0.);
	model.setColumnUpper((cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands + d, 0.);

  if (Lri.find(r0) != Lri.end())
    for (set<TLinkItrs>::iterator l=Lri[r0].begin(); l!=Lri[r0].end(); l++)
//       for (unsigned int d=1; d<=numDemands; d++)
      for (unsigned int d=0; d<numDemands; d++)
// 	set_upbo(lp, (cumNumNeighbs[l->first->info->id] + distance(l->first->NeighborBegin(),l->second))*numDemands + d, 0.);
	model.setColUpper((cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands + d, 0.);


//   for (int i=1; i<=Ncol; i++)
  for (int i=0; i<Ncol; i++)
  {
    coeff[i] = 0.;               // reset the elements of coeff
    colno[i] = i;
  }

  /* constraint on the total power consumption */
  lhs = 0.;
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);
    lhs += lhs_power.at(i);
  }

//   add_constraint(lp, coeff, LE, Ptot - lhs - numStandbyNodes*T->pm.StandbyPowerConsumption);
  buildObject.addRow(Ncol, colno, coeff, -COIN_DBL_MAX, *Ptot - lhs - powerStandbyNodes);


  /* tau constraints */
  for (unsigned int r=0; r<coeff_tau.size(); r++)
  {
    copy (coeff_tau.at(r).begin(), coeff_tau.at(r).end(), coeff);
    copy (colno_tau.at(r).begin(), colno_tau.at(r).end(), colno);

//     add_constraintex(lp, coeff_tau.at(r).size(), coeff, colno, LE, 1);
    buildObject.addRow(coeff_tau.at(r).size(), colno, coeff, -COIN_DBL_MAX, 1.);
  }

  /* node power constraints */
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    copy (coeff_power.at(i).begin(), coeff_power.at(i).end(), coeff);
    copy (colno_power.at(i).begin(), colno_power.at(i).end(), colno);

//     add_constraintex(lp, coeff_power.at(i).size(), coeff, colno, LE, nodePowerConstr->at(i) - lhs_power.at(i));
    buildObject.addRow(coeff_power.at(i).size(), colno, coeff, -COIN_DBL_MAX, nodePowerConstr->at(i) - lhs_power.at(i));
  }


  vector<vector<int> > colno_v(T->GetNumNodes()*numDemands);
  vector<vector<double> > coeff_v(T->GetNumNodes()*numDemands);

  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      int d=0;
      for (map<pair<int,int>,double>::iterator di=demands->begin(); di!=demands->end(); di++, d++) {
      // for (unsigned int d=0; d<numDemands; d++) {
	dist = distance(T->Node(u)->NeighborBegin(), v);
	/* flow leaving node u */
	coeff_v[u*numDemands+d].push_back(1);
// 	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[u*numDemands+d].push_back((cumNumNeighbs[u] + dist)*numDemands + d);
	/* flow entering node v */
	coeff_v[v->Id()*numDemands+d].push_back(-1);
// 	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	colno_v[v->Id()*numDemands+d].push_back((cumNumNeighbs[u] + dist)*numDemands + d);
	/* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	if ( (u == di->first.second) || (v->Id() == di->first.first) )
// 	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
	  model.setColumnUpper((cumNumNeighbs[u] + dist)*numDemands + d, 0.);
//	else
//	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, di->second);
      }
    }

  /* Add the flow conservation constraints one-by-one */
  k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->GetNumNodes(); i++)
    for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first.first)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, d->second, d->second);
      /* if i is the destination of the demand d */
      else if (i == d->first.second)
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, -d->second);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, -d->second, -d->second);
      else
// 	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
	buildObject.addRow(coeff_v[k].size(), colno, coeff, 0., 0.);
      k++;
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because of the
   * constraints on the total utilizations
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->NeighborBegin(); v!=u->NeighborEnd(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->NeighborBegin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

//   set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  model.addRows(buildObject);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}

void TurnOffRadioPower::updateFlows ()
{
  if (result)
  {
    cerr << endl << "WARNING: TurnOffRadioPower: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
    return;
  }

  /* If an optimal solution has been found, then update the flows, the e2e flows and the Utot */

  cout << endl << "TurnOffRadioPower: the LP was able to turn off the radio!" << endl;

  double *ptr_var, *ptr_constr;
  unsigned int numDemands = demands->size();

  ptr_var = model.primalColumnSolution();
  ptr_constr = model.primalRowSolution();

  *maxUtot = 0.;

  unsigned int linkId = 0;

  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
    {
      SetFlow (v, 0.);
      ClearFlowDemands (v);

      int i=0;
      for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++)
	if (ptr_var[linkId*numDemands + i] > 0.)
	{
	  SetFlowDemand (v, d->first, ptr_var[linkId*numDemands + i]);
	  SetFlow (v, v->Flow() + ptr_var[linkId*numDemands + i]);
	}

      SetUtot (v, ptr_constr[linkId]);
      *maxUtot = max(*maxUtot, v->UTot());

      linkId++;
    }

//     Ptot = ptr_constr[linkId] + lhs + numStandbyNodes*T->pm.StandbyPowerConsumption;
  *Ptot = ptr_constr[linkId] + lhs + powerStandbyNodes;
}




/*
 * Rearrange e2e flows to minimizes the total power consumption subject
 * to constraints on node power and maximum total utilization.
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
MinPower_eps::MinPower_eps(Topology* _T, map<pair<int,int>,double>* _demands, double u0, vector<float>* nodePowerConstr,
                           const vector< ArcEpsY >& epsy, bool _apply)
  : LP (_T, _apply),
    demands(_demands)
{
  map<TRadio, set<TLinkItrs> > Lri, Lro;  // maps each radio to the set of incoming/outgoing links using it

  /* compute Lri and Lro for all the radios */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
    {
      /* insert u->v in the set Lro of the radio on u */
      Lro[TRadio(u,distance(T->Node(u)->Info()->Radio(0),RadioIfaceIterator(v->Radio())))].insert(T->LinkItrs(T->Node(u),v));
      /* insert u->v in the set Lri of the radio on v */
//       pair<bool,unsigned int> ch = SearchChannel(v->info, v->radio->channel);
      int ri = v->Info()->findRadioWithChannel(v->RadioChannel());
//       if (ch.first)
      if (ri >= 0)
	Lri[TRadio(v->Id(), ri)].insert(T->LinkItrs(T->Node(u),v));
    }

  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands->size();

  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->GetNumNodes() + 1];

  // compute the cumulative number of arcs of type node or radio for each node. It is such that:
  // - the number of arcs of node i is cumNumArcs[i+1] - cumNumArcs[i]
  // - the last element of cumNumArcs is the total number of arcs of type node or radio
  unsigned int cumNumArcs[T->GetNumNodes() + 1];

  cumNumNeighbs[0] = 0;
  cumNumArcs[0] = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
  {
    cumNumNeighbs[u + 1] = cumNumNeighbs[u] + T->Node(u)->GetNumNeighbors();

    int radiosOn = T->Node(u)->Info()->numAvailRadios() - T->Node(u)->Info()->countRadiosWithChannel(0);
    cumNumArcs[u + 1] = cumNumArcs[u] + (radiosOn ? 2 + 2*radiosOn : 0);
  }

  // the number of variables is numDemands*numLinks (f^ab(u->v))
  Ncol = numDemands*(cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[T->GetNumNodes()]);

  // create the LP problem
//   if ((lp = make_lp(0, Ncol)) == NULL)
//     throw -1;

  model.resize(0, Ncol);

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  f_u0->u0*,d1 ... f_u0->u0*,dD f_u0*->u0,d1 ... f_u0*->u0,dD f_u0*->r^(u0,c1),d1 ... f_u0*->r^(u0,c1),dD ...
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the link variables */
  int linkId = 0;
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
//       int i=1;
      int i=0;
      for (FlowDemandsIterator d=demands->begin(); d!=demands->end(); d++, i++) {
	stringstream s;
	s << "f(" << u << "," << v->Id() << ")[" << d->first.first << d->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+i, str);
      }
      linkId++;
    }

  /* give a name to the arc variables */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    if (T->Node(u)->Info()->countRadiosWithChannel(0) != T->Node(u)->Info()->numAvailRadios())
    {
//       int d=1;
      int d=0;
      for (FlowDemandsIterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
      {
	stringstream s;
	s << "f(" << u << "," << u << "*)[" << dem->first.first << dem->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+d, str);
      }
      linkId++;

//       d=1;
      d=0;
      for (FlowDemandsIterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
      {
	stringstream s;
	s << "f(" << u << "*," << u << ")[" << dem->first.first << dem->first.second << "]";
// 	set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	string str = s.str();
	model.setColumnName(linkId*numDemands+d, str);
      }
      linkId++;

      /* for each active radio */
      for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	if (T->Node(u)->Info()->Radio(i)->Channel())
	{
// 	  d=1;
	  d=0;
	  for (FlowDemandsIterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
	  {
	    stringstream s;
	    s << "f(" << u << "*,r" << u << ":" << T->Node(u)->Info()->Radio(i)->Channel() << ")[" << dem->first.first << dem->first.second << "]";
// 	    set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	    string str = s.str();
	    model.setColumnName(linkId*numDemands+d, str);
	  }
	  linkId++;

// 	  d=1;
	  d=0;
	  for (FlowDemandsIterator dem=demands->begin(); dem!=demands->end(); dem++, d++)
	  {
	    stringstream s;
	    s << "f(r" << u << ":" << T->Node(u)->Info()->Radio(i)->Channel() << "," << u << "*" << ")[" << dem->first.first << dem->first.second << "]";
// 	    set_col_name(lp, linkId*numDemands+d, (char*)s.str().c_str());
	    string str = s.str();
	    model.setColumnName(linkId*numDemands+d, str);
	  }
	  linkId++;
	}
    }


  int* colno = new int[Ncol+1];
  double *coeff = new double[Ncol+1], *coeff2 = new double[Ncol+1];

  vector<vector<double> > coeff_tau;
  vector<vector<int> > colno_tau;

  vector<vector<double> > coeff_power;
  vector<vector<int> > colno_power;
  vector<double> lhs_power;

  int k, pos_i;

  /* With one iteration over all the radios, we determine the coefficients used for the objective
   * function, the constraints tau_TX(r) + tau_RX(r) <= 1 (for each active radio) and node power constraints */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    if (T->Node(u)->Info()->countRadiosWithChannel(0) != T->Node(u)->Info()->numAvailRadios())
    {
      k = 0;
      int r = 0;

      float lhs = 0.f;

      /* for each active radio */
      for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	if (T->Node(u)->Info()->Radio(i)->Channel())
	{
	  int j1=k;
	  RadioIfaceIterator ri = T->Node(u)->Info()->Radio(i);

	  if (Lro.find(TRadio(u,i)) != Lro.end())
	    for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->Rate());
	      // coeff2 is used for the power consumption constraint
	      fill_n(coeff2+k, numDemands, ((*ri->Rho_Tx())(DbmToW(WMS.TxPowerLevel(l->second->TxPower()))) - ri->Rho_Idle()) / l->second->Rate());
	      pos_i = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i));
	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  int j2=k;

	  if (Lri.find(TRadio(u,i)) != Lri.end())
	    for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	    {
	      fill_n(coeff+k, numDemands, 1. / l->second->Rate());
	      fill_n(coeff2+k, numDemands, (ri->Rho_Rx() - ri->Rho_Idle()) / l->second->Rate());
	      pos_i = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands;
// 	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i));
	      generate_n(colno+k, numDemands, WMLP::counter<int>(pos_i-1));
	      k += numDemands;
	    }

	  if (k != j1)
	  {
	    /* make a copy of coeff and colno so that later we can add the tau constraint */
	    coeff_tau.push_back(vector<double>(coeff+j1,coeff+k));
	    colno_tau.push_back(vector<int>(colno+j1,colno+k));
	  }

	  /* multiply coefficients by the power consumption (needed for node power constraint) */
// 	  for (int j=j1; j<j2; j++)
// 	    coeff[j] *= (T->pm.rho_tx-T->pm.rho_idle);
// 
// 	  for (int j=j2; j<k; j++)
// 	    coeff[j] *= (T->pm.rho_rx-T->pm.rho_idle);

	  /* add the contribution of arcs of type radio */
	  if (epsy.at(u).yR.at(i))
// 	    lhs += T->pm.rho_idle;
	    lhs += ri->Rho_Idle();
	  else
	  {
	    pos_i = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 2 + 2*r) * numDemands;
	    generate_n(colno+k, numDemands*2, WMLP::counter<int>(pos_i-1));
// 	    fill_n(coeff+k, numDemands*2, (T->pm.rho_idle - T->pm.rho_sleep)/epsy.at(u).epsR.at(i));
	    fill_n(coeff2+k, numDemands*2, (ri->Rho_Idle() - ri->Rho_Sleep())/epsy.at(u).epsR.at(i));
	    k += numDemands*2;

// 	    lhs += T->pm.rho_sleep;
	    lhs += ri->Rho_Sleep();
	  }

	  r++;
	}

      /* stores the coefficients for the power constraint on node u */
      if (k)
      {
// 	float node_rho_on = T->pm.BasePowerConsumption + T->pm.rho_sleep * (T->node[u].info->countRadiosWithChannel(0)),
// 	      node_rho_off = T->pm.StandbyPowerConsumption - T->pm.rho_sleep * (T->node[u].info->numAvailRadios() - T->node[u].info->countRadiosWithChannel(0));

	float node_rho_on = T->Node(u)->Info()->GetBasePowerConsumption();
	float node_rho_off = T->Node(u)->Info()->GetStandbyPowerConsumption();

	RadioIfaceIterator ri = T->Node(u)->Info()->Radio(0);
	for (int rr = 0; rr < T->Node(u)->Info()->GetNumRadios(); ri++, rr++)
	{
	  node_rho_on += (ri->Channel() ? 0 : ri->Rho_Sleep());
	  node_rho_off -= (ri->Channel() ? ri->Rho_Sleep() : 0);
	}

	/* add the contribution of arcs of type node */
	if (epsy.at(u).yN)
	  lhs += node_rho_on;
	else
	{
	  pos_i = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u]) * numDemands;
// 	  generate_n(colno+k, numDemands*2, WMLP::counter<int>(pos_i));
	  generate_n(colno+k, numDemands*2, WMLP::counter<int>(pos_i-1));
	  fill_n(coeff2+k, numDemands*2, (node_rho_on - node_rho_off)/epsy.at(u).epsN);
	  k += numDemands*2;

	  lhs += node_rho_off;
	}

	coeff_power.push_back(vector<double>(coeff2,coeff2+k));
	colno_power.push_back(vector<int>(colno,colno+k));
	lhs_power.push_back(lhs);
      }
    }

  for (int i=0; i<=Ncol; i++)
    coeff[i] = 0.;               // reset the elements of coeff

  for (unsigned int i=0; i<coeff_power.size(); i++)
    for (unsigned int j=0; j<coeff_power.at(i).size(); j++)
      coeff[colno_power.at(i).at(j)] += coeff_power.at(i).at(j);

  /* set the objective function: min sum_u\in V P(u) */
//   set_obj_fn(lp, coeff);

  for (int i=0; i<Ncol; i++)
    model.setObjectiveCoefficient(i, coeff[i]);

  /* set the object direction to minimize */
//   set_minim(lp);
  model.setOptimizationDirection(1);

//   set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  CoinBuild buildObject;

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
      k = 0;
      for (int x=0; x<T->GetNumNodes(); x++)
	for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	  if ( T->LinkItrs(T->Node(x),y).IsInTheCollDomainOf(T->LinkItrs(T->Node(u),v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->Rate());
	    for (unsigned int i=0; i<numDemands; i++)
// 	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->NeighborBegin(),y))*numDemands +i +1;
	      colno[k+i] = (cumNumNeighbs[x] + distance(T->Node(x)->NeighborBegin(),y))*numDemands +i;
	    k += numDemands;
	  }
//       add_constraintex(lp, k, coeff, colno, LE, u0);
      buildObject.addRow(k, colno, coeff, 0., u0+0.001);
    }

  /* tau constraints */
  for (unsigned int r=0; r<coeff_tau.size(); r++)
  {
    copy (coeff_tau.at(r).begin(), coeff_tau.at(r).end(), coeff);
    copy (colno_tau.at(r).begin(), colno_tau.at(r).end(), colno);

//     add_constraintex(lp, coeff_tau.at(r).size(), coeff, colno, LE, 1);
    buildObject.addRow(coeff_tau.at(r).size(), colno, coeff, 0., 1.);
  }

  /* node power constraints */
  for (unsigned int i=0; i<coeff_power.size(); i++)
  {
    copy (coeff_power.at(i).begin(), coeff_power.at(i).end(), coeff);
    copy (colno_power.at(i).begin(), colno_power.at(i).end(), colno);

//     add_constraintex(lp, coeff_power.at(i).size(), coeff, colno, LE, nodePowerConstr->at(i) - lhs_power.at(i));
    buildObject.addRow(coeff_power.at(i).size(), colno, coeff, 0., nodePowerConstr->at(i) - lhs_power.at(i));
  }

  /* flow conservation constraints on the vertices of type radio r^(u,c) */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    if (T->Node(u)->Info()->countRadiosWithChannel(0) != T->Node(u)->Info()->numAvailRadios())
    {
      int r = 0;

      /* for each active radio */
      for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	if (T->Node(u)->Info()->Radio(i)->Channel())
	{
	  // sum of the flow outgoing from r^(u,c)
// 	  for (unsigned int d=1; d<=numDemands; d++)
	  for (unsigned int d=0; d<numDemands; d++)
	  {
	    k = 0;
	    if (Lro.find(TRadio(u,i)) != Lro.end())
	      for (set<TLinkItrs>::iterator l=Lro[TRadio(u,i)].begin(); l!=Lro[TRadio(u,i)].end(); l++)
	      {
		colno[k] = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands + d;
		coeff[k++] = 1.;
	      }
	    colno[k] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 2 + 2*r)*numDemands + d;
	    coeff[k++] = -1.;

// 	    add_constraintex(lp, k, coeff, colno, EQ, 0.);
	    buildObject.addRow(k, colno, coeff, 0., 0.);
	  }

	  // sum of the flow incoming to r^(u,c)
// 	  for (unsigned int d=1; d<=numDemands; d++)
	  for (unsigned int d=0; d<numDemands; d++)
	  {
	    k = 0;
	    if (Lri.find(TRadio(u,i)) != Lri.end())
	      for (set<TLinkItrs>::iterator l=Lri[TRadio(u,i)].begin(); l!=Lri[TRadio(u,i)].end(); l++)
	      {
		colno[k] = (cumNumNeighbs[l->first->Id()] + distance(l->first->NeighborBegin(),l->second))*numDemands + d;
		coeff[k++] = 1.;
	      }
	    colno[k] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 3 + 2*r)*numDemands + d;
	    coeff[k++] = -1.;

// 	    add_constraintex(lp, k, coeff, colno, EQ, 0.);
	    buildObject.addRow(k, colno, coeff, 0., 0.);
	  }

	  r++;
	}
    }

  /* flow conservation constraints on the vertices of type node u* */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    if (T->Node(u)->Info()->countRadiosWithChannel(0) != T->Node(u)->Info()->numAvailRadios())
//       for (unsigned int d=1; d<=numDemands; d++)
      for (unsigned int d=0; d<numDemands; d++)
      {
	int r = 0;
	k = 0;

	/* for each active radio */
	for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	  if (T->Node(u)->Info()->Radio(i)->Channel())
	  {
	    colno[k] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 2 + 2*r)*numDemands + d;
	    coeff[k++] = 1.;
	    r++;
	  }
	colno[k] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u])*numDemands + d;
	coeff[k++] = -1.;

// 	add_constraintex(lp, k, coeff, colno, EQ, 0.);
	buildObject.addRow(k, colno, coeff, 0., 0.);

	r = 0;
	k = 0;

	/* for each active radio */
	for (unsigned int i=0; i<T->Node(u)->Info()->numAvailRadios(); i++)
	  if (T->Node(u)->Info()->Radio(i)->Channel())
	  {
	    colno[k] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 3 + 2*r)*numDemands + d;
	    coeff[k++] = 1.;
	    r++;
	  }
	colno[k] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 1)*numDemands + d;
	coeff[k++] = -1.;

// 	add_constraintex(lp, k, coeff, colno, EQ, 0.);
	buildObject.addRow(k, colno, coeff, 0., 0.);
      }

  /* flow conservation constraints on nodes */
  for (unsigned int u=0; u<T->GetNumNodes(); u++)
    if (T->Node(u)->Info()->countRadiosWithChannel(0) != T->Node(u)->Info()->numAvailRadios())
    {
//       int d = 1;
      int d = 0;
      for (FlowDemandsIterator dem = demands->begin(); dem != demands->end(); dem++, d++)
      {
	colno[0] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u])*numDemands + d;
	coeff[0] = 1.;

	colno[1] = (cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 1)*numDemands + d;
	coeff[1] = -1.;

	/* if u is the source of the demand */
	if (u == dem->first.first)
	{
// 	  add_constraintex(lp, 2, coeff, colno, EQ, dem->second);
	  buildObject.addRow(2, colno, coeff, dem->second, dem->second);
//	  set_upbo(lp, (cumNumNeighbs[T->node.size()] + cumNumArcs[u] + 1)*numDemands + d, 0.);
	  model.setColumnUpper((cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u] + 1)*numDemands + d, 0.);
	}
	/* if u is the destination of the demand */
	else if (u == dem->first.second)
	{
// 	  add_constraintex(lp, 2, coeff, colno, EQ, -dem->second);
	  buildObject.addRow(2, colno, coeff, -dem->second, -dem->second);
//	  set_upbo(lp, (cumNumNeighbs[T->node.size()] + cumNumArcs[u])*numDemands + d, 0.);
	  model.setColumnUpper((cumNumNeighbs[T->GetNumNodes()] + cumNumArcs[u])*numDemands + d, 0.);
	}
	else
// 	  add_constraintex(lp, 2, coeff, colno, EQ, 0.);
	  buildObject.addRow(2, colno, coeff, 0., 0.);
      }
    }

  model.addRows(buildObject);

  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;
}


void MinPower_eps::updateFlows ()
{
  if (result)
  {
    cerr << endl << "WARNING: MinPower_eps: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
    return;
  }

  /* If an optimal solution has been found, then update the flows, the e2e flows and the Utot */

  double *ptr_var, *ptr_constr;
  unsigned int numDemands = demands->size();

  ptr_var = model.primalColumnSolution();
  ptr_constr = model.primalRowSolution();

  unsigned int linkId = 0;

  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
    {
      SetUtot (v, ptr_constr[linkId]);
      SetFlow (v, 0.);
      ClearFlowDemands (v);

      int i=0;
      for (FlowDemandsIterator dem = demands->begin(); dem != demands->end(); dem++, i++)
	if (ptr_var[linkId*numDemands + i] > 0.)
	{
	  SetFlowDemand (v, dem->first, ptr_var[linkId*numDemands + i]);
	  SetFlow (v, v->Flow() + ptr_var[linkId*numDemands + i]);
	}

      linkId++;
    }
}


/* Given the initial graph, builds an LP to solve the 
   maximum (throughput) multi-commodity flow problem
*/
MaxMultiCommFlow::MaxMultiCommFlow(Topology* _T, bool _apply)
  : LP (_T, _apply)
{
	// compute the cumulative number of neighbors for each node. It is such that:
	// - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
	// - the last element of cumuNumNeighbs is the total number of links
	unsigned int cumNumNeighbs[T->GetNumNodes() + 1];

	cumNumNeighbs[0] = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
	{
		cumNumNeighbs[u + 1] = cumNumNeighbs[u];
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			cumNumNeighbs[u + 1]++;
	}

	// compute the number of commodities (each aggregation router/gateway acts both as a source and a sink)
	unsigned int numComms = T->GetNumAggregationNodes() * T->GetNumGatewayNodes() * 2;


	char var_name[100];
	unsigned int num_vars = cumNumNeighbs[T->GetNumNodes()] * numComms;
	int colno[numComms][num_vars];
	double coeff[num_vars];
	double temp[num_vars];

	// create the LP problem
	// Number of variables: |E|*|V_A|*|V_G|*2
// 	if ((lp = make_lp(0, num_vars)) == NULL)
// 		ERROR();
	model.resize(0, num_vars);

	/* Columns (variables) are ordered this way:
	   f_e1,c1 ... f_e1,cM ... f_eE,c1 ... f_eE,cM
	   where links e1 ... eE are given by
	   neighbors(node 0) neighbors(node 1) ... neighbors(node N)
	   and commodities c1 ... cM are given by
	   a_1,g_1 ... a_1,g_G ... a_A,g_1 ... a_A,g_G ... g_1,a_1 ... g_1,a_A ... g_G,a_1 ... g_G,a_A
	*/

	// give a name to the variables
	int linkId = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
		{
			for (unsigned int i=1; i <= numComms; i++)
			{
// 				sprintf(var_name,"f%d(%d->%d)",i,u->info->id,v->info->id);
// 				if (!set_col_name(lp, linkId*numComms + i, var_name))
// 					ERROR();
				stringstream s;
				s << "f" << i << "(" << u << "->" << v->Id() << ")";
				string str = s.str();
				model.setColumnName(linkId*numComms + i - 1, str);
			}
			linkId++;
		}

// 	set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

	CoinBuild buildObject;

	// add |E|/2 capacity constraints: \sum_{i=1}^{k} f_i(u->v) / c(u->v) + \sum_{i=1}^{k} f_i(v->u) / c(v->u) <= 1 for each u<->v
	unsigned int dist;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			if (u < v->Id())          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(T->Node(u)->NeighborBegin(), v);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[i-1] = 1. / v->rate;
					coeff[i] = 1. / v->Rate();
// 					colno[0][i-1] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					colno[0][i] = (cumNumNeighbs[u] + dist)*numComms + i;
				}
				/* link v->u */
				TLinkItrs invLink (T->FindInverseLink(T->LinkItrs(T->Node(u),v)));
				dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[numComms+i-1] = 1. / invLink.second->rate;
					coeff[numComms+i] = 1. / invLink.second->Rate();
// 					colno[0][numComms+i-1] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					colno[0][numComms+i] = (cumNumNeighbs[v->Id()] + dist)*numComms + i;
				}
				/* add constraint */
// 				if (!add_constraintex(lp, numComms*2, coeff, colno[0], LE, 1))
// 					ERROR();
				buildObject.addRow(numComms*2, colno[0], coeff, -COIN_DBL_MAX, 1.);
			}

	// add (|V|-|V_A|-|V_G|)*numComms flow conservation constraints
	unsigned int j, k, pos;
	for (int u=0; u<T->GetNumNodes(); u++)
		if ( !T->IsAggregationNode(u) && !T->IsGatewayNode(u) )
			{
				j = 0;
				/* outgoing links */
// 				pos = cumNumNeighbs[u->info->id]*numComms + 1;
				pos = cumNumNeighbs[u]*numComms;
				for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
				{
					for (unsigned int i=0; i < numComms; i++)
						colno[i][j] = pos++;
					coeff[j++] = 1;
				}
				/* incoming links */
				for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
				{
					TLinkItrs invLink (T->FindInverseLink(T->LinkItrs(T->Node(u),v)));
					dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 					for (unsigned int i=1; i <= numComms; i++)
					for (unsigned int i=0; i < numComms; i++)
// 						colno[i-1][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
						colno[i][j] = (cumNumNeighbs[v->Id()] + dist)*numComms + i;
					coeff[j++] = -1;
				}
				/* add constraints */
				for (unsigned int i=0; i < numComms; i++)
				{
					// Need to copy because add_constraintex modifies the vector of coefficients
					copy(coeff, coeff+j, temp);
// 					if (!add_constraintex(lp, j, temp, colno[i], EQ, 0))
// 						ERROR();
					buildObject.addRow(j, colno[i], temp, 0., 0.);
				}
			}

	// add numComms constraints to ensure that the flow cannot return back to the source: \sum_{v\in Neighbs(s_i)} f_i(v->s_i) = 0 for i=1,..numComms
	TGraphNodeIterator n;
	/* aggregation nodes */
	for (int a = 0; a < T->GetNumAggregationNodes(); a++)
	{
		j = 0;
		int a_id = T->GetAggregationNode(a);
		for (TGraphNeighborIterator v=T->Node(a_id)->NeighborBegin(); v!=T->Node(a_id)->NeighborEnd(); v++)
		{
			n = T->Node(a_id);
			TLinkItrs invLink (T->FindInverseLink( T->LinkItrs(n,v) ));
			dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[v->Id()] + dist)*numComms + a;
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}
	/* gateways */
	for (int g = 0; g < T->GetNumGatewayNodes(); g++)
	{
		j = 0;
		int g_id = T->GetGatewayNode(g);
		for (TGraphNeighborIterator v=T->Node(g_id)->NeighborBegin(); v!=T->Node(g_id)->NeighborEnd(); v++)
		{
			n = T->Node(g_id);
			TLinkItrs invLink (T->FindInverseLink( T->LinkItrs(n,v) ));
			dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[v->Id()] + dist)*numComms + g + T->GetNumAggregationNodes();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}

	// add numComms constraints to ensure that each commodity has a strictly positive flow: \sum_{v\in Neighbs(s_i)} f_i(s_i->v) > 0 for i=1,..numComms
	/* aggregation nodes */
	for (int a = 0; a < T->GetNumAggregationNodes(); a++)
	{
		j = 0;
		int a_id = T->GetAggregationNode(a);
		for (TGraphNeighborIterator v=T->Node(a_id)->NeighborBegin(); v!=T->Node(a_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(a_id)->NeighborBegin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[a_id] + dist)*numComms + a;
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}
	/* gateways */
	for (int g = 0; g < T->GetNumGatewayNodes(); g++)
	{
		j = 0;
		int g_id = T->GetGatewayNode(g);
		for (TGraphNeighborIterator v=T->Node(g_id)->NeighborBegin(); v!=T->Node(g_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(g_id)->NeighborBegin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[g_id] + dist)*numComms + g + T->GetNumAggregationNodes();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}

// 	set_add_rowmode(lp, FALSE); /* rowmode should be turned off again when done building the model */

	model.addRows(buildObject);

	// set objective function: max \sum_{i=1}^{k} \sum_{v\in Neighbs(s_i)} f_i(s_i->v)
	j = 0;
	unsigned int al_size = T->GetNumAggregationNodes();
	unsigned int gl_size = T->GetNumGatewayNodes();
	for (int a = 0; a < T->GetNumAggregationNodes(); a++)
	{
		int a_id = T->GetAggregationNode(a);
		for (TGraphNeighborIterator v=T->Node(a_id)->NeighborBegin(); v!=T->Node(a_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(a_id)->NeighborBegin(), v);
			/* there is a commodity between aggregation node (a_id)->id and every gateway */
// 			for (unsigned int i=1; i<=gl_size; i++)
			for (unsigned int i=0; i<gl_size; i++)
			{
// 				colno[0][j*gl_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(),a)*gl_size + i;
				colno[0][j*gl_size+i] = (cumNumNeighbs[a_id] + dist)*numComms + a*gl_size + i;
// 				coeff[j*gl_size+i-1] = 1;
				coeff[j*gl_size+i] = 1;
			}
			j++;
		}
	}
	k = 0;
	for (int g = 0; g < T->GetNumGatewayNodes(); g++)
	{
		int g_id = T->GetGatewayNode(g);
		for (TGraphNeighborIterator v=T->Node(g_id)->NeighborBegin(); v!=T->Node(g_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(g_id)->NeighborBegin(), v);
			/* there is a commodity between gateway (*a)->id and every aggregation node */
// 			for (unsigned int i=1; i<=al_size; i++)
			for (unsigned int i=0; i<al_size; i++)
			{
// 				colno[0][j*gl_size+k*al_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + al_size*gl_size + distance(T->gateway_list.begin(),a)*al_size + i;
				colno[0][j*gl_size+k*al_size+i] = (cumNumNeighbs[g_id] + dist)*numComms + al_size*gl_size + g*al_size + i;
// 				coeff[j*gl_size+k*al_size+i-1] = 1;
				coeff[j*gl_size+k*al_size+i] = 1;
			}
			k++;
		}
	}

// 	set_obj_fnex(lp,j*gl_size+k*al_size,coeff,colno[0]);
	
	for (int i=0; i<j*gl_size+k*al_size; i++)
	  model.setObjectiveCoefficient(colno[0][i], coeff[i]);

// 	set_maxim(lp);
	model.setOptimizationDirection(-1);
}



/* Given the initial graph and maxF, builds an LP to solve the 
   maximum minimum congestion factor problem
   Returns the value of all the variables
*/
MaxMinCongFactor::MaxMinCongFactor (Topology* _T, double maxF, bool _apply)
  : LP (_T, _apply)
{
	// compute the cumulative number of neighbors for each node. It is such that:
	// - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
	// - the last element of cumuNumNeighbs is the total number of links
	unsigned int cumNumNeighbs[T->GetNumNodes() + 1];

	cumNumNeighbs[0] = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
	{
		cumNumNeighbs[u + 1] = cumNumNeighbs[u];
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			cumNumNeighbs[u + 1]++;
	}

	// compute the number of commodities (each aggregation router/gateway acts both as a source and a sink)
	unsigned int numComms = T->GetNumAggregationNodes() * T->GetNumGatewayNodes() * 2;


	char var_name[100];
	unsigned int num_vars = cumNumNeighbs[T->GetNumNodes()] * numComms * 1.5 + 1;
	int colno[numComms][num_vars];
	double coeff[num_vars];
	double temp[num_vars];
	double * vars = new double[num_vars];

	// create the LP problem
	// Number of variables: |E|*|V_A|*|V_G|*2 + 1
// 	if ((lp = make_lp(0, num_vars)) == NULL)
// 		ERROR();
	
	model.resize(0, num_vars);

	/* Columns (variables) are ordered this way:
	   f_e1,c1 ... f_e1,cM ... f_eE,c1 ... f_eE,cM B_1,c1 ... B_1,cM ... B_(E/2),c1 ... B_(E/2),cM alpha
	   where links e1 ... eE are given by
	   neighbors(node 0) neighbors(node 1) ... neighbors(node N)
	   and commodities c1 ... cM are given by
	   a_1,g_1 ... a_1,g_G ... a_A,g_1 ... a_A,g_G ... g_1,a_1 ... g_1,a_A ... g_G,a_1 ... g_G,a_A
	*/

	// give a name to the variables
	int linkId = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
		{
			for (unsigned int i=1; i <= numComms; i++)
			{
// 				sprintf(var_name,"f_(%d->%d),%d",u->info->id,v->info->id,i);
// 				if (!set_col_name(lp, linkId*numComms + i, var_name))
// 					ERROR();
				stringstream s;
				s << "f_(" << u << "->" << v->Id() << ")," << i;
				string str = s.str();
				model.setColumnName(linkId*numComms + i - 1, str);
			}
			linkId++;
		}
	linkId = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			if (u < v->Id())       /* each undirected link just once */
			{
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					sprintf(var_name,"B_(%d->%d),%d",u->info->id,v->info->id,i);
// 					if (!set_col_name(lp, (cumNumNeighbs[T->node.size()]+linkId)*numComms + i, var_name))
// 						ERROR();
					stringstream s;
					s << "B_(" << u << "->" << v->Id() << ")," << i+1;
					string str = s.str();
					model.setColumnName((cumNumNeighbs[T->GetNumNodes()]+linkId)*numComms + i, str);
#ifdef MILP
// 					if (!set_binary(lp, (cumNumNeighbs[T->node.size()]+linkId)*numComms + i, TRUE))
					model.setInteger((cumNumNeighbs[T->GetNumNodes()]+linkId)*numComms + i);
					model.setColumnBounds((cumNumNeighbs[T->GetNumNodes()]+linkId)*numComms + i, 0., 1.);
#else
// 					if (!set_bounds(lp, (cumNumNeighbs[T->node.size()]+linkId)*numComms + i, 0, 1))
					model.setColumnBounds((cumNumNeighbs[T->GetNumNodes()]+linkId)*numComms + i, 0., 1.);
#endif
// 						ERROR();
				}
				linkId++;
			}
// 	if (!set_col_name(lp, num_vars, "alpha"))
// 		ERROR();
	{
		string str("alpha");
		model.setColumnName(num_vars-1, str);
	}


// 	set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

	CoinBuild buildObject;

	// add |E|/2 capacity constraints: \sum_{i=1}^{k} f_i(u->v) / c(u->v) + \sum_{i=1}^{k} f_i(v->u) / c(v->u) <= 1 for each u<->v
	unsigned int dist;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			if (u < v->Id())          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(T->Node(u)->NeighborBegin(), v);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[i-1] = 1. / v->rate;
					coeff[i] = 1. / v->Rate();
// 					colno[0][i-1] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					colno[0][i] = (cumNumNeighbs[u] + dist)*numComms + i;
				}
				/* link v->u */
				TLinkItrs invLink (T->FindInverseLink(T->LinkItrs(T->Node(u),v)));
				dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
// 					coeff[numComms+i-1] = 1. / invLink.second->rate;
					coeff[numComms+i] = 1. / invLink.second->Rate();
// 					colno[0][numComms+i-1] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					colno[0][numComms+i] = (cumNumNeighbs[v->Id()] + dist)*numComms + i;
				}
				/* add constraint */
// 				if (!add_constraintex(lp, numComms*2, coeff, colno[0], LE, 1))
// 					ERROR();
				buildObject.addRow(numComms*2, colno[0], coeff, -COIN_DBL_MAX, 1.);
			}

	// add |E|*numComms constraints on the minimum congestion factor: \sum_{i=1}^{k} f_i(u->v) >= alpha * c(u->v) for each u->v
	linkId = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
		{
// 			for (unsigned int i=1; i <= numComms; i++)
			for (unsigned int i=0; i < numComms; i++)
			{
// 				coeff[i-1] = 1;
				coeff[i] = 1;
// 				colno[0][i-1] = linkId*numComms + i;
				colno[0][i] = linkId*numComms + i;
			}
			coeff[numComms] = - v->Rate();
// 			colno[0][numComms] = num_vars;
			colno[0][numComms] = num_vars-1;
// 			if (!add_constraintex(lp, numComms+1, coeff, colno[0], GE, 0))
// 				ERROR();
			buildObject.addRow(numComms+1, colno[0], coeff, 0., COIN_DBL_MAX);
			linkId++;
		}

	// add (|V|-|V_A|-|V_G|)*numComms flow conservation constraints
	unsigned int j, k, pos;
	for (int u=0; u<T->GetNumNodes(); u++)
		if (!T->IsAggregationNode(u) && !T->IsGatewayNode(u))
			{
				j = 0;
				/* outgoing links */
// 				pos = cumNumNeighbs[u->info->id]*numComms + 1;
				pos = cumNumNeighbs[u]*numComms;
				for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
				{
					for (unsigned int i=0; i < numComms; i++)
						colno[i][j] = pos++;
					coeff[j++] = 1;
				}
				/* incoming links */
				for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
				{
					TLinkItrs invLink (T->FindInverseLink(T->LinkItrs(T->Node(u),v)));
					dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 					for (unsigned int i=1; i <= numComms; i++)
					for (unsigned int i=0; i < numComms; i++)
// 						colno[i-1][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
						colno[i][j] = (cumNumNeighbs[v->Id()] + dist)*numComms + i;
					coeff[j++] = -1;
				}
				/* add constraints */
				for (unsigned int i=0; i < numComms; i++)
				{
					// Need to copy because add_constraintex modifies the vector of coefficients
					copy(coeff, coeff+j, temp);
// 					if (!add_constraintex(lp, j, temp, colno[i], EQ, 0))
// 						ERROR();
					buildObject.addRow(j, colno[i], temp, 0., 0.);
				}
			}

#ifndef MILP
	// add numComms constraints to ensure that the flow cannot return back to the source: \sum_{v\in Neighbs(s_i)} f_i(v->s_i) = 0 for i=1,..numComms
	TGraphNodeIterator n;
	/* aggregation nodes */
	for (int a = 0; a < T->GetNumAggregationNodes(); a++)
	{
		j = 0;
		int a_id = T->GetAggregationNode(a);
		for (TGraphNeighborIterator v=T->Node(a_id)->NeighborBegin(); v!=T->Node(a_id)->NeighborEnd(); v++)
		{
			n = T->Node(a_id);
			TLinkItrs invLink (T->FindInverseLink( T->LinkItrs(n,v) ));
			dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[v->Id()] + dist)*numComms + a;
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}
	/* gateways */
	for (int g = 0; g < T->GetNumGatewayNodes(); g++)
	{
		j = 0;
		int g_id = T->GetGatewayNode(g);
		for (TGraphNeighborIterator v=T->Node(g_id)->NeighborBegin(); v!=T->Node(g_id)->NeighborEnd(); v++)
		{
			n = T->Node(g_id);
			TLinkItrs invLink (T->FindInverseLink( T->LinkItrs(n,v) ));
			dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 			colno[0][j] = (cumNumNeighbs[v->info->id] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[v->Id()] + dist)*numComms + g + T->GetNumAggregationNodes();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], EQ, 0))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0., 0.);
	}

	// add numComms constraints to ensure that each commodity has a strictly positive flow: \sum_{v\in Neighbs(s_i)} f_i(s_i->v) > 0 for i=1,..numComms
	/* aggregation nodes */
	for (int a = 0; a < T->GetNumAggregationNodes(); a++)
	{
		j = 0;
		int a_id = T->GetAggregationNode(a);
		for (TGraphNeighborIterator v=T->Node(a_id)->NeighborBegin(); v!=T->Node(a_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(a_id)->NeighborBegin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(), a) + 1;
			colno[0][j] = (cumNumNeighbs[a_id] + dist)*numComms + a;
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}
	/* gateways */
	for (int g = 0; g < T->GetNumGatewayNodes(); g++)
	{
		j = 0;
		int g_id = T->GetGatewayNode(g);
		for (TGraphNeighborIterator v=T->Node(g_id)->NeighborBegin(); v!=T->Node(g_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(g_id)->NeighborBegin(), v);
// 			colno[0][j] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->gateway_list.begin(), a) + T->aggregation_list.size() + 1;
			colno[0][j] = (cumNumNeighbs[g_id] + dist)*numComms + g + T->GetNumAggregationNodes();
			coeff[j++] = 1;
		}
// 		if (!add_constraintex(lp, j, coeff, colno[0], GE, 0.00001))
// 			ERROR();
		buildObject.addRow(j, colno[0], coeff, 0.00001, COIN_DBL_MAX);
	}
#endif

#if 0
	// add |E|*numComms special ordered sets (SOS)
	for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
		for (vector<TGraphNeighbor>::iterator v=u->NeighborBegin(); v!=u->NeighborEnd(); v++)
			if (u->info->id < v->info->id)          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(u->NeighborBegin(), v);
				for (unsigned int i=1; i <= numComms; i++)
					colno[i-1][0] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
				/* link v->u */
				invLink = T->FindInverseLink(TLinkItrs(u,v));
				dist = distance(invLink.first->NeighborBegin(), invLink.second);
				for (unsigned int i=1; i <= numComms; i++)
					colno[i-1][1] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
				/* add constraint */
				for (unsigned int i=1; i <= numComms; i++)
					add_SOS(lp, "SOS", 1, 1, 2, colno[i-1], NULL);
			}
#endif

	// add |E|/2*numComms pairs of constraints: f_i(u->v) <= B * c(u->v) and f_i(v->u) <= (1-B) * c(v->u)
	linkId = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			if (u < v->Id())          /* each undirected link just once */
			{
				/* link u->v */
				dist = distance(T->Node(u)->NeighborBegin(), v);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
					coeff[0] = 1;
// 					colno[i-1][0] = (cumNumNeighbs[u->info->id] + dist)*numComms + i;
					colno[i][0] = (cumNumNeighbs[u] + dist)*numComms + i;
					coeff[1] = - v->Rate();
// 					colno[i-1][1] = (cumNumNeighbs[T->node.size()]+linkId)*numComms + i;
					colno[i][1] = (cumNumNeighbs[T->GetNumNodes()]+linkId)*numComms + i;
// 					if (!add_constraintex(lp, 2, coeff, colno[i-1], LE, 0))
// 						ERROR();
					buildObject.addRow(2, colno[i], coeff, -COIN_DBL_MAX, 0.);
				}
				/* link v->u */
				TLinkItrs invLink (T->FindInverseLink(T->LinkItrs(T->Node(u),v)));
				dist = distance(invLink.first->NeighborBegin(), invLink.second);
// 				for (unsigned int i=1; i <= numComms; i++)
				for (unsigned int i=0; i < numComms; i++)
				{
					coeff[0] = 1;
// 					colno[i-1][0] = (cumNumNeighbs[v->info->id] + dist)*numComms + i;
					colno[i][0] = (cumNumNeighbs[v->Id()] + dist)*numComms + i;
					coeff[1] = invLink.second->Rate();
// 					colno[i-1][1] = (cumNumNeighbs[T->node.size()]+linkId)*numComms + i;
					colno[i][1] = (cumNumNeighbs[T->GetNumNodes()]+linkId)*numComms + i;
// 					if (!add_constraintex(lp, 2, coeff, colno[i-1], LE, invLink.second->rate))
// 						ERROR();
					buildObject.addRow(2, colno[i], coeff, -COIN_DBL_MAX, invLink.second->Rate() );
				}
				linkId++;
			}

	// add the constraint on the maximum throughput: \sum_{i=1}^{k} \sum_{v\in Neighbs(s_i)} f_i(s_i->v) = maxF
	j = 0;
	unsigned int al_size = T->GetNumAggregationNodes();
	unsigned int gl_size = T->GetNumGatewayNodes();
	for (int a = 0; a < T->GetNumAggregationNodes(); a++)
	{
		int a_id = T->GetAggregationNode(a);
		for (TGraphNeighborIterator v=T->Node(a_id)->NeighborBegin(); v!=T->Node(a_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(a_id)->NeighborBegin(), v);
			/* there is a commodity between aggregation node (*a)->id and every gateway */
// 			for (unsigned int i=1; i<=gl_size; i++)
			for (unsigned int i=0; i<gl_size; i++)
			{
// 				colno[0][j*gl_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + distance(T->aggregation_list.begin(),a)*gl_size + i;
// 				coeff[j*gl_size+i-1] = 1;
				colno[0][j*gl_size+i] = (cumNumNeighbs[a_id] + dist)*numComms + a*gl_size + i;
				coeff[j*gl_size+i] = 1;
			}
			j++;
		}
	}
	k = 0;
	for (int g = 0; g < T->GetNumGatewayNodes(); g++)
	{
		int g_id = T->GetGatewayNode(g);
		for (TGraphNeighborIterator v=T->Node(g_id)->NeighborBegin(); v!=T->Node(g_id)->NeighborEnd(); v++)
		{
			dist = distance(T->Node(g_id)->NeighborBegin(), v);
			/* there is a commodity between gateway (*a)->id and every aggregation node */
// 			for (unsigned int i=1; i<=al_size; i++)
			for (unsigned int i=0; i<al_size; i++)
			{
// 				colno[0][j*gl_size+k*al_size+i-1] = (cumNumNeighbs[*a] + dist)*numComms + al_size*gl_size + distance(T->gateway_list.begin(),a)*al_size + i;
// 				coeff[j*gl_size+k*al_size+i-1] = 1;
				colno[0][j*gl_size+k*al_size+i] = (cumNumNeighbs[g_id] + dist)*numComms + al_size*gl_size + g*al_size + i;
				coeff[j*gl_size+k*al_size+i] = 1;
			}
			k++;
		}
	}

// 	if (!add_constraintex(lp, j*gl_size+k*al_size, coeff, colno[0], EQ, maxF))
// 		ERROR();
	buildObject.addRow(j*gl_size+k*al_size, colno[0], coeff, maxF, maxF);

// 	set_add_rowmode(lp, FALSE); /* rowmode should be turned off again when done building the model */

	model.addRows(buildObject);

	// set the objective function: maximize alpha
	coeff[0] = 1;
// 	colno[0][0] = num_vars;
	colno[0][0] = num_vars-1;
// 	set_obj_fnex(lp,1,coeff,colno[0]);
	model.setObjectiveCoefficient(colno[0][0], coeff[0]);

// 	set_maxim(lp);
	model.setOptimizationDirection(-1);

}

