#ifndef WIMESH_LP
#define WIMESH_LP

#include "WMtypes.h"
#include <OsiClpSolverInterface.hpp>



/* node id-radio id pair to represent a radio */
typedef pair<int,int> TRadio;


namespace WMLP {
/* counter function objects
 * it is initialized at init and increased by incr each time it is invoked
 * Can be used with generate to produce sequences like 1,2,3,4,5... (default behavior)
 */
template <class T>
class counter {
private:
  T curr;
  T incr;
public:
  counter(T _init = 0, T _incr = 1) : curr(_init), incr(_incr) {}
  T operator()() { return (curr += incr); }
};
}

struct ArcEpsY {
  float epsN;
  vector<float> epsR;
  int yN;
  vector<int> yR;

  ArcEpsY (float eps, int y, int numRadios) : epsN(eps), yN(y)
  {
    epsR.resize(numRadios);
    yR.resize(numRadios);

    fill (epsR.begin(), epsR.end(), eps);
    fill (yR.begin(), yR.end(), y);
  }
};


/* Abstract base class for ILPs */

class LP
{
public:

  // constructor
  LP (Topology* _T, bool _apply = true)
    : T(_T),
      apply(_apply),
      result(0)
      {}

  // returns the result of solve()
  virtual int lp_solve ();

  // destructor
  virtual ~LP () {}

protected:

  // apply the solution of the program
  virtual void updateFlows () = 0;

  /* A list of common operations performed by linear programs */

//   void SetRadio (TGraphNeighborIterator ni, RadioIface* radio) { TGraphNeighborFriend::SetRadio (ni, radio); }
  void SetFlow (TGraphNeighborIterator ni, double f) { TGraphNeighborFriend::SetFlow (ni, f); }
  void SetUtot (TGraphNeighborIterator ni, double utot) { TGraphNeighborFriend::SetUtot (ni, utot); }
//   void SetRate (TGraphNeighborIterator ni, float rate) { TGraphNeighborFriend::SetRate (ni, rate); }
//   void SetTxPower (TGraphNeighborIterator ni, unsigned int txpower) { TGraphNeighborFriend::SetTxPower (ni, txpower); }
//   void SetAvailBW (TGraphNeighborIterator ni, double bw) { TGraphNeighborFriend::SetAvailBW (ni, bw); }
  void ClearFlowDemands (TGraphNeighborIterator ni) { TGraphNeighborFriend::ClearFlowDemands (ni); }
  void SetFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::SetFlowDemand (ni, d, f); }
  void IncreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::IncreaseFlowDemand (ni, d, f); }
  void DecreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { TGraphNeighborFriend::DecreaseFlowDemand (ni, d, f); }

  Topology* T;
  bool apply;        // true means that the changes will be applied to the topology
  ClpSimplex model;
  int result;
  int Ncol;
};




static vector< set< pair< int, int > > > EmptyVector;


/*
* Rearrange e2e flows of given demands to minimize the maximum Utot
* based on a Linear Program
* Sd is an (optional) vector including the set of links, for each demand (in the
* same order as in the demands map),
* that are allowed to carry the flow of the corresponding demand
* If capacity is true, the capacity constraints are included (which may
* make the LP infeasible)
* Changes are applied only if the resulting max Utot < maxUtot
* In that case, maxUtot is updated to the new max Utot
* Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
*/
class MinMaxUtot : public LP
{
public:
  // constructor, which builds the model
  MinMaxUtot (Topology* _T, map< pair< int, int >, double >* _demands, double* _maxUtot = NULL, 
              vector< set< pair< int, int > > > Sd = EmptyVector, 
              bool capacity = false, vector<float>* nodePowerConstr = NULL, bool _apply = true);

private:
  virtual void updateFlows ();

  map< pair< int, int >, double >* demands;
  double* maxUtot;
};



/*
 * Rearrange e2e flows to minimize the sum of the flows on the links L of a radio
 * such that no Utot above u0 increases nor exceeds u0
 * If the minimum is zero, returns true (the radio can be turned off)
 */
class TurnOffRadio : public LP
{
public:
  // constructor, which builds the model
  TurnOffRadio(Topology* _T, map<pair<int,int>,double>* _demands, set<TLinkItrs>* L, double u0, bool _apply = true);

  bool Succeeded () { return success; }

private:
  virtual void updateFlows ();

  map<pair<int,int>,double>* demands;
  bool success;
  unsigned int numSubDemands;
};


/*
 * This version minimizes the total power consumption subject to constraints on node power and maximum total utilization
 */
class TurnOffRadioPower : public LP
{
public:
  // constructor, which builds the model
  TurnOffRadioPower(Topology* _T, map<pair<int,int>,double>* _demands, const TRadio& r0, double u0, double* _maxUtot,
                    double* _Ptot, vector<float> *nodePowerConstr, bool _apply = true);

private:
  virtual void updateFlows ();

  map<pair<int,int>,double>* demands;
  double* maxUtot;
  double* Ptot;
  double lhs;
  float powerStandbyNodes;
};


/*
 * Rearrange e2e flows to minimizes the total power consumption subject
 * to constraints on node power and maximum total utilization.
 * Returns the result of the LP (OPTIMAL, INFEASIBLE, etc.)
 */
class MinPower_eps : public LP
{
public:
  // constructor, which builds the model
  MinPower_eps(Topology* _T, map<pair<int,int>,double>* _demands, double u0, vector<float> *nodePowerConstr,
               const vector<ArcEpsY>& epsy, bool _apply = true);

private:
  virtual void updateFlows ();

  map<pair<int,int>,double>* demands;
};


/*
 * Given the initial graph, builds an LP to solve the 
 * maximum (throughput) multi-commodity flow problem
 * Used by PFR_mmcf
 */
class MaxMultiCommFlow : public LP
{
public:
  // constructor, which builds the model
  MaxMultiCommFlow (Topology* _T, bool _apply = true);

  double getObjValue () { return model.getObjValue(); }

private:
  virtual void updateFlows () {}
};


/*
 * Given the initial graph and maxF, builds an LP to solve the
 * maximum minimum congestion factor problem
 * Returns the value of all the variables
 * Used by PFR_mmcf
 */
class MaxMinCongFactor : public LP
{
public:
  // constructor, which builds the model
  MaxMinCongFactor (Topology* _T, double maxF, bool _apply = true);

  double* getPrimalColumnSolution () { return ( result ? NULL : model.primalColumnSolution()); }

private:
  virtual void updateFlows () {}
};

#endif
