#include "WMtypes.h"

#include <math.h>
#include <iostream>
#include <set>
#include <map>
#include <algorithm>
#include <numeric>
#include <functional>
#include <cassert>
#include <fstream>
#include <sstream>
#include <iterator>

#define CHANNEL(l)              \
l.second->radio->channel

//#define DEBUG
#define VERBOSE


class MVCRAclass;

/** ------------------------------------------------------- **/
/*
 * Abstract base class to assign a cost to the replacement of channel c_to_erase with channel c on node u
 */
class ChannelCostFunc
{
public:
  virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p) = 0;

  void SetNodeChannel (int _u, unsigned int _c_selected);
  
  void SetMVCRA (MVCRAclass* _mvcra) { mvcra = _mvcra; }
  
protected:
  const vector<vector<int> >& getChangesPerChannel ();
  const set<TLinkItrs>& getQ ();
  // check if a link is repairable
  bool IsRepairable (const TLinkItrs& l);

  int u;
  unsigned c_selected;
  
private:
  MVCRAclass* mvcra;
};

// The cost is the number of times the channel to be replaced has been set on u (as per changesPerChannel)
// normalized to the number of channel changes on u
class ChannelSetCount : public virtual ChannelCostFunc
{
public:
  virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
};

// The cost is the number of disrupted links (repairable links are not counted)
class DisruptedLinkCount : public virtual ChannelCostFunc
{
public:
  virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
};

// The cost is the number of extracted links that would be disrupted (repairable links are not counted)
class DisruptExtrLinkCount : public virtual ChannelCostFunc
{
public:
  virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
};

// The cost is the sum of the F/C ratios of the non-repairable links
class DisruptedLinkFC : public virtual ChannelCostFunc
{
public:
  virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
};

class MixedCost : public ChannelSetCount, public DisruptedLinkFC
{
public:
  virtual double GetCost (const pair<unsigned int,set<TLinkItrs> >& p);
};


/** ------------------------------------------------------- **/

/*
 * Abstract base class to assign a priority to a link
 */
class LinkPriority
{
public:
  LinkPriority (Topology* _T) : T(_T) {}
  
  virtual double calc (const TLinkItrs& l) = 0;

protected:
  Topology* T;
};

// The priority is the Utot included in the Utot field
class UtotPriority : public LinkPriority
{
public:
  UtotPriority (Topology* _T) : LinkPriority(_T) {}
  
  virtual double calc (const TLinkItrs& l) { return l.second->Utot; }
};

// The priority is the F/C ratio of the link
class FCPriority : public LinkPriority
{
public:
  FCPriority (Topology* _T) : LinkPriority(_T) {}
  
  virtual double calc (const TLinkItrs& l) { return l.second->flow / l.second->rate; }
};

// The priority of a link is its f/c ratio multiplied by the number of links
// with Utot >= u0 having that link in their collision domain
class SumUtotPriority : public LinkPriority
{
public:
  SumUtotPriority(Topology* _T, double _u0) : LinkPriority(_T), u0(_u0) {}
  
  virtual double calc (const TLinkItrs& l);
  
private:
  double u0;
};

/*
 * Binary predicate used to compare the priorities of links
 */
class LinkPriorityPred {
    LinkPriority* lp;
  public:
    // constructor
    LinkPriorityPred (LinkPriority* _lp) : lp(_lp) {}
    // operator()
    bool operator()(const TLinkItrs& p, const TLinkItrs& q) {
      return lp->calc(p) < lp->calc(q);
    }
};


/** ------------------------------------------------------- **/

/*
 * A LoopDetector object records all the decisions taken by MVCRA (a channel is assigned to a link)
 * and detects a loop. CausesLoop returns true when a loop is repeated twice:
 * 
 *     ... a b c d e a b c d e a
 *        |---------|---------|
 *                             ↑
 *                           here LoopFound returns true
 */

class LoopDetector
{
public:
  // constructor
  LoopDetector (unsigned int mi) : potential_loop(false), max_iterations(mi) {}
  
  bool CausesLoop (TLinkItrs l, unsigned int ch) throw (int);  // throws an exception if the size of ListOfChanges exceeds max_iterations
  
private:
  vector<pair<TLinkItrs,unsigned int> > ListOfChanges;
  unsigned int idx_1st_loop, idx_2nd_loop;
  bool potential_loop;
  unsigned int max_iterations;
};


/** ------------------------------------------------------- **/

class MVCRAclass
{
public:
  // constructor
  MVCRAclass (Topology* _T, bool _RateAdapt, unsigned int _MaxNumChanges, ChannelCostFunc* _ccf, LinkPriority* _lp, LinkPriority* _lpp)
  : T (_T),
    RateAdapt (_RateAdapt),
    MaxNumChanges (_MaxNumChanges),
    NumChanges (0),
    origChannels (T->node.size()),
    ld (_MaxNumChanges*15),
    changesPerChannel (T->node.size(), vector<int>(T->numAvailChannels,0)),
    ccf (_ccf),
    lp (_lp),
    lpp (_lpp)
    { ccf->SetMVCRA(this); }

  // run MVCRA and return the number of changes
  int Run ();

private:  
  /*
  * Determines the highest rate for l (if any) that makes Utot(l)-f(l)/c(l) smaller than Umax.
  * If such rate is not found, the highest rate that minimizes Utot(l) is returned.
  * Note that the rate of l is not changed
  */
  float Adjust_Rate (TLinkItrs& l, double Umax);
  
  /*
  * Determines the channel and (in case RateAdapt is true) the rate for link l that minimizes the maximum among the Utot of l and 
  * of the links including l in their collision domain. The selected channel is returned, while the selected rate is assigned to
  * new_rate. Note that this function does not change anything
  */
  unsigned int Min_Utot (TLinkItrs l, const std::set< unsigned int >& S, float& new_rate);

  /*
   * Invokes Min_Utot to select a channel, but checks whether the returned choice causes a loop. If so,
   * invokes Min_Utot again to select a different channel
   */
  unsigned int Min_Utot_safe (TLinkItrs l, const std::set< unsigned int >& S, float& new_rate);
  
  /*
  * Returns the channel that would disrupt the minimum number of links if it were replaced
  * by c. If Ldisrupt is a non-null pointer, it points to such set of links (must be freed after use)
  */
  unsigned int Min_Disrupt (int u, unsigned int c, set<TLinkItrs>** Ldisrupt = NULL);

  /*
  * Assigns channel c to one of the radios of g. The channel to be replaced is selected by
  * calling Min_Disrupt. Some links may be added to the queue Qp of pending links.
  */
  void Change_If (MeshNode* g, unsigned int c);


  Topology* T;
  bool RateAdapt;
  unsigned int MaxNumChanges;
  unsigned int NumChanges;
  set<TLinkItrs> Q, Qp;
  vector<set<unsigned int> > origChannels;
  LoopDetector ld;
  vector<vector<int> > changesPerChannel;      // for each node, stores the number of times each channel is assigned to that node
  ChannelCostFunc* ccf;
  LinkPriority *lp, *lpp;

  friend class ChannelCostFunc;  
};

/** ------------------------------------------------------- **/




/** MVCRA **/

/* Given the initial graph, performs a channel assignment
   and returns the induced graph
*/
int MVCRA(Topology* T, bool RateAdapt, unsigned int MaxNumChanges, string CCFname, string QPrio, string QpPrio)
{
  ChannelCostFunc* ccf;
  LinkPriority *lp, *lpp;
  unsigned int len = string("SumUtotPriority").length();
  
  if (CCFname == "ChannelSetCount")
    ccf = new ChannelSetCount;
  else if (CCFname == "DisruptedLinkCount")
    ccf = new DisruptedLinkCount;
  else if (CCFname == "DisruptedLinkFC")
    ccf = new DisruptedLinkFC;
  else if (CCFname == "DisruptExtrLinkCount")
    ccf = new DisruptExtrLinkCount;
  else  // default
    ccf = new MixedCost;
  
  if (QPrio == "UtotPriority")
    lp = new UtotPriority(T);
  else if (QPrio == "FCPriority")
    lp = new FCPriority(T);
  else // default
  {
    double u0 = 0.5; // default value
    
    if (!QPrio.compare(0,len,"SumUtotPriority") && QPrio.length() > len)
    {
      stringstream ss(QPrio.substr(len));
      
      if (!(ss >> u0))
	u0 = 0.5;
    }
      
    lp = new SumUtotPriority(T,u0);
  }
  
  if (QpPrio == "UtotPriority")
    lpp = new UtotPriority(T);
  else if (!QpPrio.compare(0,len,"SumUtotPriority"))
  {
    double u0 = 0.5;

    if (QpPrio.length() > len)
    {
      stringstream ss(QPrio.substr(len));
      
      if (!(ss >> u0))
	u0 = 0.5;
    }
    
    lpp = new SumUtotPriority(T,u0);
  }
  else // default
    lpp = new FCPriority(T);
  
  
  int ret;
  MVCRAclass mvcra (T, RateAdapt, MaxNumChanges, ccf, lp, lpp);
  
  try
  {
    ret = mvcra.Run();
  }
  catch (int)
  {
    cout << "** ERROR ** Maximum number of iterations exceeded. MVCRA quits, likely leaving broken links" << endl;
    ret = -1;
  }

  return ret;
}

int MVCRAclass::Run()
{
  // compute the total utilizations and put all the links into Q
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
    {
      v->Utot = T->computeUtot(TLinkItrs(u,v));
      Q.insert(TLinkItrs(u,v));
    }

  // record the current channels assigned to all the radios (used to detect channel switches)
  for (vector<TGraphNode>::iterator u = T->node.begin(); u != T->node.end(); u++)
    findChannels (u, origChannels[u->info->id])
    

  set<TLinkItrs>::iterator ptr_l;
  TLinkItrs l;
  bool uFull, vFull;
  int uNumChanges, vNumChanges;
  unsigned int c_selected;

  while (!Q.empty() && NumChanges < MaxNumChanges)
  {
    // Extract the link with the highest priority from Q
    ptr_l = max_element (Q.begin(), Q.end(), LinkPriorityPred(lp));
    
    // Erase the extracted link from Q
    l = *ptr_l;
    Q.erase(ptr_l);

#ifdef VERBOSE
    cout << "Extracting link " << l.first->info->id << "->" << l.second->info->id << "(" << l.second->radio->channel << ")  Utot = " << l.second->Utot << endl << endl;
#endif

    // update the Utots before removing the extracted link
    T->updateUtot (l, 0, l.second->rate);
    
    // remove the extracted link
    l.second->radio = NULL;

    float new_rate = l.second->rate;
    
    set<unsigned int> S, uChannels, vChannels;
    S.clear();

    // put all the channels in S
    for (unsigned int i=1; i<=T->numAvailChannels; S.insert(i++));

    // if RateAdapt is true, the rate is updated
    c_selected = Min_Utot_safe (l, S, new_rate);
    
#ifdef VERBOSE
    cout << "Selected channel " << c_selected << endl << endl;
#endif
    
    Change_If (l.first->info, c_selected);
    Change_If (l.second->info, c_selected);

    // update the Utot of the other links
    T->updateUtot (l, c_selected, new_rate);

    // actually change the rate
    l.second->rate = new_rate;
    
    // the selected channel is assigned to the extracted link
    T->AssignChannel2Link(l, c_selected);

    // compute the Utot of link l
    l.second->Utot = T->computeUtot(l);


#ifdef VERBOSE
    cout << "The extracted link " << l.first->info->id << "->" << l.second->info->id << " is assigned channel " << l.second->radio->channel << " new Utot = " << l.second->Utot << endl << endl;
#endif
#ifdef DEBUG
    if (T->checkUtot(1))
      cout << endl << "checkUtot riuscito al punto A" << endl << endl;
    else
      return 0;
#endif
    
    /*
     *** Repair the pending links ***
     */
    while (!Qp.empty())
    {
      // Extract the link with the highest priority from Qp
      ptr_l = max_element (Qp.begin(), Qp.end(), LinkPriorityPred(lpp));
      
      // Erase the extracted link from Qp
      TLinkItrs e = *ptr_l;
      Qp.erase (ptr_l);

#ifdef VERBOSE
    cout << "Extracting pending link " << e.first->info->id << "->" << e.second->info->id << "  Utot = " << e.second->Utot << endl << endl;
#endif

      // Find the set of channels of the end nodes of the extracted link
      findChannels(e.first, uChannels)
      findChannels(e.second, vChannels)

#ifdef VERBOSE
      cout << "Channels on " << e.first->info->id << ": ";
      copy(uChannels.begin(), uChannels.end(), ostream_iterator<unsigned int>(cout, " "));
      cout << endl << "Channels on " << e.second->info->id << ": ";
      copy(vChannels.begin(), vChannels.end(), ostream_iterator<unsigned int>(cout, " "));
      cout << endl;
#endif
      
      // if a radio is unused, it can be assigned any channel
      if (uChannels.size() < e.first->info->numAvailRadios)
	for (unsigned int i=1; i<=T->numAvailChannels; uChannels.insert(i++));
      if (vChannels.size() < e.second->info->numAvailRadios)
	for (unsigned int i=1; i<=T->numAvailChannels; vChannels.insert(i++));
      
      // put the common channels of u and v in S
      S.clear();
      set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

      if (S.empty())
      {
	uNumChanges = accumulate (changesPerChannel[e.first->info->id].begin(), changesPerChannel[e.first->info->id].end(), 0);
	vNumChanges = accumulate (changesPerChannel[e.second->info->id].begin(), changesPerChannel[e.second->info->id].end(), 0);
	
	uFull = uNumChanges > (int)e.first->info->numAvailRadios;
	vFull = vNumChanges > (int)e.second->info->numAvailRadios;

	if (!uFull && !vFull)
	  set_union(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));
	else if (uFull && !vFull)
	  S = uChannels;
	else if (!uFull && vFull)
	  S = vChannels;
	else
	{
	  // do not touch the node which experienced the higher number of radio channel changes
	  S = (uNumChanges > vNumChanges ? uChannels : vChannels);

	  cout << "TUTTI E DUE PIENI!!!" << endl << endl;
	}
      }
      else
      {
	// NOTE this is OPTIONAL: if the end nodes share a common channel, the link is re-inserted into Q
	Q.insert(e);
      }
      
#ifdef VERBOSE
      cout << "Candidate channels: ";
      copy(S.begin(), S.end(), ostream_iterator<unsigned int>(cout, " "));
      cout << endl;
#endif
     
    // u and v have some channels in common
    c_selected = Min_Utot_safe (e, S, new_rate);    


#ifdef VERBOSE
	cout << "Selecting channel " << c_selected << endl << endl;
#endif

      // these calls immediately return if the node already has the selected channel
      Change_If (e.first->info, c_selected);
      Change_If (e.second->info, c_selected);

      T->updateUtot (e, c_selected, new_rate);
      
      e.second->rate = new_rate;
      
      // hence the selected channels already belongs to u and v
      T->AssignChannel2Link(e, c_selected);

      e.second->Utot = T->computeUtot(e);


#ifdef VERBOSE
    cout << "The pending link " << e.first->info->id << "->" << e.second->info->id << " is repaired using channel " << e.second->radio->channel << " new Utot = " << e.second->Utot << endl << endl;
#endif
#ifdef DEBUG
    if (T->checkUtot(1))
      cout << endl << "checkUtot riuscito al punto B" << endl << endl;
    else
      return 0;
#endif
    
    }

  }

  return NumChanges;

}


/** Change_If **/

void MVCRAclass::Change_If (MeshNode* g, unsigned int c)
{
  set<TLinkItrs> *Ldisrupt;
  set<unsigned int> uChannels, vChannels, S;
  
  // if the node already has channel c or an available radio, return
  for (unsigned int i=0; i < g->numAvailRadios; i++)
    if (g->radio[i].channel == c || g->radio[i].channel == 0)
      return;

  unsigned int c_to_erase = Min_Disrupt (g->id, c, &Ldisrupt);
  
  /* The links in Ldisrupt that still share a common channel after the channel switch can be
   * temporarily repaired without being put in the pending queue nor being extracted from Q.
   * The other links are put into Qp and removed from Q
   */
  for (set<TLinkItrs>::iterator l = Ldisrupt->begin(); l != Ldisrupt->end(); )
  {
    // update the Utots before removing the link
    T->updateUtot (*l, 0, l->second->rate);
    
    // remove the link
    l->second->radio = NULL;
    
    // check whether the end nodes of the link will share a channel after the channel switch
    findChannels(l->first, uChannels)
    findChannels(l->second, vChannels)
    
    // take the channel switch into account
    (l->first->info->id == g->id ? uChannels : vChannels).erase(c_to_erase);
    (l->first->info->id == g->id ? uChannels : vChannels).insert(c);

    // put the common channels of u and v in S
    S.clear();
    set_intersection(uChannels.begin(), uChannels.end(), vChannels.begin(), vChannels.end(), inserter(S, S.begin()));

    // move the link to Qp if the end nodes do not share channels
    if (S.empty())
    {
      Qp.insert (*l);
      Q.erase (*l);
    
#ifdef VERBOSE
      cout << "New pending link " << l->first->info->id << "->" << l->second->info->id << "(" << c_to_erase << ")  Utot = " << l->second->Utot << endl;
#endif
      
      // remove the link from Ldisrupt
      set<TLinkItrs>::iterator temp = l++;
      Ldisrupt->erase(temp);
    }
    else
      l++;
  }
  
  /* The links left in Ldisrupt are repairable. Make the channel switch and repair them */

#ifdef VERBOSE
    cout << "Replacing channel " << c_to_erase << " with channel " << c << " on node " << g->id << endl;
#endif

  // replace c_to_erase with c
  for (unsigned int i=0; i<g->numAvailRadios; i++)
    if (g->radio[i].channel == c_to_erase)
    {
      g->radio[i].channel = c;
      
      if ( origChannels[g->id].find(c_to_erase) != origChannels[g->id].end() && origChannels[g->id].find(c) == origChannels[g->id].end() )
	NumChanges++;
      
      if ( origChannels[g->id].find(c_to_erase) == origChannels[g->id].end() && origChannels[g->id].find(c) != origChannels[g->id].end() )
	NumChanges--;

      break;
    }
    
  changesPerChannel[g->id][c-1]++;

  // temporarily repair the repairable links (that have not been extracted by Q nor put into Qp)
  for (set<TLinkItrs>::iterator l = Ldisrupt->begin(); l != Ldisrupt->end(); l++)
  {
    S = T->haveCommonChannels (l->first->info->id, l->second->info->id);
    
    assert (!S.empty());

    float new_rate;
    unsigned int c_selected = Min_Utot (*l, S, new_rate);

    T->updateUtot (*l, c_selected, new_rate);

    l->second->rate = new_rate;

    // hence the selected channels already belongs to u and v
    T->AssignChannel2Link(*l, c_selected);

    l->second->Utot = T->computeUtot(*l);

#ifdef VERBOSE
    cout << "Pending link " << l->first->info->id << "->" << l->second->info->id << " temporarily repaired using channel " << l->second->radio->channel << " new Utot = " << l->second->Utot << endl;
#endif
  }
  
  delete Ldisrupt;

}


/** Min_Utot **/

unsigned int MVCRAclass::Min_Utot(TLinkItrs l, const std::set< unsigned int >& S, float& new_rate)
{
    map<unsigned int, double> Umax;

    // for each channel c in S
    for (set<unsigned int>::iterator c = S.begin(); c != S.end(); c++)
      Umax[*c] = 0.;

    // for every directed link x->y
    for (vector<TGraphNode>::iterator x = T->node.begin(); x != T->node.end(); x++)
      for (list<TGraphNeighbor>::iterator y = x->neighbor.begin(); y != x->neighbor.end(); y++)
	// whose potential collision domain includes the extracted link and having a channel in S
	if ((TLinkItrs(x,y) != l) && (y->radio) && (T->PotInterfLinks(l, TLinkItrs(x,y))) && (S.find(y->radio->channel)!=S.end()))
	{
	  if (y->Utot > Umax[y->radio->channel])
	    Umax[y->radio->channel] = y->Utot;
	}
    if (l.second->radio && S.find(CHANNEL(l)) != S.end() && Umax[CHANNEL(l)])
      Umax[CHANNEL(l)] -= l.second->flow/l.second->rate;

    map<unsigned int, float> rate;
    float save_rate = l.second->rate;
    double Utot;

    /* we use a fake radio to compute Utot(u->v) for every channel, otherwise the channel of the real radio
    * is changed and all the links using that radio are considered to be on the new channel
    */
    RadioIface* fakeradio = new RadioIface();
    RadioIface* save_radio = l.second->radio;
    l.second->radio = fakeradio;

    // for each channel
    for (set<unsigned int>::iterator c=S.begin(); c!=S.end(); c++)
    {
      CHANNEL(l) = *c;

      if (RateAdapt)
	l.second->rate = Adjust_Rate (l, Umax[*c]);
      
      rate[*c] = l.second->rate;

      Utot = T->computeUtot(l) - l.second->flow/l.second->rate;

      if (Utot > Umax[*c])
	Umax[*c] = Utot;

      // in this way, the different rate selected for channels is taken into account in the comparison of Utots
      Umax[*c] += l.second->flow/l.second->rate;
      
      l.second->rate = save_rate;
    }

#ifdef VERBOSE
    for (map<unsigned int,double>::iterator mm = Umax.begin(); mm != Umax.end(); mm++)
      cout << "Umax [" << mm->first << "] = " << mm->second << endl;
#endif

    unsigned int c_selected = min_element(Umax.begin(), Umax.end(), compareSecond<unsigned int,double>())->first;
    new_rate = rate[c_selected];
    
    // restore the previous radio
    l.second->radio = save_radio;
    
    delete fakeradio;
    
    return c_selected;
}


/** Min_Utot_safe **/

unsigned int MVCRAclass::Min_Utot_safe(TLinkItrs l, const std::set< unsigned int >& S, float& new_rate)
{
  unsigned int ch = Min_Utot (l, S, new_rate);
  bool loopfound;
  
  try
  {
    loopfound = ld.CausesLoop(l,ch);
  }
  catch (int)
  {
    // remove the pending links
    for (set<TLinkItrs>::iterator p = Qp.begin(); p != Qp.end(); p++)
      T->node[p->first->info->id].neighbor.erase (p->second);

    // remove the extracted link
    T->node[l.first->info->id].neighbor.erase (l.second);
    
    // rethrow the exception
    throw;
  }
  
  if (loopfound)
  {
    set<unsigned int> S1(S);
    
    S1.erase(ch);
    
    if (S1.empty())
    {
      for (unsigned int i=1; i<=T->numAvailChannels; S1.insert(i++));
      S1.erase(ch);
    }
    
    ch = Min_Utot (l, S1, new_rate);
  }
  
  return ch;
}


/** Adjust_Rate **/

float MVCRAclass::Adjust_Rate(TLinkItrs& l, double Umax)
{
  int m, mmin;
  double min, save;

  float save_rate = l.second->rate;
  
  /* determine the max rate for the given link */
  int start_rate = T->MaxAllowLinkRate(l.first->info->id, l.second->info->id, l.second->power, T->min_rate, T->max_rate);
  
  l.second->rate = availTXrates[start_rate];

  mmin = (m = start_rate);
  min = T->computeUtot(l);

  while ((m>0) && (min > Umax + l.second->flow/l.second->rate))
  {
    l.second->rate = availTXrates[--m];
    if ( (save = T->computeUtot(l)) < min )
    {
      min = save;
      mmin = m;
    }
  }
  
  // restore the previous rate
  l.second->rate = save_rate;
  
  return availTXrates[mmin];
}


/** Min_Disrupt **/

unsigned int MVCRAclass::Min_Disrupt(int u, unsigned int c, set< TLinkItrs >** Ldisrupt)
{
  map<unsigned int,set<TLinkItrs> > W;
  map<unsigned int, unsigned int> rep_count;
  set<unsigned int> S, newCh;
  unsigned int c_to_erase;

  // initialize the set of disrupted links for each channel on u
  for (unsigned int i=0; i<T->node[u].info->numAvailRadios; i++)
    if (T->node[u].info->radio[i].channel)
    {
      W[T->node[u].info->radio[i].channel].clear();
      rep_count[T->node[u].info->radio[i].channel] = 0;
    }
  
  for (list<TGraphNeighbor>::iterator w=T->node[u].neighbor.begin(); w!= T->node[u].neighbor.end(); w++)
  {
    findChannels(w,S)
    // only count this link if it is not already pending
    if (w->radio)
    {
      W[w->radio->channel].insert(TLinkItrs(T->node.begin()+u, w));
      // if the neighbor has channel c, this link is repairable
      if (S.find(c) != S.end())
	rep_count[w->radio->channel]++;
    }
    // count the inverse link(s)
    for (list<TGraphNeighbor>::iterator v=T->node[w->info->id].neighbor.begin(); v!= T->node[w->info->id].neighbor.end(); v++)
      if (v->info->id == u && v->radio)
      {
	W[v->radio->channel].insert(TLinkItrs(T->node.begin()+w->info->id, v));
	// if the neighbor has channel c, this link is repairable
	if (S.find(c) != S.end())
	  rep_count[v->radio->channel]++;
      }
  }

  ccf->SetNodeChannel(u,c);

  map<unsigned int,double> costmap;
  
  // transform (W.begin(), W.end(), costvec.begin(), bind1st(mem_fun(&ChannelCostFunc::GetCost), ccf));
  
  for (map<unsigned int,set<TLinkItrs> >::iterator mm = W.begin(); mm != W.end(); mm++)
    costmap[mm->first] = ccf->GetCost(*mm);

  c_to_erase = min_element(costmap.begin(), costmap.end(), compareSecond<unsigned int,double>())->first;

  
#if 0
  // ignore the channel that has been assigned the highest number of times to this node
  int minT = numeric_limits< int >::max(), maxT = -1;
  unsigned int minCh = 0, maxCh = 0;
  
  for (map<unsigned int,set<TLinkItrs> >::iterator mm = W.begin(); mm != W.end(); mm++)
  {
    if (changesPerChannel[u][mm->first-1] < minT)
    {
      minT = changesPerChannel[u][mm->first-1];
      minCh = mm->first;
    }
    if (changesPerChannel[u][mm->first-1] > maxT)
    {
      maxT = changesPerChannel[u][mm->first-1];
      maxCh = mm->first;
    }
  }
  
  if (minT != maxT)
    W.erase (maxCh);

  assert (!W.empty());
  
  S.clear();

  unsigned int card, min_card = W.begin()->second.size();

  // put into S all the channels corresponding to a set with the minimum cardinality
  for (map<unsigned int,set<TLinkItrs> >::iterator mm = W.begin(); mm != W.end(); mm++)
    if (mm->second.size() - rep_count[mm->first] < min_card)
    {
      min_card = mm->second.size() - rep_count[mm->first];
      S.clear();
      S.insert(mm->first);
    }
    else if (mm->second.size() - rep_count[mm->first] == min_card)
    {
      S.insert(mm->first);
    }

  if (S.size() == 1)
    c_to_erase = *S.begin();
  else
  {
    min_card = numeric_limits<unsigned int>::max();
    
    // need to break the tie
    for (set<unsigned int>::iterator cc = S.begin(); cc != S.end(); cc++)
    {
      card = 0;
      for (set<TLinkItrs>::iterator ll = W[*cc].begin(); ll != W[*cc].end(); ll++)
	if (Q.find(*ll) == Q.end())
	  card++;
	  
      if (card < min_card)
      {
	min_card = card;
	c_to_erase = *cc;
      }
    }
  }
#endif  
  if (Ldisrupt)
    *Ldisrupt = new set<TLinkItrs> (W[c_to_erase]);
  
  return c_to_erase;
}





/** LoopDetector method **/

bool LoopDetector::CausesLoop(TLinkItrs l, unsigned int ch) throw (int)
{
  pair<TLinkItrs,unsigned int> lastChange(l,ch);
  vector<pair<TLinkItrs,unsigned int> >::iterator pos;
  
  // if we are not tracking a potential loop, check if one is starting
  if (!potential_loop)
  {
    if ((pos = find (ListOfChanges.begin(), ListOfChanges.end(), lastChange)) != ListOfChanges.end() )
    {
      // this might be a potential loop...
      potential_loop = true;
      idx_1st_loop = pos - ListOfChanges.begin();
      idx_2nd_loop = ListOfChanges.size();
      
    }
  }
  // otherwise, check if the potential loop is confirmed or not
  else
  {
    if ( lastChange != ListOfChanges[idx_1st_loop + (ListOfChanges.size() - idx_2nd_loop)] )
    {
      // false alarm
      potential_loop = false;
    }
    else if ( lastChange == ListOfChanges[idx_2nd_loop] )
    {
      // the detected loop is starting for the third time
      // this change is not recorded and the caller should recompute a different change
      potential_loop = false;
      
      cout << "** ATTENTION **: Assigning channel " << ch << " to link " << l.first->info->id << "->" << l.second->info->id << " causes a loop! " << endl;
      
      return true;
    }
  }
  
  ListOfChanges.push_back(lastChange);
  
  if (ListOfChanges.size() > max_iterations)
    throw 1;
  
  return false;
}


bool ChannelCostFunc::IsRepairable(const TLinkItrs& l)
{
  set<unsigned int> S;
  
  if (l.first->info->id == u)
    findChannels (l.second, S)
  else
    findChannels (l.first, S)
  
  return (S.find(c_selected) != S.end());
}


void ChannelCostFunc::SetNodeChannel(int _u, unsigned int _c_selected)
{
  u = _u;
  c_selected = _c_selected;
}


const vector<vector<int> >& ChannelCostFunc::getChangesPerChannel ()
{
  return mvcra->changesPerChannel;
}


const std::set< TLinkItrs >& ChannelCostFunc::getQ()
{
  return mvcra->Q;
}


double ChannelSetCount::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  const vector<int>& v = getChangesPerChannel()[u];
  
  double ratio = accumulate (v.begin(), v.end(), 0);
  
  if (ratio)
    ratio = v[p.first-1] / ratio;
  
  return ratio;
}


double DisruptedLinkCount::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  int count = 0;
  
  for (set<TLinkItrs>::const_iterator l = p.second.begin(); l != p.second.end(); l++)
    if (!IsRepairable(*l))
      count++;

  return count;
}


double DisruptExtrLinkCount::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  int count = 0;
  const set<TLinkItrs>& Q = getQ();
  
  for (set<TLinkItrs>::const_iterator l = p.second.begin(); l != p.second.end(); l++)
    if (!IsRepairable(*l) && Q.find(*l) == Q.end())
      count++;

  return count;
}


double DisruptedLinkFC::GetCost(const pair<unsigned int,set<TLinkItrs> >& p)
{
  double total = 0.;
  
  for (set<TLinkItrs>::const_iterator l = p.second.begin(); l != p.second.end(); l++)
    if (!IsRepairable(*l))
      total += l->second->flow / l->second->rate;

  return total;
}


double MixedCost::GetCost (const pair<unsigned int,set<TLinkItrs> >& p)
{
  return (1 + ChannelSetCount::GetCost(p)) * DisruptedLinkFC::GetCost(p);
}


double SumUtotPriority::calc(const TLinkItrs& l)
{
  double ret = 0;

  for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
    for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++) 
      if (y->Utot >= u0 && y->radio && l.second->radio && l.IsInTheCollDomainOf(T, TLinkItrs(x,y))) 
	ret += l.second->flow / l.second->rate * y->Utot;

  return ret;

}


