#include "WMtypes.h"
#include "CAalgos.h"
#include "WMutils.h"

#include <queue>
#include <map>
#include <algorithm>
#include <cassert>

#define MAX_REROUTE_COUNT 3

namespace PACA_NS
{
  /* Binary predicate that returns true if the Utot of x is less than the Utot of y 
    Used by the priority queues in PACA */
  class UtotPriority {
    public:
      bool operator()(const TLinkItrs& x, const TLinkItrs& y) { return x.second->UTot() < y.second->UTot(); }
  };

  /* Binary predicate used to compare the priorities of links
    The priority of a link is its f/c ratio multiplied by the number of links
    with Utot >= u0 having that link in their collision domain */
  class SumUtotPriority {
      Topology* T;
      double u0;
    public:
      SumUtotPriority(Topology* _T, double _u0) : T(_T), u0(_u0) {}
      bool operator()(const TLinkItrs& p, const TLinkItrs& q) {
	int countp=0, countq=0;
	for (int x=0; x<T->GetNumNodes(); x++)
	  for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++) {
	    if ( (y->UTot() >= u0) && p.IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)) )
	      countp++;
	    if ( (y->UTot() >= u0) && q.IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)) )
	      countq++;
	  }
	return countp * p.second->Flow() / p.second->Rate() < countq * q.second->Flow() / q.second->Rate();
      } 
  };

  set<TLinkItrs> findExcessLinks(Topology* T, double u0, bool inclusive);
}  /* PACA_NS namespace */


using namespace PACA_NS;


void PACA::Configure (const list<pair<string,string> >& params)
{
  // default values
  m_u0 = 0.5;

  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("u0", m_u0);

    if (RP.notFound())
      cerr << endl << "Ignoring unknown (or duplicate) attribute (" << p->first << ") for the PACA algorithm" << endl << endl;
  }
}



/* Given a network topology, attempts to switch off as many radios as possible
   while keeping the maximum Utot below u0
   It should be always true that if a link u->v exists then a link v->u exists
   (not necessarily using the same channel)
*/
double PACA::Run(Topology *T)
{
  int num;
  bool pathFound;
  
  /* TODO controllare che se nessun link ha Utot > u0 si passa alla seconda fase */
  
  // compute Utot for each link (PACA may come after a flow variation)
  for (int u=0; u<T->GetNumNodes(); u++)
    for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
      SetUtot (v, T->computeUtot(T->LinkItrs(T->Node(u),v)));

  /* Insert all links in a priority queue, where the priority is the Utot of the link */
  set<TLinkItrs> Q = findExcessLinks(T, m_u0, false);

//   for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
//     for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
//       Q.insert(TLinkItrs(u,v));

  /* extracted link */
  
  while (!Q.empty()) {
//    l = *max_element(Q.begin(), Q.end(), UtotPriority());
    TLinkItrs l (*max_element(Q.begin(), Q.end(), SumUtotPriority(T,m_u0)));
    Q.erase(l);
  
    cout << "(" << l.first->Id() << "->" << l.second->Id() << ")  Utot = " << T->computeUtot(l) << endl;

    num = 0;
    pathFound = true;
    while ((l.second->UTot() > m_u0) && (num < MAX_REROUTE_COUNT) && pathFound) {
      /* TODO forse si può cercare di reinstradare solo la quantità di flusso necessaria per Utot<=u0 */
      cout << "cerco di reinstradare " << l.second->Flow() << " flusso" << endl;
      pathFound = Reroute(T, l, l.second->Flow(), m_u0);
      num++;
    }     
    cout << "--------------------------" << endl;
  }


  /* Q has been emptied, but it is not guaranteed that there is no Utot > u0
     Let's try to switch on some radio on the nodes in the collision domains
     of nodes with Utot >= 0 */
  
  set<TLinkItrs> Qa = findExcessLinks(T, m_u0, false);

  while (!Qa.empty()) {
    TLinkItrs l (*max_element(Qa.begin(), Qa.end(), SumUtotPriority(T,m_u0)));
    Qa.erase(l);

    cout << "** Link " << l.first->Id() << "->" << l.second->Id() << " ha ancora Utot = " << l.second->UTot() << "  **" << endl;
    
    if (!l.first->Info()->countRadiosWithChannel(0))
      continue;

    /* pairs channel, amount of flow */
    map<unsigned int, double> S;
    map<unsigned int, double>::iterator minS;
    
    /* determine potential neighbors (nodes that are in reach of l.first) */
    list<TGraphNeighbor> N;
    int r,p;

    /* find the channels of l.first and avoid to put them in S */
    set<unsigned int> uChannels;
//     findChannels(l.first, uChannels);
    uChannels = l.first->Info()->findChannels();

    for (int u=0; u<T->GetNumNodes(); u++)
      if ( (l.first!=T->Node(u)) and T->canBeNeighbors(l.first->Id(),u,r,p) ) {
	N.push_back(TGraphNeighbor(T->Node(u)->Info(),WMS.availTXrates(r),p));
// 	for (RadioIface* i=u->info->radio; i<u->info->radio+u->info->numAvailRadios(); i++)
	for (int i=0; i<T->Node(u)->Info()->GetNumRadios(); i++)
	  if (T->Node(u)->Info()->Radio(i)->Channel() && (uChannels.find(T->Node(u)->Info()->Radio(i)->Channel()) == uChannels.end()))
	    S[T->Node(u)->Info()->Radio(i)->Channel()] = 0.;
      }

    /* for every channel used by neighbors in N */
    for (map<unsigned int,double>::iterator s=S.begin(); s!=S.end(); s++) {
      cout << "provo ad accendere una radio sul canale " << s->first << endl;
      /* clone the topology */
      Topology* To = new Topology(*T);
      TGraphNodeIterator lfo = To->Node(l.first->Id());    /* refers to l.first in the new topology To */
      TGraphNeighborIterator lso;					/* refers to l.second in the new topology To */
      for (lso=lfo->NeighborBegin(); lso!=lfo->NeighborEnd() && (lso->Id()!=l.second->Id() || lso->RadioChannel()!=l.second->RadioChannel()); lso++);
      assert ( lso != lfo->NeighborEnd() );

      /* insert the potential neighbors using channel c in To */
      for (TGraphNeighborIterator n=N.begin(); n!=N.end(); n++)
	if (n->Info()->countRadiosWithChannel(s->first)) {
	  TGraphNeighborIterator nn = AddNeighbor (lfo, &(*n));
	  To->AssignChannel2Link(To->LinkItrs(lfo,nn),s->first);
	  SetUtot (nn, To->computeUtot(To->LinkItrs(lfo,nn)));
	}

      /* Reroute flows from the extracted link in To (hopefully using the newly established links */
      num = 0;
      pathFound = true;
      while ((lso->Flow() > 0) && (num < MAX_REROUTE_COUNT) && pathFound) {
	cout << "cerco di reinstradare " << lso->Flow() << " flusso" << endl;
	pathFound = Reroute(To, To->LinkItrs(lfo,lso), lso->Flow(), m_u0);
	num++;
      }     

      /* each channel is associated with the resulting flow on the extracted link */
      s->second = lso->Flow();

      delete To;      
    }

    /* assign channels (in order of increasing residual lso->flow) to the unused radios on the extracted link */
    while (l.first->Info()->countRadiosWithChannel(0) && !S.empty()) {
      minS = min_element(S.begin(), S.end(), compareSecond<unsigned int,double>());

      /* insert the potential neighbors using the selected channel */
      for (TGraphNeighborIterator n=N.begin(); n!=N.end(); n++)
	if (n->Info()->countRadiosWithChannel(minS->first)) {
	  TGraphNeighborIterator nn = AddNeighbor (l.first, &(*n));
	  T->AssignChannel2Link(T->LinkItrs(l.first,nn),minS->first);
	  SetUtot (nn, T->computeUtot(T->LinkItrs(l.first,nn)));
	  /* if it does not exist yet (using any channel), insert the inverse link */
	  TLinkItrs invlink (T->FindInverseLink(T->LinkItrs(l.first,nn)));
	  if (invlink.second == invlink.first->NeighborEnd()) {
	    cout << "Insert inverse link " << n->Id() << "->" << l.first->Id() << "[" << minS->first << "]" <<  endl;
	    TGraphNeighbor gn (l.first->Info(),n->Rate(),n->TxPower());
	    TGraphNeighborIterator nn2 = AddNeighbor (T->Node(n->Id()), &gn);
	    T->AssignChannel2Link(T->LinkItrs(T->Node(n->Id()), nn2), minS->first);
	    SetUtot (nn2, T->computeUtot(T->LinkItrs(T->Node(n->Id()), nn2)));
	  }
	  else
	    cout << "Inverse link " << n->Id() << "->" << l.first->Id() << "[" << invlink.second->RadioChannel() << "] already exists" <<  endl;
	}

      /* Reroute flows from the extracted link in T (hopefully using the newly established links */
      num = 0;
      pathFound = true;
      while ((l.second->Flow() > 0) && (num < MAX_REROUTE_COUNT) && pathFound) {
	cout << "trying to re-route " << l.second->Flow() << " amount of flow" << endl;
	pathFound = Reroute(T, l, l.second->Flow(), m_u0);
	num++;
      }     

      S.erase(minS);
    }
  }


  /* Ok, we tried our best to keep Utot <= u0 for every link 
     Now, let's try to switch some radio off, while not 
     introducing any other violation to the constrint Utot <= u0 */

  cout << endl << "** START SECOND STAGE **" << endl << endl;

  /* each pair (node id, radio id) is mapped to the associated flow */
  map<pair<int,int>,double> Qr;
  map<pair<int,int>,double>::iterator minQr;
  set<unsigned int> A,B,Ccom;
  set<TLinkItrs> Eprune;
  double ftot;
  int ri;
  
  for (int i=0; i<T->GetNumNodes(); i++)
  {
    for (int j=0; j<T->Node(i)->Info()->numAvailRadios(); j++)
      if (T->Node(i)->Info()->Radio(j)->Channel())
	Qr[pair<int,int>(i,j)] = 0.;
    for (TGraphNeighborIterator v=T->Node(i)->NeighborBegin(); v!=T->Node(i)->NeighborEnd(); v++)
    {
//       ch = SearchChannel(v->info, v->radio->channel);
      ri = v->Info()->findRadioWithChannel (v->RadioChannel());
      /* add the flow of the current link to the two radios at its endpoints */
      Qr[pair<int,int>(i, distance(T->Node(i)->Info()->Radio(0), RadioIfaceIterator(v->Radio())))] += v->Flow();
      if (ri >= 0)
	Qr[pair<int,int>(v->Id(), ri)] += v->Flow();
    }
  }

  while (!Qr.empty()) {
    
    /* extract the pair (node,radio) with the minimum flow */
    minQr = min_element(Qr.begin(), Qr.end(), compareSecond<pair<int,int>,double>());
    cout << "** Extracting node " << minQr->first.first << " radio " << minQr->first.second << "(channel " << T->Node(minQr->first.first)->Info()->Radio(minQr->first.second)->Channel() << ") flow " << minQr->second << " **" << endl;
    
    /* Eprune is the set of links using the extracted radio */
    Eprune.clear();
    for (TGraphNeighborIterator w=T->Node(minQr->first.first)->NeighborBegin(); w!=T->Node(minQr->first.first)->NeighborEnd(); w++) {
//       if (w->radio - T->node[minQr->first.first].info->radio == minQr->first.second)
      if (distance(T->Node(minQr->first.first)->Info()->Radio(0),RadioIfaceIterator(w->Radio())) == minQr->first.second)
	Eprune.insert(T->LinkItrs(T->Node(minQr->first.first),w));
      TLinkItrs invlink = T->FindInverseLink(T->LinkItrs(T->Node(minQr->first.first),w),T->Node(minQr->first.first)->Info()->Radio(minQr->first.second)->Channel());
      if (invlink.second != invlink.first->NeighborEnd())
	Eprune.insert(invlink);
    }

    /* for each neighbor reached using the extracted radio, consider all the
       existing (if any) links on other channels and all the links that can
       be possibly established using other (if any) common channels 
       The goal is to move the flow on the link that uses the extracted radio
       onto these links */
    for (TGraphNeighborIterator v=T->Node(minQr->first.first)->NeighborBegin(); v!=T->Node(minQr->first.first)->NeighborEnd(); v++) {
//       if (v->radio == T->node[minQr->first.first].info->radio + minQr->first.second) {
      if (distance(T->Node(minQr->first.first)->Info()->Radio(minQr->first.second), RadioIfaceIterator(v->Radio())) == 0) {
	/* v is one of the neighbors reached using the extracted radio */
#if 0
	/* *FIRST* check whether there is a "parallel" link using another channel (hence radio) */
	/* In realta' FIRST è inutile, in quanto realizzato da Reroute in THIRD */
	for (list<TGraphNeighbor>::iterator w=T->node[minQr->first.first].neighbor.begin(); w!=T->node[minQr->first.first].neighbor.end(); w++)
	  if ((w->info->id == v->info->id) && (w->radio->channel != v->radio->channel))
	    /* try to move flow from v to w (same neighbor, different channel) */
	    MoveFlowBetweenParallelLinks(T, T->node.begin()+minQr->first.first, v, w, u0);

	/* if we have moved all the flow from v, jump to the next neighbor reached with the extracted radio */
	if (v->flow == 0)
	  continue;
#endif
	/* *SECOND* check whether a new parallel link on another channel can be established */
	Ccom.clear();
// 	findChannels((T->node.begin()+minQr->first.first),A);
// 	findChannels(v,B);
	A = T->Node(minQr->first.first)->Info()->findChannels();
	B = v->Info()->findChannels();
	set_intersection(A.begin(), A.end(), B.begin(), B.end(), inserter(Ccom, Ccom.begin()));
	Ccom.erase(v->RadioChannel());

	for (set<unsigned int>::iterator c=Ccom.begin(); c!=Ccom.end(); c++) {
	  /* if not existent, a parallel link can be established */
	  bool trovato = false;
	  for (TGraphNeighborIterator w=T->Node(minQr->first.first)->NeighborBegin(); w!=T->Node(minQr->first.first)->NeighborEnd(); w++)
	    if ((w->Id() == v->Id()) && (w->RadioChannel() == *c))
	      trovato = true;
	  if (!trovato) {
	    TGraphNeighbor gn (v->Info(), v->Rate(), v->TxPower());
	    TGraphNeighborIterator nn = AddNeighbor (T->Node(minQr->first.first), &gn);
	    /* add the newly established link */
	    TLinkItrs lnew(T->LinkItrs(T->Node(minQr->first.first), nn));
	    T->AssignChannel2Link(lnew,*c);
	    SetUtot (lnew.second, T->computeUtot(lnew));
	    cout << "Added parallel link " << lnew.first->Id() << "->" << lnew.second->Id() << "[" << *c << "]" << endl;
	    /* if it does not exist yet (using any channel), insert the inverse link */
	    TLinkItrs invlink = T->FindInverseLink(lnew);
	    if (invlink.second == invlink.first->NeighborEnd()) {
	      cout << "Insert inverse link " << lnew.second->Id() << "->" << lnew.first->Id() << "[" << *c << "]" <<  endl;
	      TGraphNeighbor gn2 (lnew.first->Info(),lnew.second->Rate(),lnew.second->TxPower());
	      TGraphNeighborIterator nn2 = AddNeighbor (T->Node(lnew.second->Id()), &gn2);
	      T->AssignChannel2Link(T->LinkItrs(T->Node(lnew.second->Id()), nn2), *c);
	      SetUtot (nn2, T->computeUtot(T->LinkItrs(T->Node(lnew.second->Id()), nn2)));
	    }
	  }

	  /* remove the new parallel link if we were not able to move some flow on it */
/*	  if (!MoveFlowBetweenParallelLinks(T, T->node.begin()+minQr->first.first, v, lnew.second, u0)) {
	    cout << "Rimuovo link " << minQr->first.first << "->" << (--T->node[minQr->first.first].neighbor.end())->info->id << endl;
	    T->node[minQr->first.first].neighbor.pop_back();
	  }*/
#if 0
	  else {//if (!(T->FindInverseLink(lnew,*c))) {
	    /* add the inverse link */
	    cout << "Insert inverse link " << v->info->id << "->" << minQr->first.first << endl;
	    T->node[v->info->id].neighbor.push_front(TGraphNeighbor(T->node[minQr->first.first].info, v->rate, v->power));
	    T->AssignChannel2Link(TLinkItrs(T->node.begin()+v->info->id,T->node[v->info->id].neighbor.begin()),*c);
	    T->node[v->info->id].neighbor.begin()->Utot= T->computeUtot(TLinkItrs(T->node.begin()+v->info->id,T->node[v->info->id].neighbor.begin()));
	  }

	  /* if we have moved all the flow from v, jump to the next neighbor reached with the extracted radio */
	  if (v->flow == 0)
	    break;
#endif
	}
	  
// 	/* if we have moved all the flow from v, jump to the next neighbor reached with the extracted radio */
// 	if (v->flow == 0)
// 	  continue;

	/* *THIRD* reroute flow on links not using the extracted radio */
	num = 0;
	pathFound = true;
	while ((v->Flow() > 0) && (num < MAX_REROUTE_COUNT) && pathFound) {
	  cout << "trying to re-route " << v->Flow() << " amount of flow" << endl;
	  pathFound = Reroute(T, T->LinkItrs(T->Node(minQr->first.first),v), v->Flow(), m_u0, &Eprune);
	  num++;
	}     
      }
    }

    /* If there is no flow on the links using the extracted radio and their inverse links, switch off the radio */
    /* TODO maybe Eprune can be used to avoid re-scanning all the links */
    ftot = 0.;
    for (TGraphNeighborIterator v=T->Node(minQr->first.first)->NeighborBegin(); v!=T->Node(minQr->first.first)->NeighborEnd(); v++) {
//       if (v->radio == T->node[minQr->first.first].info->radio + minQr->first.second)
      if (distance(T->Node(minQr->first.first)->Info()->Radio(minQr->first.second), RadioIfaceIterator(v->Radio())) == 0)
	ftot += v->Flow();
      TLinkItrs invlink (T->FindInverseLink(T->LinkItrs(T->Node(minQr->first.first),v),v->RadioChannel()));
      if (invlink.second != invlink.first->NeighborEnd())
	ftot += invlink.second->Flow();
      }

    if (!ftot) {
      cout << "** EUREKA **" << endl;
      /* remove all the links */
      /* TODO maybe Eprune can be used to avoid re-scanning all the links */
	for (TGraphNeighborIterator v=T->Node(minQr->first.first)->NeighborBegin(); v!=T->Node(minQr->first.first)->NeighborEnd(); )
// 	  if (v->radio == T->node[minQr->first.first].info->radio + minQr->first.second) {
	  if (distance(T->Node(minQr->first.first)->Info()->Radio(minQr->first.second),RadioIfaceIterator(v->Radio())) == 0) {
	    /* first remove the inverse link */
	    TLinkItrs invlink (T->FindInverseLink(T->LinkItrs(T->Node(minQr->first.first),v),v->RadioChannel()));
	    if (invlink.second != invlink.first->NeighborEnd()) {
	      cout << "Removing " << invlink.first->Id() << "->" << invlink.second->Id() << "[" << invlink.second->RadioChannel() << "] flow = " << invlink.second->Flow() << endl;
	      EraseNeighbor (T->Node(v->Id()), invlink.second);
	    }
	    TGraphNeighborIterator vv=v;
	    vv++;
	    /* then the direct link */
	    cout << "Removing " << minQr->first.first << "->" << v->Id() << "[" << v->RadioChannel() << "] flow = " << v->Flow() << endl;
	    EraseNeighbor (T->Node(minQr->first.first), v);
	    v = vv;
	  }
	  else
	    v++;

      /* switch off the radio (set channel to 0) */
      SetChannel (T->Node(minQr->first.first)->Info()->Radio(minQr->first.second), 0);
    }

    Qr.erase(minQr);
  }

  /* erase links with null flow */
  T->eraseLinksWithNullFlow();

  return 0;
}


namespace PACA_NS
{
  /* Link cost function object used by Reroute: c(l) = Utot(l) */
  class linkCost_Utot : public linkCostFunc {
    public:
      float operator()(TLinkItrs l) { return l.second->UTot(); }
  };

  /* Link pruning function object used by Reroute: l in s */
  class linkPrune_Utot : public linkPruneFunc {
    set<TLinkItrs>* s;
    public:
      linkPrune_Utot(set<TLinkItrs>* _s) : s(_s) {}
      float operator()(TLinkItrs l) { return (s->find(l) != s->end()); }
  };

  /* Returns the set of all the links that either have Utot >0 or belong to 
    the collision domain of links with Utot > u0, i.e.,
    the union of the collision domains of all links with Utot > 0 
    If inclusive is true, the inequality is >= (not only >) u0 */
  set<TLinkItrs> findExcessLinks(Topology* T, double u0, bool inclusive) {
    set<TLinkItrs> s1, s2;
    
    /* s1 is the set of all the links whose Utot >= u0 */
    for (int x=0; x<T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	if ( (y->UTot() > u0) || (y->UTot() == u0 && inclusive) )
	  s1.insert(T->LinkItrs(T->Node(x),y));

    /* s2 is the set of all the links with Utot >= u0 or belonging to the collision domain of links with Utot >= u0 */  
    for (int x=0; x<T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++) {
	  //cout << "FE " << x->info->id << "->" << y->info->id << endl;
	for (set<TLinkItrs>::iterator l=s1.begin(); l!=s1.end(); l++)
	  if ( T->LinkItrs(T->Node(x),y).IsInTheCollDomainOf(*l) ) {
	    s2.insert(T->LinkItrs(T->Node(x),y));
	    continue;
	  }
	}
    return s2;
  }

} /* namespace PACA_NS */


  /* Attempts to reroute up to f_uv flow from link u->v subject
    to the constraint that the Utot should be less than u0
    Returns false if there is no path to reroute flow
    otherLinksToPrune is a set of additional links to prune */
  bool PACA::Reroute(Topology* T, TLinkItrs uv, double f_uv, double u0, set<TLinkItrs>* otherLinksToPrune) {

    vector<TLinkItrs>* path;
    set<TLinkItrs> s = findExcessLinks(T, u0, true);
    s.insert(uv);  

    cout << "** REROUTE ** " << uv.first->Id() << "->" << uv.second->Id() << "[" << uv.second->RadioChannel() << "]" << endl;

    if (otherLinksToPrune)
      set_union(s.begin(), s.end(), otherLinksToPrune->begin(), otherLinksToPrune->end(), inserter(s, s.begin()));
    
    /* Find the shortest path between u and v that does not include u->v and the links that
      either have Utot >=0 or belong to the collision domain of links with Utot >= u0 
      Indeed, we cannot allocate flow on such links */

    linkPruneFunc* prune = new linkPrune_Utot (&s);
    linkCostFunc* cost = new linkCost_Utot;

    path = T->ShortestPath(uv.first->Id(), uv.second->Id(), prune, cost);

    if (path->empty())
      return false;

    double d, f;
    multimap<double,TLinkItrs> count;
    multimap<double,TLinkItrs> dneg;

    // for every directed link x->y
    for (int x=0; x < T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++) {

	d = 0.;
	// for every link l along the shortest path
	for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++)
	  // if l belongs to the collision domain of x->y
	  if ( l->IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)) ) 
	    d += 1 / l->second->Rate();
	// if u->v belongs to the collision domain of x->y, subtract its contribution because flow will 
	// be moved out of u->v
	if (uv.IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)))
	  d -= 1 / uv.second->Rate();

	// if in the collision domain of x->y there is no link of the shortest path neither u->v (d=0) or
	// there is u->v but no link of the shortest path (d<0), then x->y should not be considered
	if (d>0.)
	  count.insert(pair<double,TLinkItrs>((u0 - y->UTot())/d, T->LinkItrs(T->Node(x),y)));
	if (d<0.)
	  dneg.insert(pair<double,TLinkItrs>(d, T->LinkItrs(T->Node(x),y)));
      }

    f = min(count.begin()->first, f_uv);

  /*  cout << "count vector:" << endl;
    for (multimap<double,TLinkItrs>::iterator p=count.begin(); p!= count.end(); p++)
      cout << p->second.first->info->id << "->" << p->second.second->info->id << " count = " << p->first << endl;*/
    
    cout << "shortest path ";
    for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++)
      cout << l->first->Id() << "->" << l->second->Id() << "[" << l->second->RadioChannel() << "] ";
    cout << endl << "f = " << f << endl;

    /* Since we have pruned links with Utot >= u0 or belonging to the collision domain of
      links with Utot >= u0, the flow should be always positive */
  //  assert (f > 0.);

    /* reroute traffic demands (if any) */
    if (f)
      for (FlowDemandsIterator dem=uv.second->F_e2eBegin(); dem!=uv.second->F_e2eEnd(); dem++) {
	for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++)
	  if (l->second->IsDemandPresent (dem->first))
	    IncreaseFlowDemand (l->second, dem->first, f * dem->second / uv.second->Flow());
	  else
	    SetFlowDemand (l->second, dem->first, f * dem->second / uv.second->Flow());

// 	  dem->second -= (f * dem->second / uv.second->flow);
	  DecreaseFlowDemand (uv.second, dem->first, f * dem->second / uv.second->Flow());
      }

    /* actually add the flow on the links of the shortest path */
    for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++)
      SetFlow (l->second, l->second->Flow() + f);
    /* and remove the rerouted flow f from link u->v */
    SetFlow (uv.second, uv.second->Flow() - f);

    /* update the Utot of all the affected links */
    for (multimap<double,TLinkItrs>::iterator m=count.begin(); m!=count.end(); m++)
      SetUtot (m->second.second, m->second.second->UTot() + (1/m->first)*(u0-m->second.second->UTot())*f);
    for (multimap<double,TLinkItrs>::iterator m=dneg.begin(); m!=dneg.end(); m++)
      SetUtot (m->second.second, m->second.second->UTot() + (m->first)*f);

  cout << "Utot(" << uv.first->Id() << "->" << uv.second->Id() << "[" << uv.second->RadioChannel() << "]) = " << uv.second->UTot() << endl << endl;
  //   /* update the Utot of all the affected links */
  //   for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
  //     for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
  //       // if u->v belongs to the collision domain of x->y
  //       if ( uv.IsInTheCollDomainOf(T, TLinkItrs(x,y)) ) 
  // 	y->Utot -= f/uv.second->rate;

    delete prune;
    delete cost;
    delete path;

    return true;  
  }



  /* DEPRECATED
    Attempts to move flow from u->v to u->w 
    It makes sense when u->v and u->w are two "parallel" links between the same pair of nodes 
    Returns true if it has been able to move some flow */
  bool PACA::MoveFlowBetweenParallelLinks(Topology *T, TGraphNodeIterator u, TGraphNeighborIterator v, TGraphNeighborIterator w, double u0) {
    double f = v->Flow();	/* flow to be moved on parallel links */
    vector<TGraphNeighborIterator> Etemp;

    for (int x=0; x<T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	// if the parallel link belongs to the collision domain of x->y
	if ( T->LinkItrs(u,w).IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)) ) {
	  Etemp.push_back(y);
	  f = min(f, (u0 - y->UTot()) * w->Rate());
	}

    if (f<=0.)
      return false;

    /* actually add the flow on w */
    SetFlow (w, w->Flow() + f);

    /* update the Utot of all the affected links */
    for (vector<TGraphNeighborIterator>::iterator i=Etemp.begin(); i!=Etemp.end(); i++)
      SetUtot (*i, (*i)->UTot() + f/w->Rate());

    /* remove the rerouted flow f from link v */
    SetFlow (v, v->Flow() - f);

    /* update the Utot of all the affected links */
    for (int x=0; x<T->GetNumNodes(); x++)
      for (TGraphNeighborIterator y=T->Node(x)->NeighborBegin(); y!=T->Node(x)->NeighborEnd(); y++)
	// if u->v belongs to the collision domain of x->y
	if ( T->LinkItrs(u,v).IsInTheCollDomainOf(T->LinkItrs(T->Node(x),y)) ) 
	  SetUtot (y, y->UTot() - f/v->Rate());

    return true;
  }
