// -*- C++ -*-

/* myparser.cc
 *
 * Copyright (C) 2002 The libxml++ development team
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "parsers.h"
#include "WMutils.h"
#include "WMtypes.h"
#include "rndvars.h"

#include <glibmm/convert.h> //For Glib::ConvertError
#include <iostream>
#include <cassert>

using namespace std;

/** WiMesh config parser **/

WiMeshParser::WiMeshParser ()
  : xmlpp::SaxParser(),
    nodePCrv (NULL),
    nodePCseed (RndVar::getSeed()),
    radioPCrv (NULL),
    radioPCseed (RndVar::getSeed()),
    nodeUrng (NULL),
    radioUrng (NULL),
    CDrv (NULL),
    CDseed (RndVar::getSeed()),
    CDtype (0),
    CDid ("XX"),
    postCAopt (false)
{
}

WiMeshParser::~WiMeshParser()
{
  if (nodeUrng)
    delete nodeUrng;
  if (radioUrng)
    delete radioUrng;
}

void WiMeshParser::on_start_document()
{
  cout << "Reading WiMesh config..." << endl << endl;
}

void WiMeshParser::on_end_document()
{
  cout << "Finished reading WiMesh config" << endl << endl;
}

void WiMeshParser::on_start_element(const Glib::ustring& name,
                                   const AttributeList& attributes)
{
  static int id;

  if (name == "Topology")
  {
    if (!xml_topo_file.empty() || !serialized_topo_file.empty())
    {
      // a topology file name has been already specified
      cerr << endl << "Ignoring duplicate Topology element" << endl << endl;
      return;
    }
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("xml_file", xml_topo_file);
      RP.read ("serialized_file", serialized_topo_file);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Topology element" << endl << endl;
    }
  }
  else if (name == "Simulation")
  {
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("numChannels", WMS.m_numAvailChannels);
      RP.read ("id", WMS.m_simulId);
      RP.read ("SINRfactor", WMS.m_SINRthreshfactor);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Simulation element" << endl << endl;
    }
  }
  else if (name == "BuildGiMethod")
  {
    if (!buildGiMethod.empty())
    {
      // a buildGi method has been already specified
      cerr << endl << "Ignorind duplicate BuildGiMethod element" << endl << endl;
      return;
    }

    // "name" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring BuildGiMethod element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring BuildGiMethod element with first attribute other than name" << endl << endl;
      return;
    }

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      buildGiParams.push_back(pair<string,string>(iter->name,iter->value));
    
    buildGiMethod = attributes.front().value;
  }
  else if (name == "PFRmethod")
  {
    if (!pfrMethod.empty())
    {
      // a pfr method has been already specified
      cerr << endl << "Ignoring duplicate PFRmethod element" << endl << endl;
      return;
    }

    // "name" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring PFRmethod element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring PFRmethod element with first attribute other than name" << endl << endl;
      return;
    }

    // save all the attributes in the params list
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      pfrParams.push_back(pair<string,string>(iter->name,iter->value));
    
    pfrMethod = attributes.front().value;
  }
  else if (name == "InterferenceModel")
  {
    // "name" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring InterferenceModel element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring InterferenceModel element with first attribute other than name" << endl << endl;
      return;
    }

    if (attributes.front().value == "SINRbased")
    {
      if (WMS.m_interferenceModel)
	delete WMS.m_interferenceModel;

      WMS.m_interferenceModel = new SINRbasedInterferenceModel;
      return;
    }

    if (attributes.front().value == "khopNeighbor")
    {
      int k = 2;

      if (WMS.m_interferenceModel)
	delete WMS.m_interferenceModel;

      // the only expected parameter is k
      if (attributes.size() > 1)
      {
	xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1;

	ReadParams RP (iter->name, iter->value);
	
	RP.read ("k", k);

	if (RP.notFound())
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for the k-hop interfernce model" << endl << endl;
      }

      WMS.m_interferenceModel = new kHopNeighbInterferenceModel (k);
      return;
    }

    cerr << endl << "Ignoring InterferenceModel element with unknown interference model (" << attributes.front().value << ")" << endl << endl;
  }
  else if (name == "PropagationLossModel")
  {
    // "name" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring PropagationLossModel element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring PropagationLossModel element with first attribute other than name" << endl << endl;
      return;
    }

    if (attributes.front().value == "LogDistance")
    {
      double exponent = LogDistancePropagationLossModel::default_exponent,
             referenceDistance = LogDistancePropagationLossModel::default_referenceDistance,
             referenceLoss = LogDistancePropagationLossModel::default_referenceLoss;

      if (WMS.m_propagationLossModel)
	delete WMS.m_propagationLossModel;

      for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      {
	ReadParams RP (iter->name, iter->value);
	
	RP.read ("exponent", exponent);
	RP.read ("ref_distance", referenceDistance);
	RP.read ("ref_loss", referenceLoss);

	if (RP.notFound())
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for the LogDistance propagation loss model" << endl << endl;
      }

      WMS.m_propagationLossModel = new LogDistancePropagationLossModel (exponent, referenceDistance, referenceLoss);
      return;
    }

    if (attributes.front().value == "ThreeLogDistance")
    {
      double distance0 = ThreeLogDistancePropagationLossModel::default_distance0,
             distance1 = ThreeLogDistancePropagationLossModel::default_distance1,
             distance2 = ThreeLogDistancePropagationLossModel::default_distance2,
             exponent0 = ThreeLogDistancePropagationLossModel::default_exponent0,
             exponent1 = ThreeLogDistancePropagationLossModel::default_exponent1,
             exponent2 = ThreeLogDistancePropagationLossModel::default_exponent2,
             referenceLoss = ThreeLogDistancePropagationLossModel::default_referenceLoss;

      if (WMS.m_propagationLossModel)
	delete WMS.m_propagationLossModel;

      for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      {
	ReadParams RP (iter->name, iter->value);
	
	RP.read ("distance0", distance0);
	RP.read ("distance1", distance1);
	RP.read ("distance2", distance2);
	RP.read ("exponent0", exponent0);
	RP.read ("exponent1", exponent1);
	RP.read ("exponent2", exponent2);
	RP.read ("ref_loss", referenceLoss);

	if (RP.notFound())
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for the ThreeLogDistance propagation loss model" << endl << endl;
      }

      WMS.m_propagationLossModel = new ThreeLogDistancePropagationLossModel (distance0, distance1, distance2,
                                                                             exponent0, exponent1, exponent2, referenceLoss);
      return;
    }

    cerr << endl << "Ignoring PropagationLossModel element with unknown propagation loss model (" << attributes.front().value << ")" << endl << endl;
  }
  else if (name == "NodePCprofile")
  {
    float bpc = defaultNodePCprofile.BasePowerConsumption,
          spc = defaultNodePCprofile.StandbyPowerConsumption;

    // "id" is expected as the first attribute, followed by the other parameters

    if (attributes.empty())
    {
      cerr << endl << "Ignoring NodePCprofile element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "id")
    {
      cerr << endl << "Ignoring NodePCprofile element with first attribute other than id" << endl << endl;
      return;
    }

    id = atoi (attributes.front().value.c_str());
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("bpc", bpc);
      RP.read ("spc", spc);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a NodePCprofile element" << endl << endl;
    }
    
    nodePCprofile.insert(pair<int,NodePowerConsumptionProfile>(id,NodePowerConsumptionProfile(bpc,spc)));
  }
  else if (name == "RadioPCprofile")
  {
    foRadioTxPower* rho_tx = defaultRadioPCprofile.rho_tx->Clone();
    float rho_rx = defaultRadioPCprofile.rho_rx,
          rho_idle = defaultRadioPCprofile.rho_idle,
	  rho_sleep = defaultRadioPCprofile.rho_sleep;

    // "id" is expected as the first attribute, followed by the other parameters

    if (attributes.empty())
    {
      cerr << endl << "Ignoring RadioPCprofile element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "id")
    {
      cerr << endl << "Ignoring RadioPCprofile element with first attribute other than id" << endl << endl;
      return;
    }

    id = atoi (attributes.front().value.c_str());

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      RP.read ("rho_rx", rho_rx);
      RP.read ("rho_idle", rho_idle);
      RP.read ("rho_sleep", rho_sleep);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a RadioPCprofile element" << endl << endl;
    }
    
    radioPCprofile.insert(pair<int,RadioPowerConsumptionProfile>(id,RadioPowerConsumptionProfile(rho_tx, rho_rx, rho_idle, rho_sleep)));
  }
  else if (name == "Rho_tx")
  {
    string fname;

    // "name" is expected as the first attribute, followed by the other parameters

    if (attributes.empty())
    {
      cerr << endl << "Ignoring Rho_tx element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "name")
    {
      cerr << endl << "Ignoring Rho_tx element with first attribute other than name" << endl << endl;
      return;
    }

    fname = attributes.front().value;
    
    // check if it is the name of a RadioTxPower function object
    if (fname == foRadioTxPowerLinear::name)
    {
      float a, b;
      bool hasA = false, hasB = false;

      // read foRadioTxPowerLinear parameters
      for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
      {
	if (iter->name == "a")
	{
	  a = atof(iter->value.c_str());
	  hasA = true;
	}
	else if (iter->name == "b")
	{
	  b = atof(iter->value.c_str());
	  hasB = true;
	}
	else
	  cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a Rho_tx::" << fname << " element" << endl << endl;
      }

      if (!hasA || !hasB)
      {
	cerr << endl << "Ignoring Rho_tx::" << fname << " element with incomplete list of parameters (a, b)" << endl << endl;
	return;
      }

      // check whether an entry with the current id is present in radioPCprofile
      map<int, RadioPowerConsumptionProfile>::iterator mi = radioPCprofile.find(id);
      
      if (mi == radioPCprofile.end())
      {
	cerr << endl << "Ignoring Rho_tx::" << fname << " element, being unable to find a corresponding RadioPCprofile element" << endl << endl;
	return;
      }

      if (mi->second.rho_tx)
	delete mi->second.rho_tx;

      mi->second.rho_tx = new foRadioTxPowerLinear (a, b);
    }
    else
    {
      cerr << endl << "Ignoring Rho_tx element with unknown RadioTxPower function object name (" << fname << ")" << endl << endl;
      return;
    }
  }
  else if (name == "NodePCRndVar")
  {
    if (nodePCrv)
    {
      // a nodePC random variable has been already specified
      cerr << endl << "Ignoring duplicate NodePCRndVar element" << endl << endl;
      return;
    }

    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (!nodePCrv)
	RP.read ("rv", nodePCrv);
      RP.read ("seed", nodePCseed);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid (or duplicate) attribute (" << iter->name << ") for a NodePCRndVar element" << endl << endl;
    }

    if (!nodePCrv)
    {
      cerr << endl << "Ignoring NodePCRndVar element with no valid rv attribute" << endl << endl;
      return;
    }
  }
  else if (name == "RadioPCRndVar")
  {
    if (radioPCrv)
    {
      // a radioPC random variable has been already specified
      cerr << endl << "Ignoring duplicate RadioPCRndVar element" << endl << endl;
      return;
    }
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);

      if (!radioPCrv)
	RP.read ("rv", radioPCrv);
      RP.read ("seed", radioPCseed);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid (or duplicate) attribute (" << iter->name << ") for a RadioPCRndVar element" << endl << endl;
    }

    if (!radioPCrv)
    {
      cerr << endl << "Ignoring RadioPCRndVar element with no valid rv attribute" << endl << endl;
      return;
    }
  }
  else if (name == "ChangeDemands")
  {
    if (CDrv)
    {
      // a CD random variable has been already specified
      cerr << endl << "Ignoring duplicate ChangeDemands element" << endl << endl;
      return;
    }
    
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin(); iter != attributes.end(); ++iter)
    {
      ReadParams RP (iter->name, iter->value);
      
      if (!CDrv)
	RP.read ("rv", CDrv);
      RP.read ("seed", CDseed);
      RP.read ("type", CDtype);
      RP.read ("id", CDid);

      if (RP.notFound())
	cerr << endl << "Ignoring invalid attribute (" << iter->name << ") for a ChangeDemands element" << endl << endl;
    }

    if (!CDrv || CDtype > 2)
    {
      cerr << endl << "Ignoring ChangeDemands element with invalid attributes" << endl << endl;
      return;
    }
  }
  else if (name == "ChannelAssign")
  {
    if (!caAlgo.empty())
    {
      // a channel assignment algorithm has been already specified
      cerr << endl << "Ignoring duplicate ChannelAssign element" << endl << endl;
      return;
    }

    // "algo" is expected as the first attribute, followed by parameters where requested

    if (attributes.empty())
    {
      cerr << endl << "Ignoring ChannelAssign element with no attributes" << endl << endl;
      return;
    }

    if (attributes.front().name != "algo")
    {
      cerr << endl << "Ignoring ChannelAssign element with first attribute other than algo" << endl << endl;
      return;
    }

    // save all the attributes (but postopt) in the params list
    for(xmlpp::SaxParser::AttributeList::const_iterator iter = attributes.begin()+1; iter != attributes.end(); ++iter)
    {
      if (iter->name == "postopt")
      {
	if (iter->value == "true" || iter->value == "1")
	  postCAopt = true;
	if (iter->value == "false" || iter->value == "0")
	  postCAopt = false;
      }
      else
	caParams.push_back(pair<string,string>(iter->name,iter->value));
    }

    caAlgo = attributes.front().value;
  }
  else if (name != "WiMeshCfg")
  {
    cerr << endl << "Ignoring invalid element (" << name << ")" << endl << endl;
  }
}

void WiMeshParser::on_end_element(const Glib::ustring& name)
{
//   if (name == "WiMeshCfg")
//     // stop the parser
//     throw xmlpp::exception("Done");
}

void WiMeshParser::on_characters(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_characters(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "WiMeshParser::on_characters(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void WiMeshParser::on_comment(const Glib::ustring& text)
{
//   try
//   {
//     std::cout << "on_comment(): " << text << std::endl;
//   }
//   catch(const Glib::ConvertError& ex)
//   {
//     std::cerr << "WiMeshParser::on_comment(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
//   }
}

void WiMeshParser::on_warning(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_warning(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshParser::on_warning(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshParser::on_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshParser::on_error(): Exception caught while converting text for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshParser::on_fatal_error(const Glib::ustring& text)
{
  try
  {
    std::cout << "on_fatal_error(): " << text << std::endl;
  }
  catch(const Glib::ConvertError& ex)
  {
    std::cerr << "WiMeshParser::on_characters(): Exception caught while converting value for std::cout: " << ex.what() << std::endl;
  }
}

void WiMeshParser::setNodePCprofile (MeshNode* info)
{
  if (!nodePCrv)
    return;

  if (!nodeUrng)
  {
    nodeUrng = new MotherOfAll (nodePCseed);    // declare uniform random number generator
    Random::Set (*nodeUrng);                    // set urng as generator to be used
  }

  int id = (int)nodePCrv->rv->Next();

  // find a profile with such an id
  map<int, NodePowerConsumptionProfile>::const_iterator mi = nodePCprofile.find (id);

  if (mi == nodePCprofile.end())
  {
    cerr << endl << "Warning: node profile id=" << id << " not found. Making no changes" << endl << endl;
    return;
  }

  info->BasePowerConsumption = mi->second.BasePowerConsumption;
  info->StandbyPowerConsumption = mi->second.StandbyPowerConsumption;
}

void WiMeshParser::setRadioPCprofile (RadioIface* radio)
{
  if (!radioPCrv)
    return;

  if (!radioUrng)
  {
    radioUrng = new MotherOfAll (radioPCseed);    // declare uniform random number generator
    Random::Set (*radioUrng);                     // set urng as generator to be used
  }

  int id = (int)radioPCrv->rv->Next();

  // find a profile with such an id
  map<int, RadioPowerConsumptionProfile>::const_iterator mi = radioPCprofile.find (id);

  if (mi == radioPCprofile.end())
  {
    cerr << endl << "Warning: radio profile id=" << id << " not found. Making no changes" << endl << endl;
    return;
  }

  radio->rho_tx = mi->second.rho_tx->Clone();
  radio->rho_rx = mi->second.rho_rx;
  radio->rho_idle = mi->second.rho_idle;
  radio->rho_sleep = mi->second.rho_sleep;
}

