/* This files contains the function to compute the maxflow on the initial graph */

#include "pfr_methods.h"
#include "WMutils.h"
#include "CAalgos.h"
#include "lp.h"
#include "types_pr.h"
#include <limits>
#include <sstream>
#include <iostream>

extern "C" int hipr(char**,node**,arc**,long**,double*,long*);


#define N_NODE( i ) ( (i) - ndp + nmin )
#define N_ARC( a )  ( ( a == NULL )? -1 : (a) - arp )

using namespace std;


PFR_method* PFR_method::Create (string name, const list<pair<string,string> >& params)
{
  if (name == "pfr_Raniwala")
    return new PFR_Raniwala (params);
  if (name == "pfr_MinMaxUtot")
    return new PFR_MinMaxUtot (params);
  if (name == "pfr_Random")
    return new PFR_Random (params);
  if (name == "pfr_AllpairsMF")
    return new PFR_allpairsMF;
  if (name == "pfr_MMCF")
    return new PFR_mmcf;
  if (name == "pfr_Maxflow")
    return new PFR_maxflow;

  return NULL;
}


/* compute maxflow
   the network parameter describes the problem */
void maxflow(Topology *T, char **network)
{
	double  flow;
	/* needed by hipr function */
	node *ndp;
	arc *arp;
	long *cap;
	long ni,na,nmin;
	TGraphNeighborIterator pos;

	hipr(network,&ndp,&arp,&cap,&flow,&nmin);

	for (node* in = ndp; in < ndp + T->GetNumNodes()+2; in ++ )
	{
		ni = N_NODE(in);
		for ( arc* a = in -> first; a != NULL; a = a -> next )
		{
			na = N_ARC(a);
			if ( (cap[na] > 0) && (ni != (int)T->GetNumNodes()) && (N_NODE(a->head) != (int)T->GetNumNodes()+1) )  // exclude fictitious links
			{
				for (pos = T->Node(ni)->NeighborBegin(); pos != T->Node(ni)->NeighborEnd() && pos->Id() != N_NODE(a->head); pos++);
				if (pos != T->Node(ni)->NeighborEnd())
					PFR_method::SetFlow (pos, pos->Flow() + cap[na]-(a->r_cap));
			}
		}
	}
	/* free memory */
	free(ndp);
	free(arp);
	free(cap);
}


/* computes the maxflow between supersource and supersink on the initial graph (given by T)
   and records the flow on each link */

void PFR_maxflow::pfr_calc(Topology* T)
{
	int     numarcs=0;

	/* data structure passed to the hipr function */
	char** network = NULL;

	/* adds each directed link to the list that feeds the hipr function */
	for (int i=0; i<T->GetNumNodes(); i++)
		for (TGraphNeighborIterator j=T->Node(i)->NeighborBegin(); j!=T->Node(i)->NeighborEnd(); j++)
		{
			numarcs++;
			network=(char **)realloc(network,(numarcs+4)*sizeof(char *));
			network[numarcs+2]=(char *)malloc(20*sizeof(char));
			sprintf(network[numarcs+2],"a %d %d %d",i, j->Id(), (int)j->Rate());
		}

	/* adds directed links between the supersource (id=T->node.size()) and each of the aggregation devices */
	for (int i = 0; i < T->GetNumAggregationNodes(); i++)
	{
		numarcs++;
		network=(char **)realloc(network,(numarcs+4)*sizeof(char *));
		network[numarcs+2]=(char *)malloc(20*sizeof(char));
		sprintf(network[numarcs+2],"a %zu %d %lu", T->GetNumNodes(), T->GetAggregationNode(i), numeric_limits<long>::max());
	}

	/* adds directed links between each of the gateways and the supersink (id=T->node.size()+1)  */
	for (int i = 0; i < T->GetNumGatewayNodes(); i++)
	{
		numarcs++;
		network=(char **)realloc(network,(numarcs+4)*sizeof(char *));
		network[numarcs+2]=(char *)malloc(20*sizeof(char));
		sprintf(network[numarcs+2],"a %d %zu %lu", T->GetGatewayNode(i), T->GetNumNodes()+1, numeric_limits<long>::max());
	}

	network[numarcs+3] = NULL;       /* NULL terminated array */

	network[0] = (char *)calloc(20,sizeof(char));   /* problem description */
	network[1] = (char *)calloc(20,sizeof(char));   /* source node */
	network[2] = (char *)calloc(20,sizeof(char));   /* destination node */
	sprintf(network[0],"p max %zu %d",T->GetNumNodes()+2,numarcs);
	sprintf(network[1],"n %zu s",T->GetNumNodes());
	sprintf(network[2],"n %zu t",T->GetNumNodes()+1);

//	for (int i=0; i<numarcs+3; i++)
//		printf("%s\n", network[i]);
	maxflow(T, network);

//	printf("\nMaxflow=%lf\n",flow);

	/* free memory */
	for (int i=0; i<numarcs+3; i++)
		free(network[i]);
	free(network);
}


/* computes the maxflow between every source-destination pair on the initial graph (given by T)
   and records the flow on each link */

void PFR_allpairsMF::pfr_calc(Topology *T)
{
	int     numarcs;

	/* data structure passed to the hipr function */
	char** network;

	// reset flow values
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			SetFlow (v, 1.);

	for (int s = 0; s < T->GetNumAggregationNodes(); s++)
	{
		numarcs = 0;
		network = NULL;

		/* adds each directed link to the list that feeds the hipr function */
		for (int i=0; i<T->GetNumNodes(); i++)
			for (TGraphNeighborIterator j=T->Node(i)->NeighborBegin(); j!=T->Node(i)->NeighborEnd(); j++)
			{
				numarcs++;
				network=(char **)realloc(network,(numarcs+4)*sizeof(char *));
				network[numarcs+2]=(char *)malloc(20*sizeof(char));
				sprintf(network[numarcs+2],"a %d %d %d",i, j->Id(), (int)j->Rate());
			}

		/* adds directed links between each of the gateways and the supersink (id=T->node.size())  */
		for (int i = 0; i < T->GetNumGatewayNodes(); i++)
		{
			numarcs++;
			network=(char **)realloc(network,(numarcs+4)*sizeof(char *));
			network[numarcs+2]=(char *)malloc(20*sizeof(char));
			sprintf(network[numarcs+2],"a %d %zu %lu", T->GetGatewayNode(i), T->GetNumNodes(), numeric_limits<long>::max());
		}

		network[numarcs+3] = NULL;       /* NULL terminated array */
	
		network[0] = (char *)calloc(20,sizeof(char));   /* problem description */
		network[1] = (char *)calloc(20,sizeof(char));   /* source node */
		network[2] = (char *)calloc(20,sizeof(char));   /* destination node */
		sprintf(network[0],"p max %zu %d",T->GetNumNodes(),numarcs);
		sprintf(network[1],"n %d s", T->GetAggregationNode(s));
		sprintf(network[2],"n %zu t",T->GetNumNodes());

	//	for (int i=0; i<numarcs+3; i++)
	//		printf("%s\n", network[i]);
		maxflow(T, network);

	//	printf("\nMaxflow=%lf\n",flow);
	
		/* free memory */
		for (int i=0; i<numarcs+3; i++)
			free(network[i]);
		free(network);
	}
}

PFR_Raniwala::PFR_Raniwala (const list<pair<string,string> >& params)
 : Bwdr (NULL),
   seed (RndVar::getSeed())
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("seed", seed);
    RP.read ("Bwdr", Bwdr);
  }
  
  if (!Bwdr)
    cerr << endl << "pfr_Raniwala: ERROR: No random variable set for bandwidth" << endl << endl;
  
  setStr();
}

void PFR_Raniwala::setStr()
{
  stringstream ss;
  
  ss << "pfr_Raniwala " << seed << " " << Bwdr->getName();
  
  char s[500];
  
  ss.getline(s, 500);
  
  PFR_method::setStr(s);
}


void PFR_Raniwala::pfr_calc(Topology *T)
{
	// reset flow values
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++) {
			SetFlow (v, 0.);
			ClearFlowDemands (v);
		}

	MotherOfAll urng(seed);              // declare uniform random number generator
	Random::Set(urng);                   // set urng as generator to be used

	// for each source-destination pair
	for (int s = 0; s < T->GetNumAggregationNodes(); s++)
		for (int d = 0; d < T->GetNumGatewayNodes(); d++)
		{
			int s_id = T->GetAggregationNode(s);
			int d_id = T->GetGatewayNode(d);
			// reset availBW values
			for (int u=0; u<T->GetNumNodes(); u++)
				for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
					SetAvailBW (v, 1.0);

			vector<TLinkItrs>* path;
			int numPaths = 0;

			linkPruneFunc* prune = new linkPrune_Bwd (1.0);
			path = T->ShortestPath(s_id, d_id, prune);

			while (!path->empty())
			{
				numPaths++;

				/* decrease availBW along the path */
				for (vector<TLinkItrs>::iterator l=path->begin(); l!=path->end(); l++)
					SetAvailBW (l->second, l->second->AvailBW() - 1.0);

				/* find another path */
				path = T->ShortestPath(s_id, d_id, prune);
			}

			/* randomly generate the traffic demand between the source-destination pair */
			double bw = (double)Bwdr->rv->Next();

			if (numPaths)
				/* Add a flow on the links involved in routing */
				for (int u=0; u<T->GetNumNodes(); u++)
					for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
						if (v->AvailBW() == 0.0)
						{
							SetFlow (v, v->Flow() + bw / numPaths);
							SetFlowDemand (v, pair<int,int>(s_id, d_id), bw / numPaths);
							TLinkItrs invlink = T->FindInverseLink(T->LinkItrs(T->Node(u),v));
							if (invlink.second != invlink.first->NeighborEnd()) {
								SetFlow (invlink.second, invlink.second->Flow() + bw / numPaths);
								SetFlowDemand (invlink.second, pair<int,int>(d_id, s_id), bw / numPaths);
							}
						}
			delete prune;
			delete path;
		}
}


PFR_Random::PFR_Random(const list<pair<string,string> >& params)
 : max (1.0),
   seed (RndVar::getSeed())
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("seed", seed);
    RP.read ("max", max);
  }

  setStr();
}

void PFR_Random::setStr()
{
  stringstream ss;
  
  ss << "pfr_Random " << seed << " " << max;
  
  char s[500];
  
  ss.getline(s, 500);
  
  PFR_method::setStr(s);
}


void PFR_Random::pfr_calc(Topology *T)
{
	MotherOfAll urng(seed);                    // declare uniform random number generator
	Random::Set(urng);                   // set urng as generator to be used
	Uniform flow;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
			SetFlow (v, flow.Next()*max);

}

PFR_MinMaxUtot::PFR_MinMaxUtot (const list<pair<string,string> >& params)
 : DemVar (NULL),
  seed (RndVar::getSeed()),
  ksp (false),
  alpha (2),
  k (10),
  capacity (false)
{
  for (list<pair<string,string> >::const_iterator p = params.begin(); p != params.end(); p++)
  {
    ReadParams RP (p->first, p->second);
    
    RP.read ("seed", seed);
    RP.read ("demands", DemVar);
    RP.read ("ksp", ksp);
    RP.read ("alpha", alpha);
    RP.read ("k", k);
    RP.read ("capacity", capacity);
  }
  
  if (!DemVar)
    cerr << endl << "pfr_MinMaxUtot: ERROR: No random variable set for demands" << endl << endl;

  setStr();
}

void PFR_MinMaxUtot::setStr()
{
  stringstream ss;
  
  ss << "pfr_MinMaxUtot " << seed << " " << DemVar->getName() << " " << (ksp ? "1 " : "0 ");
  
  if (ksp)
    ss << alpha << " " << k;
  
  ss << (capacity ? "1 " : "0 ");
  
  char s[500];
  
  ss.getline(s, 500);
  
  PFR_method::setStr(s);
}

/*
 * Uses MinMaxUtot, which routes a given set of demands to minimize the maximum
 * total utilization, to compute the initial flow rates
 * Since MinMaxUtot requires that channels have been already assigned, 1Channel is
 * invoked at the beginning and ResetChannels is invoked at the end
 */
void PFR_MinMaxUtot::pfr_calc(Topology* T)
{
  vector< set< pair< int, int > > > Sd;
  double maxUtot = numeric_limits<double>::max();
  MotherOfAll* urng;

  /* Determine the set of demands */
  urng = new MotherOfAll(seed);         // declare uniform random number generator
  Random::Set(*urng);                   // set urng as generator to be used

  T->SetThroughput (0);
  
  OneChannel OC;
  OC.Run(T);

  map<pair<int,int>,double> demands;
  for (int a = 0; a < T->GetNumAggregationNodes(); a++)
    for (int g = 0; g < T->GetNumGatewayNodes(); g++)
    {
      int a_id = T->GetAggregationNode(a);
      int g_id = T->GetGatewayNode(g);
      
      demands[pair<int,int>(a_id,g_id)] = DemVar->rv->Next();
      demands[pair<int,int>(g_id,a_id)] = DemVar->rv->Next();
      
      T->SetThroughput (T->GetThroughput() + demands[pair<int,int>(a_id,g_id)] + demands[pair<int,int>(g_id,a_id)]);
    }

  delete urng;

  if (ksp) {
    /* find the set of links carrying the traffic of each demand */      
    for (map<pair<int,int>,double>::iterator d=demands.begin(); d!=demands.end(); d++)
      Sd.push_back(T->find_path(pair<int,int>(d->first.first,d->first.second), alpha, k));
  }
  
//   if (MinMaxUtot(T, &demands, maxUtot, Sd, capacity) != OPTIMAL)
  MinMaxUtot mmu (T, &demands, &maxUtot, Sd, capacity);
  if (mmu.lp_solve())
    cerr << endl << "WARNING: PFR_MinMaxUtot: An optimal solution to the LP has not been found,"
    << " hence the initial flow rates have not been set" << endl << endl;

  T->ResetChannels();
}


/* determines the pre-computed flow rates by solving the Maximum Minimum Congestion Factor problem and returns the sum of all the commodity flows */
void PFR_mmcf::pfr_calc(Topology *T)
{
	// compute the number of commodities (each aggregation router/gateway acts both as a source and a sink)
	unsigned int numComms = T->GetNumAggregationNodes() * T->GetNumGatewayNodes() * 2;

	MaxMultiCommFlow mmcf (T);
	mmcf.lp_solve();
	double maxF = mmcf.getObjValue();

	MaxMinCongFactor mmc (T, maxF);
	mmc.lp_solve();
	double * vars = mmc.getPrimalColumnSolution();

	unsigned int linkId = 0;
	for (int u=0; u<T->GetNumNodes(); u++)
		for (TGraphNeighborIterator v=T->Node(u)->NeighborBegin(); v!=T->Node(u)->NeighborEnd(); v++)
		{
			SetFlow (v, 0.);
			ClearFlowDemands (v);
			int i=0;
			for (int a = 0; a < T->GetNumAggregationNodes(); a++)
				for (int g = 0; g < T->GetNumGatewayNodes(); g++)
				{
					if (vars[linkId*numComms + i])
					{
						SetFlowDemand (v, pair<int,int>(T->GetAggregationNode(a),T->GetGatewayNode(g)), vars[linkId*numComms + i]);
						SetFlow (v, v->Flow() + vars[linkId*numComms + i]);
					}
				}
			for (int g = 0; g < T->GetNumGatewayNodes(); g++)
				for (int a = 0; a < T->GetNumAggregationNodes(); a++)
				{
					if (vars[linkId*numComms + i])
					{
						SetFlowDemand (v, pair<int,int>(T->GetGatewayNode(g),T->GetAggregationNode(a)), vars[linkId*numComms + i]);
						SetFlow (v, v->Flow() + vars[linkId*numComms + i]);
					}
				}
			linkId++;
		}

	T->SetThroughput (maxF);
}
#if 0
void PFR_ksp::pfr_calc(Topology* T)
{
  lprec *lp;
  int Ncol;
  set<pair<int,int> > demands;

  for (vector<int>::iterator s=T->aggregation_list.begin(); s!=T->aggregation_list.end(); s++)
    for (vector<int>::iterator d=T->gateway_list.begin(); d!=T->gateway_list.end(); d++) {
      demands.insert(pair<int,int>(*s,*d));
      demands.insert(pair<int,int>(*d,*s));
    }

  cout << "Demands:" << endl;
  for(set<pair<int,int> >::iterator m=demands.begin(); m!=demands.end(); m++)
    cout << "d[" << m->first << "," << m->second << "]" << endl;
  
  // compute the number of demands (each aggregation router/gateway acts both as a source and a sink)
  unsigned int numDemands = demands.size();
  
  // compute the cumulative number of neighbors for each node. It is such that:
  // - the number of outgoing links of node i is cumNumNeighbs[i+1] - cumNumNeighbs[i]
  // - the last element of cumNumNeighbs is the total number of links
  unsigned int cumNumNeighbs[T->node.size() + 1];

  cumNumNeighbs[0] = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
  	  cumNumNeighbs[u->info->id + 1] = cumNumNeighbs[u->info->id] + u->neighbor.size();

  // the number of variables is numDemands*numLinks (f^ab(u->v)) + numDemands (F^d)
  Ncol = numDemands*cumNumNeighbs[T->node.size()] + numDemands;

  // create the LP problem
  if ((lp = make_lp(0, Ncol)) == NULL)
    throw -1;

  /* Columns (variables) are ordered this way:
      f_e1,d1 ... f_e1,dD ... f_eE,d1 ... f_eE,dD  F^d1 ... F^dD
      where links e1 ... eE are given by
      neighbors(node 0) neighbors(node 1) ... neighbors(node N)
      and demands d1 ... dD are given in the order they are present in demands
  */

  /* give a name to the variables */
  int linkId = 0;
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int i=1;
      for (set<pair<int,int> >::iterator d=demands.begin(); d!=demands.end(); d++, i++) {
	stringstream s;
	s << "f(" << u->info->id << "," << v->info->id << ")[" << d->first << d->second << "]";
	set_col_name(lp, linkId*numDemands+i, (char*)s.str().c_str());
      }
      linkId++;
    }

  for (set<pair<int,int> >::iterator d=demands.begin(); d!=demands.end(); d++){
    stringstream s;
    s << "F[" << d->first << "," << d->second << "]";
    set_col_name(lp, numDemands*cumNumNeighbs[T->node.size()]+1+distance(demands.begin(),d), (char*)s.str().c_str());
  }

  int* colno = new int[Ncol];
  REAL* coeff = new REAL[Ncol];
  
  /* set the objective function: max sum_d in D F^d */
  for (int i=0; i<numDemands; i++) {
    coeff[i] = 1;
    colno[i] = numDemands*cumNumNeighbs[T->node.size()]+i+1;
  }
  set_obj_fnex(lp, numDemands, coeff, colno);

  /* set the object direction to minimize */
  set_maxim(lp);

  
  set_add_rowmode(lp, TRUE);  /* makes building the model faster if it is done rows by row */

  vector<vector<int> > colno_v(T->node.size()*numDemands);
  vector<vector<REAL> > coeff_v(T->node.size()*numDemands);
  
  /* flow conservation constraints  ( |V| * numDemands constraints in total)
   * We calculate them all together, in order to perform just a single iteration over all the network links
   * of course, more memory is needed
   * We also exploit this loop to set some variables to zero
   */
  unsigned int dist;

  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      int d=0;
      for (set<pair<int,int> >::iterator di=demands.begin(); di!=demands.end(); di++, d++) {
      // for (unsigned int d=0; d<numDemands; d++) {
	dist = distance(u->neighbor.begin(), v);
	/* flow leaving node u */
	coeff_v[u->info->id*numDemands+d].push_back(1);
	colno_v[u->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	/* flow entering node v */
	coeff_v[v->info->id*numDemands+d].push_back(-1);
	colno_v[v->info->id*numDemands+d].push_back((cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1);
	/* if v is the source of the demand or u is the destination of the demand, f(u->v)[d] must be null */
	if ( (u->info->id == di->second) || (v->info->id == di->first) )
	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, 0.);
//	else
//	  set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + d + 1, di->second);
      }
    }

  /* Add the flow conservation constraints one-by-one */
  int k=0;           // index of the constraint to be added
  for (unsigned int i=0; i<T->node.size(); i++)
    for (set<pair<int,int> >::iterator d=demands.begin(); d!=demands.end(); d++) {
      copy(coeff_v[k].begin(), coeff_v[k].end(), coeff);
      copy(colno_v[k].begin(), colno_v[k].end(), colno);

      /* if i is the source of the demand d */
      if (i == d->first) {
	coeff[coeff_v[k].size()] = -1;
	colno[colno_v[k].size()] = numDemands*cumNumNeighbs[T->node.size()]+1+distance(demands.begin(),d);
	add_constraintex(lp, coeff_v[k].size()+1, coeff, colno, EQ, 0);
      }
      /* if i is the destination of the demand d */
      else if (i == d->second) {
	coeff[coeff_v[k].size()] = 1;
	colno[colno_v[k].size()] = numDemands*cumNumNeighbs[T->node.size()]+1+distance(demands.begin(),d);
	add_constraintex(lp, coeff_v[k].size()+1, coeff, colno, EQ, 0);
      }
      else
	add_constraintex(lp, coeff_v[k].size(), coeff, colno, EQ, 0);
      k++;
    }

  /* capacity constraints  ( |E| constraints in total) */
  /* This set of constraints often make the LP infeasible 
   * because it is not possible to satisfy them while routing
   * the requested e2e demands.
   * It is anyway safe to remove them because the goal of minimizing M
   * will force to minimize the e2e flows
   */
#if 0
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      fill(coeff, coeff+numDemands, 1);
      for (int i=0; i<numDemands; i++)
	colno[i] = (cumNumNeighbs[u->info->id] + distance(u->neighbor.begin(),v))*numDemands +i +1;
      
      add_constraintex(lp, numDemands, coeff, colno, LE, v->rate);
    }
#endif

  /* constraints on the total utilizations ( |E| constraints in total) */
  for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
    for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++) {
      k = 0;
      for (vector<TGraphNode>::iterator x=T->node.begin(); x!=T->node.end(); x++)
	for (list<TGraphNeighbor>::iterator y=x->neighbor.begin(); y!=x->neighbor.end(); y++)
	  if ( TLinkItrs(x,y).IsInTheCollDomainOf(T,TLinkItrs(u,v)) ) {
	    fill(coeff+k, coeff+k+numDemands, 1. / y->rate);
	    for (unsigned int i=0; i<numDemands; i++)
	      colno[k+i] = (cumNumNeighbs[x->info->id] + distance(x->neighbor.begin(),y))*numDemands +i +1;
	    k += numDemands;
	  }
      add_constraintex(lp, k, coeff, colno, LE, 1 );
    }
    
  /* constraints to only have non-null flow on links belonging to the first (at
   * most k) shortest paths with length at most alpha times the shortest length */
  /* ksp false means not to include such constraints */
  if (ksp) {
    set<pair<int,int> > fksp;
    
    for (set<pair<int,int> >::iterator d=demands.begin(); d!=demands.end(); d++) {
      fksp = T->find_path(pair<int,int>(d->first,d->second), alpha, K);
      for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
	for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
	  if (fksp.find(pair<int,int>(u->info->id,v->info->id)) == fksp.end()) {
	    dist = distance(u->neighbor.begin(), v);
	    set_upbo(lp, (cumNumNeighbs[u->info->id] + dist)*numDemands + distance(demands.begin(),d) + 1, 0.);
	  }
    }
  }
    
  set_add_rowmode(lp, FALSE);            /* rowmode should be turned off again when done building the model */

  // write_LP(lp, stdout);
  /* I only want to see important messages on screen while solving */
  // set_verbose(lp, IMPORTANT);

  /* Now let lpsolve calculate a solution */
  int ret = solve(lp);
  if (ret != OPTIMAL)
    cerr << endl << "WARNING: ksp: An optimal solution to the LP has NOT been found and hence this function made no changes to the topology" << endl << endl;
//   else
//     print_solution(lp,4);

  /* If an optimal solution has been found and maxUtot < prevMaxUtot, update the flows, the e2e flows and the Utot */
  if (ret==OPTIMAL) {

    T->throughput = get_objective(lp);
    cout << endl << "Throughput: " << T->throughput << endl;
    
    REAL *ptr_var, *ptr_constr;
    
    get_ptr_variables(lp, &ptr_var);
    get_ptr_constraints(lp, &ptr_constr);
    
    unsigned int linkId = 0;
    for (vector<TGraphNode>::iterator u=T->node.begin(); u!=T->node.end(); u++)
      for (list<TGraphNeighbor>::iterator v=u->neighbor.begin(); v!=u->neighbor.end(); v++)
      {
	v->flow = 0.;
	v->f_e2e.clear();

	int i=0;
	for (set<pair<int,int> >::iterator d=demands.begin(); d!=demands.end(); d++, i++)
	  if (ptr_var[linkId*numDemands + i])
	    v->flow += (v->f_e2e[pair<int,int>(d->first,d->second)] = ptr_var[linkId*numDemands + i]);
	
	v->Utot = ptr_constr[T->node.size()*numDemands+linkId];

	linkId++;
      }
  }
  
  /*  Before quitting, clear memory*/
  delete colno;
  delete coeff;

  /* clean up such that all used memory by lpsolve is freed */
  if (lp)
    delete_lp(lp);
}
#endif

