#ifndef WIMESH_TYPES
#define WIMESH_TYPES

#include <vector>
#include <list>
#include <set>
#include <map>
#include <algorithm>
#include <limits>
#include <fstream>
#include <sstream>
#include <cmath>
#include "rndvars.h"



using namespace std;


class InterferenceModel;
class PropagationLossModel;

/*
 * Class specifying various parameters of a simulation
 */
class WiMeshSimulation
{
public:
  // constructor (defined in topology.cpp). Sets all the parameters to their default values
  WiMeshSimulation ();

  unsigned int numAvailChannels () { return m_numAvailChannels; }

  float availTXrates (int i) { return m_availTXrates.at(i); }
  int min_rate () { return m_min_rate; }
  int max_rate () { return m_max_rate; }
  int numTXrates () { return m_availTXrates.size(); }

  float SINRthresholds (float rate) { return m_SINRthresholds[rate] / m_SINRthreshfactor; }
  float noise () { return m_noise; }
  float SINRthreshfactor () { return m_SINRthreshfactor; }

  float minTxPowerLevel () { return m_minTxPowerLevel; }
  float maxTxPowerLevel () { return m_maxTxPowerLevel; }
  unsigned int numTxPowerLevels () { return m_numTxPowerLevels; }
  unsigned int defaultTxPowerLevel () { return m_defaultTxPowerLevel; }

  // return the i-th tx power level in dBm (0 <= i <= numTxPowerLevels-1)
  float TxPowerLevel (unsigned int i) { return m_minTxPowerLevel + i * (m_maxTxPowerLevel - m_minTxPowerLevel) / (m_numTxPowerLevels - 1); }

  const InterferenceModel* InterfModel () { return m_interferenceModel; }
  const PropagationLossModel* PropagationModel () { return m_propagationLossModel; }

  string simulId () { return m_simulId; }
  void appendToSimulId (string s) { m_simulId += s; }

  void SerializeTo (ofstream& ofs);
  void DeserializeFrom (ifstream& ifs);
  
private:

  unsigned int m_numAvailChannels;

  vector<float> m_availTXrates;
  // minimum and maximum rates as passed to buildGi_FixedPowerMaxAllowRate. Used by some CA algorithms
  int m_min_rate;
  int m_max_rate;
  
  map<float,float> m_SINRthresholds;           // pairs (rate, SINR threshold)
  float m_noise;
  // a multiplicative factor (>=1) for the SINR thresholds, ONLY used when determining
  // whether a link can be established at a given rate
  float m_SINRthreshfactor;

  // Tx power levels (dBm)
  // For ns-3 compatibility, we specify the minimum and maximum values and the number of levels
  float m_minTxPowerLevel;
  float m_maxTxPowerLevel;
  unsigned int m_numTxPowerLevels;
  unsigned int m_defaultTxPowerLevel;

  InterferenceModel* m_interferenceModel;
  PropagationLossModel* m_propagationLossModel;

  string m_simulId;

  friend class WiMeshParser;
  friend class Topology;
};


// global variable, declared in topology.cpp
extern WiMeshSimulation WMS;



// P = 30 + 10*log(W)  where P is the power in dbm and W is the power in Watt
// For ns-3 compatibility, values (in dbm) should be equally spaced among the minimum
// and the maximum
// const float availTxPwr[] = {
// 	0.00100f,	//  0 dbm
// 	0.00125f,	//  1 dbm
// 	0.00158f,	//  2 dbm
// 	0.00200f,	//  3 dbm
// 	0.00251f,	//  4 dbm
// 	0.00316f,	//  5 dbm
// 	0.00398f,	//  6 dbm
// 	0.00501f,	//  7 dbm
// 	0.00631f,	//  8 dbm
// 	0.00794f,	//  9 dbm
// 	0.01000f,	// 10 dbm
// 	0.01259f,	// 11 dbm
// 	0.01585f,	// 12 dbm
// 	0.01995f,	// 13 dbm
// 	0.02512f,	// 14 dbm
// 	0.03162f,	// 15 dbm
// 	0.03981f,	// 16 dbm
// 	0.05012f,	// 17 dbm
// 	0.06310f,	// 18 dbm
// 	0.07943f,	// 19 dbm
// 	0.10000f	// 20 dbm
// };
// #define NUMTXPWR sizeof(availTxPwr)/sizeof(float) 


// P = 30 + 10*log(W) = 10*log(1000*W) where P is the power in dBm and W is the power in Watt
inline float DbmToW (float dBm) { return pow (10.0f, dBm / 10.0f) / 1000.0f; }




typedef enum { router, aggregation, gateway } TMeshRouter;

const char MeshTypes[sizeof(TMeshRouter)] = "RAG";


#define almost_null exp(-10.)



// Power consumption values (Watts)

class NodePowerConsumptionProfile
{
public:
  float BasePowerConsumption;
  float StandbyPowerConsumption;
  
  NodePowerConsumptionProfile (float bpc, float spc)
   : BasePowerConsumption (bpc), StandbyPowerConsumption (spc) {}
};

const NodePowerConsumptionProfile defaultNodePCprofile =
      NodePowerConsumptionProfile (2.29f, 0.50f);



/***************************************************************************************
 * Abstract base class for radio power consumption in transmission mode function object
 ***************************************************************************************/
class foRadioTxPower
{
public:
  virtual ~foRadioTxPower() {}

  virtual foRadioTxPower* Clone () = 0;
  virtual float operator()(float power) = 0;

  virtual void SerializeTo (ofstream& ofs) = 0;
  static foRadioTxPower* DeserializeFrom (ifstream& ifs);

  virtual string GetNameWithParams () = 0;
};

/* The radio power consumption is a linear function of the transmission
 * power level (availTxPwr values): a * power + b
 * This is the default method, with a = 0 and b = 3.17
 * NOTE If you define new classes derived from foRadioTxPower, remember to define 
 * the string name in topology.cpp and update foRadioTxPower::DeserializeFrom
 * (topology.cpp) and WiMeshParser::on_start_element (parsers.cpp)
 */
class foRadioTxPowerLinear: public foRadioTxPower
{
  float a, b;
public:
  foRadioTxPowerLinear (float _a, float _b) : a(_a), b(_b) {}

  float operator()(float power) { return a * power + b; }
  foRadioTxPower* Clone () { return new foRadioTxPowerLinear (*this); }

  void SerializeTo (ofstream& ofs);
  static foRadioTxPowerLinear* DeserializeFrom (ifstream& ifs);

  float getA () { return a; }
  float getB () { return b; }

  static constexpr char name[] = "linear";

  string GetNameWithParams () { stringstream ss; ss << name << "(" << a << ", " << b << ")"; return ss.str(); }
};


class RadioPowerConsumptionProfile
{
public:
  foRadioTxPower* rho_tx;
  float rho_rx;
  float rho_idle;
  float rho_sleep;

  RadioPowerConsumptionProfile (foRadioTxPower* tx, float rx, float idle, float sleep)
   : rho_tx(tx), rho_rx(rx), rho_idle(idle), rho_sleep(sleep) {}
};

const RadioPowerConsumptionProfile defaultRadioPCprofile =
      RadioPowerConsumptionProfile (new foRadioTxPowerLinear(0, 3.17f), 2.36f, 1.22f, 0.29f);



/*******************
 * RadioIface class
 *******************/

class RadioIface
/* contains properties of a radio interface */
{
public:

  /* constructor */
  RadioIface (float _tx_gain_db = default_db_gain,
              float _rx_gain_db = default_db_gain,
              unsigned int _minangle = default_min_angle,
              unsigned int _maxangle = default_max_angle,
              foRadioTxPower* _rho_tx = defaultRadioPCprofile.rho_tx->Clone(),
              float _rho_rx = defaultRadioPCprofile.rho_rx,
              float _rho_idle = defaultRadioPCprofile.rho_idle,
              float _rho_sleep = defaultRadioPCprofile.rho_sleep
  ) : 
    channel (0),
    tx_gain_db (_tx_gain_db),
    rx_gain_db (_rx_gain_db),
    minangle (_minangle),
    maxangle (_maxangle),
    rho_tx (_rho_tx),
    rho_rx (_rho_rx),
    rho_idle (_rho_idle),
    rho_sleep (_rho_sleep)
  {}
  
  /* copy constructor */
  RadioIface (const RadioIface& ri)
  {
    channel = ri.channel;
    tx_gain_db = ri.tx_gain_db;
    rx_gain_db = ri.rx_gain_db;
    minangle = ri.minangle;
    maxangle = ri.maxangle;
    rho_tx = ri.rho_tx->Clone();
    rho_rx = ri.rho_rx;
    rho_idle = ri.rho_idle;
    rho_sleep = ri.rho_sleep;
  }

  ~RadioIface () { delete rho_tx; }

  int Channel() const { return channel; }
  float TxGainDb () const { return tx_gain_db; }
  float RxGainDb () const { return rx_gain_db; }
  unsigned int MinAngle () const { return minangle; }
  unsigned int MaxAngle () const { return maxangle; }
  foRadioTxPower* Rho_Tx () const { return rho_tx; }
  float Rho_Rx () const { return rho_rx; }
  float Rho_Idle () const { return rho_idle; }
  float Rho_Sleep () const { return rho_sleep; }

  void SerializeTo (ofstream& ofs);
  void DeserializeFrom (ifstream& ifs);

  // default values
  static constexpr float default_db_gain = 1.0;
  static const unsigned int default_min_angle = 0;
  static const unsigned int default_max_angle = 360;

private:

  unsigned int channel;	/* 0 stands for unset (for unused after channel assignment) */
  float tx_gain_db;
  float rx_gain_db;
  unsigned int minangle;
  unsigned int maxangle;
  foRadioTxPower* rho_tx;
  float rho_rx;
  float rho_idle;
  float rho_sleep;

  friend class RadioIfaceFriend;
  friend class Topology;
  friend class TopologyParser;
  friend class WiMeshParser;
};

typedef vector<RadioIface>::iterator RadioIfaceIterator;



/*****************
 * MeshNode class
 *****************/

class MeshNode
/* contains properties of a mesh node */
{
public:

  /* constructor */
  MeshNode(int _id, float _xpos, float _ypos, TMeshRouter _role, int _numAvailRadios, 
           float bpc = defaultNodePCprofile.BasePowerConsumption, float spc = defaultNodePCprofile.StandbyPowerConsumption) :
    id (_id),
    xpos (_xpos),
    ypos (_ypos),
    role (_role),
    BasePowerConsumption (bpc),
    StandbyPowerConsumption (spc),
    radio(vector<RadioIface>(_numAvailRadios))
  {}

  /* copy constructor */
// 	MeshNode(const MeshNode& m) : id(m.id), xpos(m.xpos), ypos(m.ypos), role(m.role), numAvailRadios(m.numAvailRadios), radio(new RadioIface[m.numAvailRadios])
// 	{
// 		for (unsigned int i=0; i<m.numAvailRadios; i++)
// 			radio[i] = m.radio[i];
// 	}

  /* destructor */
// 	~MeshNode() { if (radio) delete [] radio; }

  unsigned int numAvailRadios() { return radio.size(); }

  void SerializeTo (ofstream& ofs);
  static MeshNode* DeserializeFrom (ifstream& ifs);

  int Id () const { return id; }
  float Xpos () const { return xpos; }
  float Ypos () const { return ypos; }

  TMeshRouter Role () const { return role; }

  float GetBasePowerConsumption () const { return BasePowerConsumption; }
  float GetStandbyPowerConsumption () const { return StandbyPowerConsumption; }

  /* returns the number of radios */
  int GetNumRadios () { return radio.size(); }

  /* returns an iterator pointing to the i-th radio */
  RadioIfaceIterator Radio (int i) { return radio.begin() + i; }

  int countRadiosWithChannel(int c);

  /* returns the index of the radio set on the given channel or -1 if such a radio is not found */
  int findRadioWithChannel (int c);

  /* returns the set of channels assigned to the radios (except channel 0) */
  set<unsigned int> findChannels ();

private:
  
  int id;                              /* node unique id */
  float xpos, ypos;                    /* node X and Y coordinates */
  const TMeshRouter role;              /* role of the node */
  float BasePowerConsumption;          /* base power consumption */
  float StandbyPowerConsumption;       /* standby power consumption */
  vector<RadioIface> radio;            /* vector of radio objects */

friend class TopologyParser;
friend class WiMeshParser;
};

class Topology;

typedef map<pair<int,int>,double>::const_iterator FlowDemandsIterator;



/***********************
 * TGraphNeighbor class
 ***********************/

class TGraphNeighbor
/* characterizes a neighbor in Topology */
{
public:

  /* constructor */
  TGraphNeighbor(MeshNode *_info, float _rate, unsigned int _txpower): info(_info), radio(NULL), availBW(0.), flow(0.), Utot(0.), rate(_rate), txpower(_txpower) {}

  void SerializeTo (ofstream& ofs);
  static TGraphNeighbor* DeserializeFrom (ifstream& ifs, Topology* T, int me);

  /* returns a pointer to node's info structure */
  MeshNode* Info () { return info; }

  /* returns the id of the neighbor */
  int Id () const { return info->Id(); }

  /* returns a pointer to the info structure of the radio used to reach the neighbor */
  RadioIface* Radio () { return radio; }

  /* returns the channel used to reach the neighbor */
  unsigned int RadioChannel () const { if (radio) return radio->Channel();  return 0; }

  /* returns the value of availBW */
  double AvailBW () const { return availBW; }

  /* returns the value of flow */
  double Flow () const { return flow; }

  /* returns the value of Utot */
  double UTot () const { return Utot; }

  /* returns the value of rate */
  float Rate () const { return rate; }

  /* returns the value of txpower */
  unsigned int TxPower () const { return txpower; }

  /* returns true if some flow of the given demand is routed on this link */
  bool IsDemandPresent (pair<int,int> p) const { return f_e2e.find (p) != f_e2e.end(); }

  /* returns the value of the flow associated with the given demand */
  double F_e2e (pair<int,int> p) { return f_e2e[p]; }

  /* returns an iterator pointing to the first demand */
  FlowDemandsIterator F_e2eBegin () const { return f_e2e.begin(); }

  /* returns an iterator pointing to past the last demand */
  FlowDemandsIterator F_e2eEnd () const { return f_e2e.end(); }

private:

  MeshNode *info;                      /* pointer to neighbor's info structure */
  RadioIface *radio;                   /* pointer to info structure of the radio used to reach the neighbor
                                          Must be set to NULL in the initial graph */
  double availBW;
  double flow;                         /* flow rate on the directed link */
  double Utot;                         /* the total utilization of the collision domain of this directed link */
  float rate;                          /* the transmission rate on this directed link, initialized to the maximum rate, but
                                          set according to TXrateFunc (specified in the config file) by buildInitialGraph */
  unsigned int txpower;                /* an index corresponding to the Tx power level in dBm */

  map<pair<int,int>,double> f_e2e;     /* amount of flow due to the end-to-end flow between i and j */
                                       /* The sum of all these values MUST equal the flow field */

  friend class TGraphNeighborFriend;
  friend class Topology;
};

typedef list<TGraphNeighbor>::iterator TGraphNeighborIterator;



/*******************
 * TGraphNode class
 *******************/

class TGraphNode
/* characterizes a node in Topology */
{
public:

  /* constructor */
  TGraphNode (MeshNode* _info = NULL) : info(_info) {}
  /* copy constructor */
  TGraphNode(const TGraphNode& n) : info(new MeshNode(*(n.info))), neighbor(n.neighbor) {}
  /* destructor */
  ~TGraphNode() { if (info) delete info; }

  /* returns a pointer to node's info structure */
  MeshNode* Info () { return info; }

  /* returns the id of the node */
  int Id () const { return info->Id(); }

  /* returns the number of neighbors */
  int GetNumNeighbors () { return neighbor.size(); }

  /* returns an iterator pointing to the first node */
  TGraphNeighborIterator NeighborBegin () { return neighbor.begin(); }

  /* returns an iterator pointing to past the last node */
  TGraphNeighborIterator NeighborEnd () { return neighbor.end(); }

  /* Searches node v among the neighbors of u (using the specified channel, if any)
   * Returns NeighborEnd() if not found, otherwise an iterator 
   * pointing to the (first) found TGraphNeighbor element
   */
  TGraphNeighborIterator FindNeighbor (int v, unsigned int c = 0);

  /* Start the search from a specified neighbor (useful to find successive links to v */
  TGraphNeighborIterator FindNeighbor (int v, TGraphNeighborIterator n, unsigned int c = 0);

private:

  MeshNode *info;                      /* pointer to node's info structure */

  list<TGraphNeighbor> neighbor;       /* list of neighbors */

  friend class TGraphNodeFriend;
  friend class Topology;
};

typedef vector<TGraphNode>::iterator TGraphNodeIterator;



/* Function object that can be used to sort containers of pairs based on the second element of the pairs */
template <class T1, class T2>
class compareSecond {
  public:
    bool operator()(const pair<T1,T2>& x, const pair<T1,T2>& y) { return x.second < y.second; }
};


typedef pair<TGraphNodeIterator, TGraphNeighborIterator> _TLinkItrs;

class Topology;



/******************
 * TLinkItrs class
 ******************/

/*
 * The TLinkItrs class is basically a pair <TGraphNodeIterator, TGraphNeighborIterator>
 * The constructor is private, so use the Topology::LinkItrs method instead
 * The default copy constructor is instead accessible
 */
class TLinkItrs : public _TLinkItrs
{
public:
  bool PotInterferesWith (const TLinkItrs &l) const;

  bool IsInTheCollDomainOf (const TLinkItrs &l) const;

  Topology* topology () const { return m_topology; }

private:
  // constructor
  TLinkItrs(Topology* T, TGraphNodeIterator u, TGraphNeighborIterator v)
  : m_topology (T),
    _TLinkItrs(u,v)
  {}

  Topology* m_topology;

  friend class Topology;  
};


/* required to compare TLinkItrs objects
   defined in topology.cpp */
bool operator<(const list<TGraphNeighbor>::iterator& x, const list<TGraphNeighbor>::iterator& y);

bool operator<(const TGraphNeighbor& x, const TGraphNeighbor& y);



/**********************************************************************
 * Abstract base class for link cost function objects used by Dijkstra
 **********************************************************************/
class linkCostFunc
{
  public:
    virtual float operator()(TLinkItrs l) = 0;
};

/* Default link cost function object: c(l) = 1 */
class linkCost_Unary : public linkCostFunc
{
  public:
    float operator()(TLinkItrs l) { return 1.0; }
};



/*************************************************************************
 * Abstract base class for link pruning function objects used by Dijkstra
 *************************************************************************/
class linkPruneFunc
{
  public:
    virtual float operator()(TLinkItrs l) = 0;
};

/* Link pruning function object: availBW < BW 
   BW=0 implies all links are considered */
class linkPrune_Bwd : public linkPruneFunc
{
  double bw;
  public:
    linkPrune_Bwd(double _bw) : bw(_bw) {}
    float operator()(TLinkItrs l) { return l.second->AvailBW() < bw; }
};



/*************************************************
 * Abstract base class for the interference model
 *************************************************/

class InterferenceModel
{
public:
  virtual ~InterferenceModel() {}

  /* returns true if x->y potentially interferes with u->v */
  virtual bool PotInterfLinks (const TLinkItrs &xy, const TLinkItrs &uv) const = 0;

  virtual void SerializeTo (ofstream& ofs) = 0;
  static InterferenceModel* DeserializeFrom (ifstream& ifs);

  virtual string Description () const = 0;
};


/* NOTE If you define new classes derived from InterferenceModel, remember to define 
 * the string name in topology.cpp and update InterferenceModel::DeserializeFrom
 * (topology.cpp) and WiMeshParser::on_start_element (parsers.cpp)
 */
class SINRbasedInterferenceModel : public InterferenceModel
{
public:
  virtual bool PotInterfLinks (const TLinkItrs &xy, const TLinkItrs &uv) const;

  void SerializeTo (ofstream& ofs);
  static SINRbasedInterferenceModel* DeserializeFrom (ifstream& ifs);

  virtual string Description () const;

  static constexpr char name[] = "SINRbased";
};

class kHopNeighbInterferenceModel : public InterferenceModel
{
public:
  kHopNeighbInterferenceModel (int _k = 2) : k (_k) {}

  virtual bool PotInterfLinks (const TLinkItrs &xy, const TLinkItrs &uv) const;

  void SerializeTo (ofstream& ofs);
  static kHopNeighbInterferenceModel* DeserializeFrom (ifstream& ifs);

  virtual string Description () const;

  static constexpr char name[] = "khopNeighbor";

private:
  // returns 0 if u == v, 1 if u and v are neighbors, 2 if u and v are 2-hop neighbors (and k==3), 3 otherwise
  int dist (Topology* T, int u, int v) const;

  int k;
};



/*****************************************************
 * Abstract base class for the propagation loss model
 *****************************************************/

/* NOTE Only parameters of type double are allowed */
class PropagationLossModel
{
public:
  virtual ~PropagationLossModel() {}

  /* return the loss (dB, negative value) due to signal propagation */
  virtual double Loss (double distance) const = 0;

  vector<pair<string,double> >::const_iterator Begin () const { return params.begin(); }
  vector<pair<string,double> >::const_iterator End () const { return params.end(); }

  virtual void SerializeTo (ofstream& ofs) = 0;
  static PropagationLossModel* DeserializeFrom (ifstream& ifs);

  virtual const char* Name () const = 0;
  virtual string Description () const;

protected:
  vector<pair<string,double> > params;
};


/* NOTE If you define new classes derived from PropagationLossModel, remember to define 
 * the string name in topology.cpp and update PropagationLossModel::DeserializeFrom
 * (topology.cpp) and WiMeshParser::on_start_element (parsers.cpp)
 */
class LogDistancePropagationLossModel : public PropagationLossModel
{
public:
  // constructor
  LogDistancePropagationLossModel (double exponent = default_exponent,
                                   double referenceDistance = default_referenceDistance,
                                   double referenceLoss = default_referenceLoss)
  {
    params.push_back (pair<string,double> ("Exponent", exponent));
    params.push_back (pair<string,double> ("ReferenceDistance", referenceDistance));
    params.push_back (pair<string,double> ("ReferenceLoss", referenceLoss));
  }

  virtual double Loss (double distance) const;

  void SerializeTo (ofstream& ofs);
  static LogDistancePropagationLossModel* DeserializeFrom (ifstream& ifs);

  virtual const char* Name () const { return name; }

  // ns-3 default values
  static constexpr double default_exponent = 3.0;
  static constexpr double default_referenceDistance = 1.0;
  static constexpr double default_referenceLoss = 46.6777;       // Friis at 1m with 5.15 GHz

  static constexpr char name[] = "LogDistance";
};

class ThreeLogDistancePropagationLossModel : public PropagationLossModel
{
public:
  // constructor
  ThreeLogDistancePropagationLossModel (double distance0 = default_distance0,
                                        double distance1 = default_distance1,
                                        double distance2 = default_distance2,
                                        double exponent0 = default_exponent0,
                                        double exponent1 = default_exponent1,
                                        double exponent2 = default_exponent2,
                                        double referenceLoss = default_referenceLoss)
  {
    params.push_back (pair<string,double> ("Distance0", distance0));
    params.push_back (pair<string,double> ("Distance1", distance1));
    params.push_back (pair<string,double> ("Distance2", distance2));
    params.push_back (pair<string,double> ("Exponent0", exponent0));
    params.push_back (pair<string,double> ("Exponent1", exponent1));
    params.push_back (pair<string,double> ("Exponent2", exponent2));
    params.push_back (pair<string,double> ("ReferenceLoss", referenceLoss));
  }

  virtual double Loss (double distance) const;

  void SerializeTo (ofstream& ofs);
  static ThreeLogDistancePropagationLossModel* DeserializeFrom (ifstream& ifs);

  virtual const char* Name () const { return name; }

  // ns-3 default values
  static constexpr double default_distance0 = 1.0;
  static constexpr double default_distance1 = 200.0;
  static constexpr double default_distance2 = 500.0;
  static constexpr double default_exponent0 = 1.9;
  static constexpr double default_exponent1 = 3.8;
  static constexpr double default_exponent2 = 3.8;
  static constexpr double default_referenceLoss = 46.6777;       // Friis at 1m with 5.15 GHz

  static constexpr char name[] = "ThreeLogDistance";
};

/* class that holds power consumption and time usage (tau) of each radio for a given network node */
class TPower
{
public:
  TPower (int r, float p = 0.f) : power(p) { tau.resize(r,0.f); }

  float power;
  vector<float> tau;
};


class TopologyParser;



/*************************
 * NETWORK TOPOLOGY CLASS
 *************************/

class Topology
{
friend class TopologyParser;

public:
  /* constructors */
//   Topology(unsigned int _numAvailChannels = defaultNumChannels) : numAvailChannels(_numAvailChannels), noise(0.00001), maxP(numTxPowerLevels-1) {}

  Topology (const char* filename);
  
  /* copy constructor */
  Topology(const Topology&);		/* node information (e.g. MeshNode and RadioIface structures) are not duplicated  */

  /* saves the topology information to a file */
  void SerializeTo (const char *filename); 

  /* reads the topology from a file */
  static Topology* DeserializeFrom (const char *filename);

  const char* Toponame () const { return toponame.c_str(); }

  /* returns the nuber of aggregation nodes */
  int GetNumAggregationNodes ()  { return aggregation_list.size(); }

  /* returns the nuber of aggregation nodes */
  int GetNumGatewayNodes ()  { return gateway_list.size(); }

  /* returns true if n is an aggregation node */
  bool IsAggregationNode (int n) { return (find(aggregation_list.begin(), aggregation_list.end(), n) != aggregation_list.end()); }

  /* returns true if n is a gateway node */
  bool IsGatewayNode (int n) { return (find(gateway_list.begin(), gateway_list.end(), n) != gateway_list.end()); }

  /* returns the i-th aggregation node */
  int GetAggregationNode (int i) { return aggregation_list.at(i); }

  /* returns the i-th gateway node */
  int GetGatewayNode (int i) { return gateway_list.at(i); }

  /* returns the number of nodes */
  int GetNumNodes () { return node.size(); }

  /* returns an iterator pointing to the i-th node */
  TGraphNodeIterator Node (int i) { return node.begin() + i; }

  /* returns an iterator pointing to the first node */
  TGraphNodeIterator NodeBegin () { return node.begin(); }

  /* returns an iterator pointing to past the last node */
  TGraphNodeIterator NodeEnd () { return node.end(); }

  float GetThroughput () { return throughput; }

  void SetThroughput (float tput) { throughput = tput; }

  /* returns a TLinkItrs starting from a TGraphNodeIterator and a TGraphNeighborIterator */
  TLinkItrs LinkItrs (const TGraphNodeIterator &u, const TGraphNeighborIterator &v) { return TLinkItrs (this, u, v); }

  void addNeighbor(int u, int v, float rate, unsigned int power);

  int MaxAllowLinkRate (int u, int v, unsigned int power, int minrate, int maxrate);
  
  void buildGi_FixedPowerMaxAllowRate(unsigned int power, int minrate, int maxrate);
  
  int buildGi_FromFile(const char * filename);

  void findToBePossiblyAddedLinks(int maxp = WMS.numTxPowerLevels());

  void printTopology(ostream& f = cout);

  void printStats (const char* name, bool verbose);

  void drawTopology();

  /*
   * Computes the power (Watts) received by v when u transmits at the given power level.
   * Used by methods attempting to determine if a link between u and v using the given power may exist.
   */
  float ReceivedPower (int u, int v, unsigned int power);

  /*
   * Computes the power (Watts) received by node x when transmission occurs on link l.
   * If node x is not provided, it defaults to the receiver of link l.
   * If link l has been assigned a channel, this function uses the exact values of tx power, tx and rx gain.
   */
  float ReceivedPower (TLinkItrs l, int x = -1);

  /* returns the distance between two nodes */
  float dist(int x, int y);

  /* adjust the TX power of each interface so as to just reach the farthest neighbor reachable on the channel that interface is set to */
  // void AdjustPower();
  /* PRE: radios must have been assigned a channel */

  void ResetChannels();

  /*
   * Checks whether two nodes u and v can be neighbors using some values for TX power and rate
   * Starts from the highest TX rate and the lowest TX power, then increases the TX power
   * If the SNR is still below the threshold, lowers the TX rate
   * Thus, determines the highest possible capacity
   */
  bool canBeNeighbors(int u, int v, int & c, int & p, int maxp = WMS.numTxPowerLevels());

  /* Given a TLinkItrs u->v returns the TLinkItrs v->u */
  TLinkItrs FindInverseLink(TLinkItrs l, unsigned int channel = 0);

  /* Assign channel c to link l
  PRE: for both end nodes, either a radio assigned to c already exists or an available radio exists */
  void AssignChannel2Link(TLinkItrs l, unsigned int c);

  set<unsigned int> haveCommonChannels (int u, int v);
  
  /* Find all the links that potentially interfere with l */
  vector<TLinkItrs> computeNeighborhood(TLinkItrs l);

  /* computes the set of channels used in the topology */
  void getChannelSet(set<unsigned int> &channelSet);

  /* erases links with null flow */
  void eraseLinksWithNullFlow();

  /* returns true if at least one link has no channel assigned (i.e., channel 0) */
  bool isThereLinkWithoutChannel ();

  /*
   * Computes the Utot of all the links and returns true if they all match the value in the Utot field.
   * If vec is non-null, it points to a vector (that must be de-allocated after use) containing
   * all the pairs <link, computed Utot>. Values for verbose:
   * - 0: do not print anything to stdout
   * - 1: only print a message if the computed Utot differs from the value in the Utot field
   * - 2: print the Utot for all the links
   * - 3: print the collision domain of all the links
   */
  bool checkUtot (int verbose, vector< std::pair< TLinkItrs, double > >** vec = NULL); 

  /*
   * Computes the Utot of the collision domain of a link. If verbose is true,
   * prints all the links belonging to the collision domain of link l
   */
  double computeUtot(TLinkItrs l, bool verbose = false);

  /* returns the maximum among the utilization of the collision domains of all the links */
  pair<TLinkItrs,double> getUtotMax();

  /*
   * Update the Utot for every link involved (except l itself) in case l is assigned channel c and rate r.
   * Must be invoked before making the change
   */
  void updateUtot (TLinkItrs l, unsigned int c, float r);

  /* scales all the precomputed flow rates and the total utilizations by the max Utot */
  void scaleFlowsByUtotMax();

  /* attempts to make the graph connected if connect is true */
  bool isConnected(bool connect);

  /* computes the degree of the network (i.e. average number of neighbors per node) */
  double getDegree();

  int countRadiosOff();

  int countNodesOff();

  int countNumRadios();

  /* 
   * Determines how many radios change channel between the current configuration and Tnew 
   * oldChSet and newChSet contains, for each node, the channels differing between the two configurations
   */
  int countRadioChanged (Topology* Tnew, vector<multiset<unsigned int> >& oldChSet, vector<multiset<unsigned int> >& newChSet);

  /*
   * Computes the power consumption of all the nodes and the time usage (tau) for each of their radios.
   * Results are stored in the vector consumption (which can also be empty when passed)
   */
  void calcAllNodesPowerConsumption (vector<TPower>& consumption);

  /*
   * checks whether the power constraints on all the nodes and the time constraints (tau) on all the radios are satisfied.
   * If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
   * can be passed so that calcAllNodesPowerConsumption is not called again
   */
  bool checkPowerConstraints (vector<float>* nodePowerConstr, vector<TPower>* consumption = NULL);

  /*
   * computes the total power consumption of the whole network.
   * If the vector of power consumptions is available (i.e., calcAllNodesPowerConsumption has been called), its address
   * can be passed so that calcAllNodesPowerConsumption is not called again
   */
  float TotalPowerConsumption (vector<TPower>* consumption = NULL);

  /*
   * Applies Dijkstra's algorithm to the topology
   * Returns a vector containing the distance of each node to the source
   * Links for which prune is true are not considered
   * The link cost function is computed by the cost function object
   */
  vector<float> Dijkstra (int source, bool isSink = false, linkPruneFunc* prune = NULL, linkCostFunc* cost = NULL);	        /* applies Dijkstra to the topology */

  /*
   * Applies Dijkstra algorithm to topology T
   * Returns the shortest feasible path between source and dest
   */
  vector<TLinkItrs>* ShortestPath(int source, int dest, linkPruneFunc* prune = NULL, linkCostFunc* cost = NULL);

  /*
   * Find the first (at most k) shortest paths with length at most alpha times the length
   * of the shortest path. Returns the set of links on those paths.
   */
  set<pair<int,int> > find_path (pair<int,int> d, float alpha, int k);

  /* Returns the set of links on the shortest path */
  set<pair<int,int> > find_path (pair<int,int> d) { return find_path (d, 1.0f, 1); }

  /* Returns the set of links carrying the flow of the given demand */
  set<pair<int,int> > find_links_for_demand (pair<int,int> d);

  /*
   * Implement the RDAS algorithm. The last input parameter is a function pointer to determine how to build the initial DAG.
   * It defaults to &Topology::find_path, which returns the set of links on the shortest path. Alternatively,
   * &Topology::find_links_for_demand can be passed, which returns the set of links carrying the flow of the given demand
   */
  set<pair<int,int> > FindLooplessResilientDAG(int s, int d, float alpha, set<pair<int,int> > (Topology::*pmi) (pair<int,int>) = &Topology::find_path);

  /*
   * For each demand, determines the set of links carrying the associated flow and runs the RDAS algorithm to compute a
   * DAG that "protects" the involved nodes (each node has at least two neighbors for the given demand, for resiliency
   * purposes). This is useful if an ns-3 simulation with MPLS and NHLFE deactivation is to be run.
   * alpha is the multiplicative factor used by RDAS.
   */
  void AddNullFlowOnRDASlinks (float alpha);

  list<TLinkItrs>* changeFlowOnRandomPath(int maxLength, bool increase, double percent);

  /* 
   * If type = 2, swaps all the demands, in the sense that demands with less flow are assigned the flow of the demands with more flow.
   * Each demand is multiplied (type = 0 or 2) or set (type = 1) by a random value derived from the given distribution
   */
  void changeDemands(RndVar* RV, double seed, unsigned int type = 0);

  /* copies the flow value in the availBW field */
  void InitRouting();

  /* Route flows along the shortest path; AvailBW is used as link capacity; returns the cross-section goodput */
  double RouteFlows(int NumFlows, RndVar *Source, RndVar *Dest, RndVar *Bwdr, double seed);

  /* Checks whether the e2e flows are consistent */
  bool checkE2Eflows(map<pair<int,int>,double>* ret = NULL);


private:
  // default constructor, used by DeserializeFrom
  Topology () {}
  
  string toponame;
  string topopath;

  vector<int> aggregation_list;
  vector<int> gateway_list;

  vector<TGraphNode> node;

  float throughput;
};



/*
 * FRIEND CLASSES
 */

class RadioIfaceFriend
{
private:

  /* A list of operations allowed to friend classes */

  static void SetChannel (RadioIfaceIterator ri, unsigned int channel) { ri->channel = channel; }

friend class CAalgo;
friend class MILP;
};

class TGraphNeighborFriend
{
private:

  /* A list of operations allowed to friend classes */

  static void SetRadio (TGraphNeighborIterator ni, RadioIface* radio) { ni->radio = radio; }
  static void SetFlow (TGraphNeighborIterator ni, double f) { ni->flow = f; }
  static void SetUtot (TGraphNeighborIterator ni, double utot) { ni->Utot = utot; }
  static void SetRate (TGraphNeighborIterator ni, float rate) { ni->rate = rate; }
  static void SetTxPower (TGraphNeighborIterator ni, unsigned int txpower) { ni->txpower = txpower; }
  static void SetAvailBW (TGraphNeighborIterator ni, double bw) { ni->availBW = bw; }
  static void ClearFlowDemands (TGraphNeighborIterator ni) { ni->f_e2e.clear(); }
  static void SetFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { ni->f_e2e[d] = f; }
  static void IncreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { ni->f_e2e[d] += f; }
  static void DecreaseFlowDemand (TGraphNeighborIterator ni, pair<int,int> d, double f) { ni->f_e2e[d] -= f; }

friend class CAalgo;
friend class LP;
friend class MILP;
friend class PFR_method;
};

class TGraphNodeFriend
{
private:

  /* A list of operations allowed to friend classes */

  static TGraphNeighborIterator AddNeighbor (TGraphNodeIterator nodei, const TGraphNeighbor* ni) { return nodei->neighbor.insert (nodei->neighbor.end(), *ni); }
  static void EraseNeighbor (TGraphNodeIterator nodei, TGraphNeighborIterator ni) { nodei->neighbor.erase (ni); }

friend class CAalgo;
friend class MILP;
};

#endif
