#ifndef __UDP_TRACE_BASED_CLIENT__
#define __UDP_TRACE_BASED_CLIENT__

#include "ns3/application.h"
#include "ns3/event-id.h"
#include "ns3/ptr.h"
#include "ns3/ipv4-address.h"
#include <vector>
#include "ns3/application-container.h"
#include "ns3/node-container.h"

namespace ns3 {

class Socket;
class Packet;

/**
 * \ingroup udpclientserver
 * \class UdpTraceBasedClient
 * \brief A trace based streamer
 *
 * sends udp packets based on a trace file
 * A valid trace file is a file with 2 columns:
 * -1- the first one indicates the time (seconds) at which the packet must be sent
 * -2- the second one indicates the packet size in byte
 * if no valid trace file is provided to the application, no packet is sent
 */
class UdpTraceBasedClient : public Application
{
public:
  static TypeId
  GetTypeId (void);
  /**
   * \brief creates a traceBasedStreamer application
   */
  UdpTraceBasedClient ();

  /**
   * \brief creates a traceBasedStreamer application
   * \param ip the destination ip address to which the stream will be sent
   * \param port the destination udp port to which the stream will be sent
   * \param traceFile a path to a trace file formatted as follows:
   *  Time[s]    Length [byte]
   *  Time[s]    Length [byte]
   *  ...
   *
   *
   */

  UdpTraceBasedClient (Ipv4Address ip, uint16_t port, char *traceFile);
  ~UdpTraceBasedClient ();

  /**
   * \brief set the destination IP address and port
   * \param ip the destination ip address to which the stream will be sent
   * \param port the destination udp port to which the stream will be sent
   */
  void SetRemote (Ipv4Address ip, uint16_t port);

  /**
   * \brief set the trace file to be used by the application
   * \param filename a path to a trace file formatted as follows:
   *  Time[s]    Length [byte]
   *  Time[s]    Length [byte]
   *  ...
   */
  void SetTraceFile (std::string filename);

  /**
   * \return the maximum packet size
   */
  uint16_t GetMaxPacketSize(void);

  /**
   * \param maxPacketSize The maximum packet size
   */
  void SetMaxPacketSize(uint16_t maxPacketSize);

protected:
  virtual void DoDispose (void);

private:
  virtual void StartApplication (void);
  virtual void StopApplication (void);
  void Send (void);
  void SendPacket (uint32_t size);


  struct TraceEntry
  {
    double timeToSend;
    uint16_t packetSize;
  };
  
  uint32_t m_sent;
  Ptr<Socket> m_socket;
  Ipv4Address m_peerAddress;
  uint16_t m_peerPort;
  EventId m_sendEvent;
  std::vector<struct TraceEntry> m_entries;
  uint32_t m_currentEntry;
  uint16_t m_maxPacketSize;
};


/** Helper **/

/**
 * Create udpTraceClient application which sends udp packets based on a trace file
 * A valid trace file is a file with 2 columns:
 * -1- the first one indicates the time at which the packet must be sent
 * -2- the second one indicates the packet size in byte
*/
class UdpTraceBasedClientHelper
{
public:
  /**
   * Create UdpTraceBasedClientHelper which will make life easier for people trying
   * to set up simulations with udp-client-server.
   *
   */
  UdpTraceBasedClientHelper ();

  /**
   * Create UdpTraceBasedClientHelper which will make life easier for people trying
   * to set up simulations with udp-client-server.
   *
   * \param ip The IP address of the remote udp server
   * \param port The port number of the remote udp server
   * \param filename the file from which packet traces will be loaded
   */
  UdpTraceBasedClientHelper (Ipv4Address ip, uint16_t port, std::string filename);

  /**
    * Record an attribute to be set in each Application after it is is created.
    *
    * \param name the name of the attribute to set
    * \param value the value of the attribute to set
    */
  void SetAttribute (std::string name, const AttributeValue &value);

  /**
    * \param c the nodes
    *
    * Create one udp trace client application on each of the input nodes
    *
    * \returns the applications created, one application per input node.
    */
  ApplicationContainer Install (NodeContainer c);

private:
  ObjectFactory m_factory;
};

} // namespace ns3

#endif // __UDP_TRACE_BASED_CLIENT__
