#ifndef WIMESH_MPLS_NS3
#define WIMESH_MPLS_NS3

#include "wimesh-ns3.h"
#include "ns3/mpls-module.h"


// the following two classes store the split ratios passed to ConfigureForwarding
class TSplitRatio
{
public:
  TSplitRatio (pair<int,int> link, int channel, double ratio) :
    m_link(link),
    m_channel(channel),
    m_ratio(ratio) {}
    
  pair<int,int> getLink () const { return m_link; }

  int getChannel () const { return m_channel; }

  double getRatio () const { return m_ratio; }

private:
  pair<int,int> m_link;
  int m_channel;
  double m_ratio;
};

class DemSplitRatios
{
public:
  DemSplitRatios (pair<int,int> edges, double flow) :
    m_edges(edges),
    m_flow(flow),
    m_ratios() {}

  void AddSplitRatio (const TSplitRatio& sr) { m_ratios.push_back (sr); }

  pair<int,int> getEdges () const { return m_edges; }

  double getFlow () const { return m_flow; }

  int getNumRatios () const { return m_ratios.size(); }

  const TSplitRatio& getSplitRatio (int i) const { return m_ratios.at(i); }

private:
  pair<int,int> m_edges;
  double m_flow;
  vector<TSplitRatio> m_ratios;
};
  

/*
 * Given a WiMesh topology, configures an ns-3 simulation
 * using the MPLS splitting technique (requires a file
 * produced by SplitRatios as input)
 */

class WiMeshMplsNs3 : public WiMeshNs3
{
public:
  WiMeshMplsNs3 ();

  WiMeshNs3* Clone () { return new WiMeshMplsNs3 (*this); }

  virtual void PrintConfiguration ();

  void ReadSplitRatiosFromFile (istream& is);

  void CalcSplitRatiosFromTopology ();

  /*
   * Takes care of switching channels, updating IP addresses and MPLS tables to match
   * the new topology. Returns the number of radios changing channel
   */
  void ChangeTopology (Topology* T, int* changes = 0);

  /* If set to true, Nhlfes are deactivated upon a failed transmission (default: false) */
  void SetDisableNhlfe (bool disable);

  /* default: 1 second */
  void SetOutTime (double ot);

  const DemSplitRatios& getDemSplitRatios (int i) { return SplitRatiosVec.at(i); }

private:
  // used by ConfigureForwarding to store nhlfes and split ratios
  typedef struct {
    vector<mpls::Nhlfe> outgoing_nhlfes;
    vector<double> split_ratios;
    set<uint32_t> incoming_labels;
    // channel <-> index of the device using that channel
    map<uint16_t, uint32_t> IfIndex;
  } NodeInfo;

  vector<DemSplitRatios> SplitRatiosVec;

  MplsNetworkConfigurator mpls;

  // if true, an Nhlfe is deactivated on a failed transmission and then re-activated (default: false)
  bool disableNhlfe;

  /* when a transmission fails, the nhlfe is deactivated for out_time seconds (in order to recover from failures) */
  double out_time;

  void FindChannelIfIndexMap (int u, map<uint16_t, uint32_t>& m);
  
  // overridden from base class
  virtual void DoCreateNodes ();
  /*
   * For each mesh node, create a wifi device for each channel it uses
   * Each wifi device has the IP address 10.1.<channel>.<meshnode_id+1>
   */
  virtual void CreateNs3MeshDevices ();
  
  virtual void ConfigureForwarding ();

  /*
   * Searches all the NHLFEs associated with all the ILMs and FTNs configured on a node
   * for those having the same IP address as the (unreachable) neighbor device and deactivates them
   */
  virtual void DoReceiveFinalDataFailed (int me, Mac48Address neighbor);
};


#endif
